/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"
#include "util/critical.h"

#include "stm32f10x_rcc.h"
#include "stm32f10x_iwdg.h"

#include "FreeRTOS.h"
#include "task.h"

#include "hal/led.h"
#include "hal/usart1.h"
#include "hal/nrf24.h"
#include "wifi.h"
#include "constants.h"
#include "config.h"
#include "network/igmpv3.h"
#include "vision.h"

CLIGlobal cli;

void processCommand();

void CLIInit()
{
	cli.hProcBufMutex = xSemaphoreCreateRecursiveMutex();

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];
	static uint8_t printData[CONSOLE_PRINT_SIZE];
	static uint8_t scanData[CONSOLE_SCAN_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.pPrintBuf = printData;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.pScanBuf = scanData;
	conData.scanSize = CONSOLE_SCAN_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);
}

void CLITask(void* params)
{
	params = params;

	portTickType xLastWakeTime;
	portTickType xLastLEDTime;

	int16_t result = 0;
	uint16_t length;

	xLastWakeTime = xTaskGetTickCount();
	xLastLEDTime = xLastWakeTime;

	LEDOff();

	while(1)
	{
		TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
		{
			// check user input to terminal
			while(FifoPacketSyncedNumPackets(&usart1.rxFifo) > 0)
			{
				// get packet
				result = FifoPacketSyncedPop(&usart1.rxFifo, &length, cli.procBuf, CLI_PROC_SIZE-1);
				if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
				{
					FifoPacketSyncedDrop(&usart1.rxFifo);
					continue;
				}

				if(result)
					continue;

				ConsolePrint("\r\n");

				cli.procBuf[length] = '\0';

				// process onboard
				result = FifoPacketSyncedBegin(&console.rxFifo);
				if(result)
					continue;

				result = FifoPacketSyncedPush(&console.rxFifo, length, cli.procBuf);
				if(result)
				{
					FifoPacketSyncedCancel(&console.rxFifo);
					continue;
				}

				FifoPacketSyncedComplete(&console.rxFifo);
			}

			// check data from main console to terminal
			while(FifoPacketSyncedNumPackets(&console.txFifo) > 0)
			{
				result = FifoPacketSyncedPop(&console.txFifo, &length, cli.procBuf, CLI_PROC_SIZE);
				if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
				{
					FifoPacketSyncedDrop(&console.txFifo);
					continue;
				}

				if(result)
					continue;

				FifoSyncedPush(&usart1.txFifo, length, cli.procBuf);

				USART1StartTx();
			}
		}

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		// Heartbeat LED
		if(xLastWakeTime - xLastLEDTime > 500/portTICK_RATE_MS)
		{
			xLastLEDTime = xLastWakeTime;

			LEDToggle();
		}

		vTaskDelayUntil(&xLastWakeTime, 50/portTICK_RATE_MS);

		IWDG_ReloadCounter();
	}
}

void processCommand()
{
	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	if(ConsoleCmpCmd("task list"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tState\tPrio\tStack\tID");

		ConsoleLockPrintBuf();
		vTaskList((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	if(ConsoleCmpCmd("task stats"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tTicks\t\tRatio");

		ConsoleLockPrintBuf();
		vTaskGetRunTimeStats((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	if(ConsoleCmpCmd("led on"))
	{
		ConsolePrint("Setting LED on\r\n");

		LEDOn();
	}

	if(ConsoleCmpCmd("led off"))
	{
		ConsolePrint("Setting LED off\r\n");

		LEDOff();
	}

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Base Station - v1.0.0\r\n");
	}

	if(ConsoleCmpCmd("runtime"))
	{
		ConsolePrint("Last wifi runtime: %u.%04us\r\n", wifi.runtime/10000, wifi.runtime%10000);
	}

	uint8_t reg;
	if(ConsoleScanCmd("nrf read %hhX", &reg) == 1)
	{
		uint8_t value;
		NRF24ReadRegister(&wifi.nrf24, reg, &value);
		ConsolePrint("Reg 0x%hhX: 0x%hhX\r\n", reg, value);
	}

	if(ConsoleScanCmd("nrf id %hhu", &reg) == 1)
	{
		NRF24ChangeId(&wifi.nrf24, reg);

		ConsolePrint("New id: %hhu\r\n", reg);
	}

	if(ConsoleCmpCmd("igmp test"))
	{
		ConsolePrint("Sending IGMPv3 Membership Report\r\n");

		uint8_t data[IGMPV3_GROUP_RECORD_SIZE];

		IGMPv3GroupRecord* pRecord = (IGMPv3GroupRecord*)data;

		pRecord->auxDataLen = 0;
		pRecord->multicastAddress = IPv4AddressSet(225, 42, 42, 42);
		pRecord->numberOfSources = 0;
		pRecord->recordType = IGMPV3_MODE_IS_EXCLUDE;

		int16_t result = NetworkSendIGMPv3MembershipReport(1, data, IGMPV3_GROUP_RECORD_SIZE);

		ConsolePrint("Result: %d\r\n", result);
	}

	if(ConsoleCmpCmd("wifi scan"))
	{
		wifi.scan = 0;
	}

	if(ConsoleCmpCmd("wifi stats"))
	{
		ConsolePrint("-- TX --\r\n");
		ConsolePrint("Bytes:   %10u/%10u\r\n", wifi.nrf24.stats.txBytesSent, wifi.nrf24.stats.txBytesLost);
		ConsolePrint("Packets: %10u/%10u\r\n", wifi.nrf24.stats.txPacketsSent, wifi.nrf24.stats.txPacketsLost);
		ConsolePrint("Retrans: %10u\r\n", wifi.nrf24.stats.retransmissions);
		ConsolePrint("-- RX --\r\n");
		ConsolePrint("Bytes:   %10u/%10u\r\n", wifi.nrf24.stats.rxBytesRecv, wifi.nrf24.stats.rxBytesLost);
		ConsolePrint("Packets: %10u/%10u\r\n", wifi.nrf24.stats.rxPacketsRecv, wifi.nrf24.stats.rxPacketsLost);
	}

	if(ConsoleScanCmd("vision invert %hhu", &reg) == 1)
	{
		if(reg)
			vision.invertPos = 1;
		else
			vision.invertPos = 0;
	}

	if(ConsoleCmpCmd("config"))
	{
		ConsolePrint("--- Configuration ---\r\n");
		ConsolePrint("IP:      %hhu.%hhu.%hhu.%hhu:%hu\r\n", config.values.ip.u8[0], config.values.ip.u8[1],
				config.values.ip.u8[2], config.values.ip.u8[3], config.values.port);
		ConsolePrint("Freq. channel: %hhu (2.%03huGHz)\r\n", config.values.channel, ((uint16_t)config.values.channel)+400);
		ConsolePrint("Max. RT: %hhu\r\n", config.values.maxRt);

		if(config.values.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(config.values.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(config.values.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");

		ConsolePrint("ACK delay: %huus\r\n", config.values.ackDelayUs);
		ConsolePrint("Vision freq: %hhuHz\r\n", config.values.maxVisionRate);

		ConsolePrint("Max no-link tries: %hu\r\n", config.values.maxNoLinkTries);
	}

	IPv4Address ip;
	if(ConsoleScanCmd("set ip %hhu %hhu %hhu %hhu", ip.u8, ip.u8+1, ip.u8+2, ip.u8+3) == 4)
	{
		NetworkChangeIP(ip);

		ConsolePrint("New IP: %hhu.%hhu.%hhu.%hhu:%hu\r\n", config.values.ip.u8[0], config.values.ip.u8[1],
				config.values.ip.u8[2], config.values.ip.u8[3], config.values.port);
	}

	uint16_t port;
	if(ConsoleScanCmd("set port %hu", &port) == 1)
	{
		NetworkChangePort(port);

		ConsolePrint("New port: %hu\r\n", config.values.port);
	}

	uint8_t ch;
	if(ConsoleScanCmd("set ch %hhu", &ch) == 1)
	{
		WifiChangeChannel(ch);

		ConsolePrint("New channel: %hhu\r\n", config.values.channel);
	}

	uint8_t maxRt;
	if(ConsoleScanCmd("set rt %hhu", &maxRt) == 1)
	{
		WifiChangeMaxRt(maxRt);

		ConsolePrint("New max. retransmissions: %hhu\r\n", maxRt);
	}

	uint8_t speed;
	if(ConsoleScanCmd("set speed %hhu", &speed) == 1)
	{
		switch(speed)
		{
			case 0: speed = NRF24_SPEED_250K; break;
			case 1: speed = NRF24_SPEED_1M; break;
			default: speed = NRF24_SPEED_2M; break;
		}

		WifiChangeSpeed(speed);

		if(config.values.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(config.values.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(config.values.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");
	}

	uint16_t u16;
	if(ConsoleScanCmd("set ack delay %hu", &u16) == 1)
	{
		int16_t result = WifiChangeAckDelay(u16);

		ConsolePrint("ACK delay set to %huus, result: %hX\r\n", u16, result);
	}

	uint8_t u8;
	if(ConsoleScanCmd("set vision rate %hhu", &u8) == 1)
	{
		WifiSetVisionRate(u8);

		ConsolePrint("Vision rate: %hhu\r\n", u8);
	}

	if(ConsoleScanCmd("set no link %hu", &u16) == 1)
	{
		config.values.maxNoLinkTries = u16;

		ConsolePrint("Max no-link tries: %hu\r\n", u16);
	}

	if(ConsoleCmpCmd("config write"))
	{
		ConsolePrint("Writing config to EEPROM...");

		if(ConfigSave())
			ConsolePrint("failed\r\n");
		else
			ConsolePrint("OK\r\n");
	}

	if(ConsoleCmpCmd("config reset"))
	{
		ConfigReset();

		ConsolePrint("Config reset, please save and restart base station\r\n");
	}
}
