/*
 * eeprom.c
 *
 *  Created on: 07.03.2013
 *      Author: AndreR
 */

#include "eeprom.h"

#include "errors.h"
#include "util/critical.h"

#include "task.h"

#include <string.h>

#define EEPROM_SIZE	128 // Byte
#define EEPROM_PAGE_SIZE 16
#define EEPROM_CMD_SIZE 2

// command set
#define EEPROM_READ		0x03
#define EEPROM_WRITE	0x02
#define EEPROM_WREN		0x06

static void spiCallback(SPIHeader* pHeader, uint8_t* pData, void* pUserData)
{
	EEPROM* pEEPROM = (EEPROM*)pUserData;

	if(pHeader->cmd == EEPROM_READ)
	{
		memcpy(pEEPROM->pReadDst, pData+EEPROM_CMD_SIZE, pHeader->length-EEPROM_CMD_SIZE);

		xSemaphoreGive(pEEPROM->hOPComplete);
	}

	if(pHeader->cmd == EEPROM_WRITE)
	{
		xSemaphoreGive(pEEPROM->hOPComplete);
	}
}

int16_t EEPROMInit(EEPROM* pEEPROM, SPI* pSPI, GPIO_TypeDef* pCSPort, uint16_t csPin)
{
	pEEPROM->pSPI = pSPI;
	pEEPROM->hInUse = xSemaphoreCreateMutex();
	pEEPROM->pReadDst = 0;
	vSemaphoreCreateBinary(pEEPROM->hOPComplete);

	SPISlave slave;
	slave.cpha = 0;
	slave.cpol = 0;
	slave.csPin = csPin;
	slave.pCSPort = pCSPort;
	slave.prescaler = SPI_BaudRatePrescaler_32;
	slave.callback = &spiCallback;
	slave.pUserData = pEEPROM;

	int16_t result = SPIAddSlave(pSPI, slave, &pEEPROM->slaveId);
	if(result)
		return result;

	return 0;
}

int16_t EEPROMRead(EEPROM* pEEPROM, uint8_t address, uint8_t* pData, uint8_t length)
{
	if(address+length > EEPROM_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	TASK_MUTEX(pEEPROM->hInUse)	// this function is non-reentrant
	{
		pEEPROM->pReadDst = pData;

		SPIHeader header;
		header.cmd = EEPROM_READ;
		header.length = EEPROM_CMD_SIZE + length;
		header.slaveId = pEEPROM->slaveId;

		uint8_t cmd[EEPROM_CMD_SIZE];
		cmd[0] = header.cmd;
		cmd[1] = address;

		xSemaphoreTake(pEEPROM->hOPComplete, 0);	// clear semaphore

		int16_t result = SPIPutPacket(pEEPROM->pSPI, &header, EEPROM_CMD_SIZE, cmd);
		if(result)
			return result;

		if(xSemaphoreTake(pEEPROM->hOPComplete, 50/portTICK_RATE_MS) == pdFALSE)
			return ERROR_EEPROM_TIMEOUT;
	}

	return 0;
}

int16_t EEPROMWrite(EEPROM* pEEPROM, uint8_t address, uint8_t* pData, uint8_t length)
{
	if(address+length > EEPROM_SIZE)
		return ERROR_NOT_ENOUGH_MEMORY;

	int16_t result = 0;

	SPIHeader wrenHeader;
	wrenHeader.cmd = EEPROM_WREN;
	wrenHeader.length = 1;
	wrenHeader.slaveId = pEEPROM->slaveId;

	SPIHeader writeHeader;
	writeHeader.cmd = EEPROM_WRITE;
	writeHeader.slaveId = pEEPROM->slaveId;

	uint8_t writeCmd[2];
	writeCmd[0] = writeHeader.cmd;


	TASK_MUTEX(pEEPROM->hInUse)	// this function is non-reentrant
	{
		while(length > 0)
		{
			uint8_t pagePosition = address%EEPROM_PAGE_SIZE;	// position inside a page
			uint8_t pageSizeLeft = EEPROM_PAGE_SIZE-pagePosition;
			uint8_t writeSize = length;

			if(pageSizeLeft < length)	// we want to write more as in this one page fits
				writeSize = pageSizeLeft;

			// unlock write protection
			result = SPIPutPacket(pEEPROM->pSPI, &wrenHeader, 1, &wrenHeader.cmd);
			if(result)
				return result;

			xSemaphoreTake(pEEPROM->hOPComplete, 0);	// clear semaphore

			// write data
			writeHeader.length = EEPROM_CMD_SIZE+writeSize;
			writeCmd[1] = address;
			result = SPIPutPacket2(pEEPROM->pSPI, &writeHeader, EEPROM_CMD_SIZE, writeCmd, writeSize, pData);
			if(result)
				return result;

			if(xSemaphoreTake(pEEPROM->hOPComplete, 50/portTICK_RATE_MS) == pdFALSE)
				return ERROR_EEPROM_TIMEOUT;

			vTaskDelay(10/portTICK_RATE_MS);	// write delay specified by EEPROM

			address += writeSize;
			pData += writeSize;
			length -= writeSize;
		}
	}

	return 0;
}
