/*
 * console.c
 *
 *  Created on: 25.05.2011
 *      Author: AndreR
 */

#include "console.h"

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>

#include "util/mem.h"
#include "util/critical.h"

Console console;

uint8_t verifyCmdReady()
{
	if(console.cmdReady)
		return 1;

	TASK_MUTEX(console.hScanMutex)
	{
		if(FifoPacketSyncedNumPackets(&console.rxFifo) > 0)
		{
			uint16_t length;
			FifoPacketSyncedPop(&console.rxFifo, &length, console.data.pScanBuf, console.data.scanSize-1);
			console.data.pScanBuf[length] = '\0';
			console.cmdReady = 1;
			break;
		}
	}

	return console.cmdReady;
}

void ConsoleInit(ConsoleData data)
{
	console.data = data;

	console.hTxMutex = xSemaphoreCreateRecursiveMutex();
	console.hRxMutex = xSemaphoreCreateRecursiveMutex();
	console.hScanMutex = xSemaphoreCreateMutex();
	console.hPrintMutex = xSemaphoreCreateMutex();

	FifoPacketSyncedInit(&console.txFifo, console.data.pDataTx, console.data.txSize, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, console.hTxMutex);
	FifoPacketSyncedInit(&console.rxFifo, console.data.pDataRx, console.data.rxSize, &FifoPacketSyncedMutexLock, &FifoPacketSyncedMutexRelease, console.hRxMutex);

	console.cmdReady = 0;
}

uint16_t ConsoleNumCmdsReady()
{
	return FifoPacketSyncedNumPackets(&console.rxFifo);
}

uint8_t ConsoleIsCmdReady()
{
	if(console.cmdReady)
		return 1;

	return ConsoleNumCmdsReady() > 0 ? 1 : 0;
}

int16_t ConsoleCmdProcessed()
{
	console.cmdReady = 0;

	return 0;
}

int16_t ConsoleCmpCmd(const char* cmd)
{
	if(verifyCmdReady() == 0)
		return 0;

	int16_t result = 0;

	TASK_MUTEX(console.hScanMutex)
	{
		if(strcmp((char*)console.data.pScanBuf, cmd) == 0)
			result = 1;
	}

	return result;
}

int32_t ConsoleScanCmd(const char* fmt, ...)
{
	if(verifyCmdReady() == 0)
		return 0;

	va_list args;
	int32_t storedFields = 0;

	TASK_MUTEX(console.hScanMutex)
	{
		va_start(args, fmt);
		storedFields = vsscanf((const char*)console.data.pScanBuf, fmt, args);
		va_end(args);
	}

	return storedFields;
}


int32_t ConsolePrint(const char* fmt, ...)
{
	va_list args;
	int32_t bytesWritten;

	TASK_MUTEX(console.hPrintMutex)
	{
		va_start(args, fmt);
		bytesWritten = vsnprintf((char*)console.data.pPrintBuf, console.data.printSize, fmt, args);
		va_end(args);

		int16_t result = FifoPacketSyncedBegin(&console.txFifo);
		if(result)
			return 0;

		result = FifoPacketSyncedPush(&console.txFifo, bytesWritten, console.data.pPrintBuf);
		if(result)
		{
			FifoPacketSyncedCancel(&console.txFifo);
			return 0;
		}

		FifoPacketSyncedComplete(&console.txFifo);

		if(console.data.txCallback)
			(*console.data.txCallback)(console.data.pUserData);
	}

	return bytesWritten;
}

char* ConsoleGetPrintBuf()
{
	return (char*)console.data.pPrintBuf;
}

// this function assumes that the print buf is locked by user!
int16_t ConsoleSendPrintBuf()
{
	int16_t result = FifoPacketSyncedBegin(&console.txFifo);
	if(result)
		return result;

	result = FifoPacketSyncedPush(&console.txFifo, strlen(console.data.pPrintBuf), console.data.pPrintBuf);
	if(result)
	{
		FifoPacketSyncedCancel(&console.txFifo);
		return result;
	}

	FifoPacketSyncedComplete(&console.txFifo);

	if(console.data.txCallback)
		(*console.data.txCallback)(console.data.pUserData);

	return 0;
}

void ConsoleLockPrintBuf()
{
	xSemaphoreTake(console.hPrintMutex, portMAX_DELAY);
}

void ConsoleUnlockPrintBuf()
{
	xSemaphoreGive(console.hPrintMutex);
}
