/*
 * intercom.c
 *
 *  Created on: 10.11.2012
 *      Author: AndreR
 */

#include "intercom.h"

#include "critical.h"

#include "task.h"

static void usartRxIRQ(void* pUserData)
{
	Intercom* pIntercom = (Intercom*)pUserData;

	signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

	xSemaphoreGiveFromISR(pIntercom->hRxDone, &xHigherPriorityTaskWoken);

	if(xHigherPriorityTaskWoken == pdTRUE)
		taskYIELD();
}

void IntercomInit(Intercom* pIntercom, USART* pUSART, uint8_t* pProcBuf, uint16_t procLength, uint8_t ownAddr)
{
	pIntercom->pUSART = pUSART;
	pIntercom->numClientsUsed = 0;
	pIntercom->pUSART->callback = &usartRxIRQ;
	pIntercom->pUSART->pUserData = pIntercom;
	vSemaphoreCreateBinary(pIntercom->hRxDone);
	pIntercom->pProcBuf = pProcBuf;
	pIntercom->procLength = procLength;
	pIntercom->ownAddress = ownAddr;
}

int16_t IntercomSend(Intercom* pIntercom, IntercomHeader* pHeader, uint16_t dataLength, uint8_t* pData)
{
	TASK_MUTEX_RECURSIVE(pIntercom->pUSART->hTxInUse)
	{
		if(pIntercom->pUSART->txFifo.fifo.EmptyBytes < FIFO_PACKET_SYNCED_HEADER_SIZE + INTERCOM_HEADER_SIZE + dataLength)
		{
			// that data does not fit in our fifo :(
			return ERROR_FIFO_SYNCED_NOT_ENOUGH_MEMORY;
		}

		FifoPacketSyncedBegin(&pIntercom->pUSART->txFifo);
		FifoPacketSyncedPush(&pIntercom->pUSART->txFifo, INTERCOM_HEADER_SIZE, pHeader->_data);
		if(dataLength > 0)
			FifoPacketSyncedPush(&pIntercom->pUSART->txFifo, dataLength, pData);
		FifoPacketSyncedComplete(&pIntercom->pUSART->txFifo);

		return USARTStartTransfer(pIntercom->pUSART);
	}

	return 0;
}

int16_t IntercomAddClient(Intercom* pIntercom, IntercomClient client)
{
	if(pIntercom->numClientsUsed == INTERCOM_MAX_CLIENTS)
		return ERROR_INTERCOM_NO_FREE_SLOT;

	pIntercom->clients[pIntercom->numClientsUsed] = client;
	pIntercom->numClientsUsed++;

	return 0;
}

int16_t IntercomProcess(Intercom* pIntercom, int32_t blockTime)
{
	if(xSemaphoreTake(pIntercom->hRxDone, blockTime) == pdFALSE)
		return 0;

	uint16_t length;
	int16_t result;

	while(1)
	{
		result = FifoPacketSyncedPop(&pIntercom->pUSART->rxFifo, &length, pIntercom->pProcBuf, pIntercom->procLength);
		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
		{
			FifoPacketSyncedDrop(&pIntercom->pUSART->rxFifo);
			return ERROR_INTERCOM_DROPPED_PACKET;
		}

		if(result)
		{
			return 0;
		}

		IntercomHeader* pHeader = (IntercomHeader*)pIntercom->pProcBuf;
		uint8_t* pData = pIntercom->pProcBuf + INTERCOM_HEADER_SIZE;

		if(pHeader->address != pIntercom->ownAddress)
			continue; // that packet is not for us

		for(uint8_t i = 0; i < pIntercom->numClientsUsed; i++)
		{
			if(pIntercom->clients[i].section == pHeader->section)
				(*pIntercom->clients[i].rxCallback)(pHeader, length-INTERCOM_HEADER_SIZE, pData, pIntercom->clients[i].pUserData);
		}
	}

	return 0;
}
