/*
 * sicp.c
 *
 *  Created on: 04.12.2012
 *      Author: AndreR
 */

#include "sicp.h"

#include "errors.h"

#include "util/console.h"

#define STATE_IDLE				0
#define STATE_TX_WAIT_FOR_HANDLER	1
#define STATE_TX_READY_TO_SEND		2
#define STATE_TX_WAIT_FOR_ACK		3
#define STATE_RX_WAIT_FOR_DATA		4
#define STATE_RX_WAIT_FOR_ACK		5

static void usartRxCallback(void* pUserData)
{
	SICP* pSICP = (SICP*)pUserData;
	signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

	xSemaphoreGiveFromISR(pSICP->hNewRxData, &xHigherPriorityTaskWoken);

	if(xHigherPriorityTaskWoken == pdTRUE)
		taskYIELD();
}

void SICPInit(SICP* pSICP, USART* pUSART)
{
	pSICP->pUSART = pUSART;
	pSICP->numHandlersUsed = 0;
	pSICP->rxCallback = 0;
	pSICP->txCallback = 0;
	pSICP->hTransferInProgress = xSemaphoreCreateRecursiveMutex();
	vSemaphoreCreateBinary(pSICP->hNewRxData);
	pSICP->hStartAnswerQueue = xQueueCreate(1, 1);
	pSICP->pUSART->callback = &usartRxCallback;
	pSICP->pUSART->pUserData = pSICP;
	pSICP->rxState = STATE_IDLE;
	pSICP->txState = STATE_IDLE;
}

int16_t SICPStartTransfer(SICP* pSICP, uint16_t dataType, uint32_t dataSize, TxCallback txCb)
{
	if(pSICP->txState != STATE_IDLE)
		return ERROR_SICP_BUSY;

	// drain queue
	xQueueReset(pSICP->hStartAnswerQueue);

	pSICP->txState = STATE_TX_WAIT_FOR_HANDLER;

	// send start transfer package
	SICPHeader header;
	header.type = SICP_TYPE_START;
	header.seq = 0;
	header.start.dataType = dataType;
	header.start.dataSize = dataSize;

	int16_t result = USARTSend(pSICP->pUSART, sizeof(SICPHeader), (uint8_t*)&header);
	if(result)
		return result;

	uint8_t startAnswer = SICP_TYPE_START_NACK;

	if(xQueueReceive(pSICP->hStartAnswerQueue, &startAnswer, 5000/portTICK_RATE_MS) == pdFALSE)
	{
		pSICP->txState = STATE_IDLE;
		return ERROR_SICP_START_TIMEOUT;
	}

	if(startAnswer == SICP_TYPE_START_ACK)
	{
		pSICP->txCallback = txCb;
		pSICP->txState = STATE_TX_READY_TO_SEND;

		pSICP->txTransferStatus.dataSize = dataSize;
		pSICP->txTransferStatus.dataProcessed = 0;

		return 0;
	}
	else
	{
		pSICP->txState = STATE_IDLE;
		return ERROR_SICP_NO_HANDLER;
	}
}

int16_t SICPSendDataChunk(SICP* pSICP, uint8_t* pData, uint16_t dataSize)
{
	if(pSICP->txState != STATE_TX_READY_TO_SEND)
		return ERROR_SICP_BUSY;

	pSICP->txState = STATE_TX_WAIT_FOR_ACK;

	pSICP->txTransferStatus.dataProcessed += dataSize;

	SICPHeader header;
	header.type = SICP_TYPE_DATA;
	header.seq = 1;

	int16_t result = USARTSendConcat(pSICP->pUSART, sizeof(SICPHeader), (uint8_t*)&header, dataSize, pData);
//	int16_t result = USARTSend(pSICP->pUSART, sizeof(SICPHeader), (uint8_t*)&header);
	if(result)
	{
		pSICP->txState = STATE_TX_READY_TO_SEND;
		return ERROR_SICP_SEND_FAILED;
	}

	return 0;
}

int16_t SICPAckPacket(SICP* pSICP)
{
	if(pSICP->rxState != STATE_RX_WAIT_FOR_ACK)
		return ERROR_SICP_BUSY;

	SICPHeader header;
	header.type = SICP_TYPE_ACK;
	header.seq = 1;

	int16_t result = USARTSend(pSICP->pUSART, sizeof(SICPHeader), (uint8_t*)&header);
	if(result)
		return ERROR_SICP_SEND_FAILED;

	if(pSICP->rxTransferStatus.dataProcessed >= pSICP->rxTransferStatus.dataSize)
		pSICP->rxState = STATE_IDLE;
	else
		pSICP->rxState = STATE_RX_WAIT_FOR_DATA;

	return 0;
}

int16_t SICPInstallHandler(SICP* pSICP, NewTransferCallback cb)
{
	if(pSICP->numHandlersUsed == SICP_MAX_HANDLERS)
		return ERROR_SICP_NO_FREE_SLOT;

	pSICP->handlers[pSICP->numHandlersUsed] = cb;
	pSICP->numHandlersUsed++;

	return 0;
}

void SICPTask(void* params)
{
	SICP* pSICP = (SICP*)params;

	static SICPHeader header;

	while(42)
	{
		xSemaphoreTake(pSICP->hNewRxData, portMAX_DELAY);

		uint16_t dataLength = 0;
		FifoPacketSyncedGetNextPacketDataLength(&pSICP->pUSART->rxFifo, &dataLength);

		int16_t result = FifoPacketSyncedPeekData(&pSICP->pUSART->rxFifo, (uint8_t*)&header, sizeof(SICPHeader));
		if(result)
			continue;

//		ConsolePrint("Rx packet. Type: 0x%X, length: %d\r\n", header.type, dataLength);
//		vTaskDelay(500/portTICK_RATE_MS);

		if(pSICP->txState == STATE_TX_WAIT_FOR_HANDLER && (header.type == SICP_TYPE_START_ACK || header.type == SICP_TYPE_START_NACK))
		{
			xQueueSendToBack(pSICP->hStartAnswerQueue, &header.type, portMAX_DELAY);

			FifoPacketSyncedDrop(&pSICP->pUSART->rxFifo);

			continue;
		}

		if(pSICP->txState == STATE_TX_WAIT_FOR_ACK)
		{
			if(header.type == SICP_TYPE_ACK)
			{
				if(pSICP->txTransferStatus.dataProcessed >= pSICP->txTransferStatus.dataSize)
					pSICP->txState = STATE_IDLE;
				else
					pSICP->txState = STATE_TX_READY_TO_SEND;

				(*pSICP->txCallback)(pSICP, &header, pSICP->txTransferStatus);

				FifoPacketSyncedDrop(&pSICP->pUSART->rxFifo);
			}

			if(header.type == SICP_TYPE_NACK)
			{
				pSICP->txState = STATE_IDLE;

				FifoPacketSyncedDrop(&pSICP->pUSART->rxFifo);
			}

			continue;
		}

		if(pSICP->rxState == STATE_IDLE && header.type == SICP_TYPE_START)
		{
			result = SICP_TRANSFER_NACK;

			for(uint16_t i = 0; i < pSICP->numHandlersUsed; i++)
			{
				result = (*pSICP->handlers[i])(pSICP, &header, &pSICP->rxCallback);
				if(result == SICP_TRANSFER_ACCEPTED)
					break;
			}

			vTaskDelay(4/portTICK_RATE_MS);

			SICPHeader startAnswer;
			startAnswer.seq = 0;

			if(result == SICP_TRANSFER_ACCEPTED)
			{
				startAnswer.type = SICP_TYPE_START_ACK;
			}
			else
			{
				startAnswer.type = SICP_TYPE_START_NACK;
			}

			int16_t sendResult = USARTSend(pSICP->pUSART, sizeof(SICPHeader), (uint8_t*)&startAnswer);
			if(sendResult || result == SICP_TRANSFER_NACK)
			{
				pSICP->rxState = STATE_IDLE;

				ConsolePrint("Sending start answer failed.\r\n");
			}
			else
			{
				pSICP->rxState = STATE_RX_WAIT_FOR_DATA;
				pSICP->rxTransferStatus.dataSize = header.start.dataSize;
				pSICP->rxTransferStatus.dataProcessed = 0;
			}

			FifoPacketSyncedDrop(&pSICP->pUSART->rxFifo);

			continue;
		}

		if(pSICP->rxState == STATE_RX_WAIT_FOR_DATA && header.type == SICP_TYPE_DATA)
		{
			pSICP->rxTransferStatus.dataProcessed += dataLength-sizeof(SICPHeader);

			pSICP->rxState = STATE_RX_WAIT_FOR_ACK;

			(*pSICP->rxCallback)(pSICP, &header, pSICP->rxTransferStatus);

			continue;
		}

		FifoPacketSyncedDrop(&pSICP->pUSART->rxFifo);
	}
}
