/*
 * sicp.h
 *
 *  Created on: 04.12.2012
 *      Author: AndreR
 *
 *	Serial Intercommunication Protocol
 *	Used between to USART to transfer large data
 */

#ifndef SICP_H_
#define SICP_H_

#include "stm32f4xx.h"

#include "FreeRTOS.h"
#include "semphr.h"
#include "task.h"

#include "usart.h"

#define SICP_MAX_HANDLERS 10

#define SICP_TYPE_START 		1
#define SICP_TYPE_ACK			2
#define SICP_TYPE_DATA			3
#define SICP_TYPE_NACK			4
#define SICP_TYPE_START_ACK		5
#define SICP_TYPE_START_NACK	6

#define SICP_TRANSFER_ACCEPTED	0
#define SICP_TRANSFER_NACK		1

typedef struct _SICPHeader
{
	uint8_t type;
	uint32_t seq;

	struct
	{
		uint16_t dataType;
		uint32_t dataSize;
	} __attribute__ ((packed)) start;
} __attribute__ ((packed)) SICPHeader;

typedef struct _SICPTransferStatus
{
	uint32_t dataSize;
	uint32_t dataProcessed;
} SICPTransferStatus;

struct _SICP;

/**
 * Called everytime a new data chunk is received.
 * Data must be taken from USART fifo. Keep in mind fifo data starts with SICPHeader.
 * After processing the data the client must call SICPAckPacket to signal the other side it is done.
 */
typedef void(*DataRxCallback)(struct _SICP*, SICPHeader*, SICPTransferStatus);

/**
 * Called after the sent chunk of data has been (n)acknowledged by the other side.
 * Usually it is used to send the next piece of data via SICPSendDataChunk.
 */
typedef void(*TxCallback)(struct _SICP*, SICPHeader*, SICPTransferStatus);

/**
 * New Transfer callback.
 * After a START packet is received the SICP Task scans all NewTransferCallbacks and
 * checks if one is able to handle this request.
 * The Handler then calls SICPAcceptTransfer to install a data handler and returns
 * TRANSFER_ACCEPTED. Otherwise it returns TRANSFER_NACK.
 */
typedef int16_t(*NewTransferCallback)(struct _SICP*, SICPHeader*, DataRxCallback* pCb);

// SICP structure
typedef struct _SICP
{
	USART* pUSART;

	NewTransferCallback handlers[SICP_MAX_HANDLERS];
	uint16_t numHandlersUsed;

	// current transfer callbacks, cleared after transfer is done
	DataRxCallback rxCallback;
	TxCallback txCallback;

	// current transfer infos
	SICPTransferStatus rxTransferStatus;
	SICPTransferStatus txTransferStatus;

	xSemaphoreHandle hTransferInProgress;
	xSemaphoreHandle hNewRxData;
	xQueueHandle hStartAnswerQueue;

	volatile uint8_t rxState;
	volatile uint8_t txState;
} SICP;

// initialize SICP structure
void	SICPInit(SICP* pSICP, USART* pUSART);

/**
 * Start a new transfer.
 *
 * @param pSICP SICP structure
 * @param dataType User data type to identify content
 * @param dataSize Complete data length to be transmitted
 * @param txCb Callback for handling ACKs and NACKs from the other side.
 * @return 0 on success, NO_HANDLER if no handler for this type is installed on the receiving side
 */
int16_t SICPStartTransfer(SICP* pSICP, uint16_t dataType, uint32_t dataSize, TxCallback txCb);

/**
 * Send a chunk of data.
 *
 * @param pSICP SICP structure
 * @param pData pointer to the data
 * @param dataSize data size to transmit in this chunk
 * @return 0 on success
 */
int16_t SICPSendDataChunk(SICP* pSICP, uint8_t* pData, uint16_t dataSize);

/**
 * Acknowledge the last received packet.
 * This also signals the processing of data is done. The other side
 * can now send more data if available.
 * Most commonly called from within the DataRxCallback.
 *
 * @param pSICP SICP structure
 * @return 0 on success
 */
int16_t SICPAckPacket(SICP* pSICP);

/**
 * Register a new client which could possibly handle some transfer types.
 *
 * @param pSICP SICP structure
 * @param cb Callback is called after a new START header has been received.
 * @return 0 on success
 */
int16_t SICPInstallHandler(SICP* pSICP, NewTransferCallback cb);

void SICPTask(void* params);

#endif /* SICP_H_ */
