/*
 * spi.c
 *
 *  Created on: 06.03.2011
 *      Author: AndreR
 */

#include "spi.h"
#include "errors.h"
#include "critical.h"
#include "bit_banding.h"
#include "mem.h"
#include "dma_gen.h"

#define SPIClearSPE(pSPI)	BBPeriphReset(pSPI->base, 6)
#define SPISetSPE(pSPI)		BBPeriphSet(pSPI->base, 6)
#define SPIGetSPE(pSPI)		BBPeriphGet(pSPI->base, 6)

#define DMAClearEn(pDMA)	BBPeriphReset((uint32_t)&pDMA->CR, 0)
#define DMASetEn(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 0)
#define DMAGetEn(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 0)

#define DMAClearTCI(pDMA) 	BBPeriphReset((uint32_t)&pDMA->CR, 4)
#define DMASetTCI(pDMA)		BBPeriphSet((uint32_t)&pDMA->CR, 4)
#define DMAGetTCI(pDMA)		BBPeriphGet((uint32_t)&pDMA->CR, 4)

static int16_t startTransfer(SPI* pSPI);

uint8_t rxFifoLock(void* pData)
{
	SPI* pSPI = (SPI*)pData;

	uint8_t state = 0;

	// make sure the following if is not interrupted by a task switch
	TASK_CRITICAL_SECTION()
	{
		if(DMAGetTCI(pSPI->pDmaRx))
		{
			state = 1;
			DMAClearTCI(pSPI->pDmaRx);
		}
	}

	return state;
}

void rxFifoRelease(uint8_t state, void* pData)
{
	SPI* pSPI = (SPI*)pData;

	if(state)
		DMASetTCI(pSPI->pDmaRx);
}

// IRQ -> no task switch can occur
int16_t SPIDmaRxInterrupt(SPI* pSPI)
{
//	SPIClearSPE(pSPI);
	SPISlave* pSlave = &pSPI->slaves[pSPI->currentPacket.slaveId];
	pSlave->pCSPort->BSRRL = pSlave->csPin;

	DMAClearEn(pSPI->pDmaRx);
	DMAClearEn(pSPI->pDmaTx);
	*pSPI->pDMARxIFCR |= pSPI->IFCRRxMask; // clear rx flags
	*pSPI->pDMATxIFCR |= pSPI->IFCRTxMask; // clear tx flags

	// RX done, complete packet
	// begin new RX packet
	if(pSPI->rxFifo.fifo.EmptyBytes < pSPI->currentPacket.length + SPI_HEADER_SIZE + FIFO_PACKET_SYNCED_HEADER_SIZE)
	{
		pSPI->busy = 0;
		startTransfer(pSPI);
		return ERROR_NOT_ENOUGH_MEMORY;
	}

	FifoPacketSyncedBegin(&pSPI->rxFifo);
	FifoPacketSyncedPush(&pSPI->rxFifo, SPI_HEADER_SIZE, pSPI->currentPacket._data);
	FifoPacketSyncedPush(&pSPI->rxFifo, pSPI->currentPacket.length, pSPI->data.pDmaDataRx);
	FifoPacketSyncedComplete(&pSPI->rxFifo);

	pSPI->busy = 0;

	if(pSPI->callback)
		(*pSPI->callback)();

	int16_t result = startTransfer(pSPI);

	return result;
}

void SPIInit(SPI* pSPI, SPI_TypeDef* pRegister, SPIData data, SPIPacketDoneCallback doneCallback)
{
	pSPI->pRegister = pRegister;
	pSPI->base = (uint32_t)pRegister;

	pSPI->data = data;

	FifoPacketSyncedInit(&pSPI->txFifo, data.pDataTx, data.size, &rxFifoLock, &rxFifoRelease, pSPI);
	FifoPacketSyncedInit(&pSPI->rxFifo, data.pDataRx, data.size, &rxFifoLock, &rxFifoRelease, pSPI);

	pSPI->hTxFifoInUse = xSemaphoreCreateMutex();
	pSPI->hExclusiveInUse = xSemaphoreCreateMutex();

	pSPI->numSlavesUsed = 0;
	pSPI->exclusiveSlave = 0xFF;	// none
	pSPI->busy = 0;

	pSPI->callback = doneCallback;

	// generate DMA_Stream_TypeDef from DMAx_BASE
	pSPI->pDmaRx = DMAGenerateStreamTypeDef(pSPI->data.pDma, pSPI->data.dmaChRx);
	pSPI->pDmaTx = DMAGenerateStreamTypeDef(pSPI->data.pDma, pSPI->data.dmaChTx);
	pSPI->pDMARxIFCR = DMAGenerateIFCR(pSPI->data.pDma, pSPI->data.dmaChRx);
	pSPI->pDMATxIFCR = DMAGenerateIFCR(pSPI->data.pDma, pSPI->data.dmaChTx);
	pSPI->IFCRRxMask = DMAGenerateMaskAll(pSPI->data.dmaChRx);
	pSPI->IFCRTxMask = DMAGenerateMaskAll(pSPI->data.dmaChTx);

	// Select correct channel
	if(pRegister == SPI1) // DMA2 Channel3
	{
		pSPI->pDmaRx->CR = DMA_SxCR_CHSEL_0 | DMA_SxCR_CHSEL_1;
		pSPI->pDmaTx->CR = DMA_SxCR_CHSEL_0 | DMA_SxCR_CHSEL_1;
	}
	if(pRegister == SPI2 || pRegister == SPI3) // DMA1 Channel0
	{
		pSPI->pDmaRx->CR = 0;
		pSPI->pDmaTx->CR = 0;
	}

	// configure DMA
	// Medium Priority, Memory Increment, Memory-to-peripheral
	pSPI->pDmaTx->CR |= DMA_SxCR_PL_0 | DMA_SxCR_MINC | DMA_SxCR_DIR_0;
	pSPI->pDmaTx->NDTR = 0;	// Number of data to transfer
	pSPI->pDmaTx->PAR = (uint32_t)&pSPI->pRegister->DR;	// peripheral address
	pSPI->pDmaTx->M0AR = (uint32_t)pSPI->data.pDmaDataTx;	// memory address

	// Medium Priority, Memory Increment, Peripheral-to-Memory, Transfer complete  interrupt enable
	pSPI->pDmaRx->CR |= DMA_SxCR_PL_0 | DMA_SxCR_MINC | DMA_SxCR_TCIE;
	pSPI->pDmaRx->NDTR = 0;
	pSPI->pDmaRx->PAR = (uint32_t)&pSPI->pRegister->DR;
	pSPI->pDmaRx->M0AR = (uint32_t)pSPI->data.pDmaDataRx;

	pSPI->pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI;
	pSPI->pRegister->CR2 = SPI_CR2_TXDMAEN | SPI_CR2_RXDMAEN;

	*pSPI->pDMARxIFCR |= pSPI->IFCRRxMask; // clear rx flags
	*pSPI->pDMATxIFCR |= pSPI->IFCRTxMask; // clear tx flags
}

int16_t SPIPutPacket(SPI* pSPI, SPIHeader* pHeader, uint16_t dataLength, uint8_t* pData)
{
	if(dataLength > pHeader->length)
		return ERROR_SPI_INVALID_LENGTH;

	if(pHeader->length > pSPI->data.dmaSize)
		return ERROR_SPI_INVALID_LENGTH;

	if(pHeader->slaveId >= pSPI->numSlavesUsed)
		return ERROR_SPI_INVALID_SLAVE;

	if(pHeader->length == 0)
		return ERROR_SPI_INVALID_LENGTH;

	if(pSPI->exclusiveSlave != 255 && pHeader->slaveId != pSPI->exclusiveSlave)
		return ERROR_SPI_EXCLUSIVE_LOCK;

	TASK_MUTEX(pSPI->hTxFifoInUse)
	{
		int16_t result = FifoPacketSyncedBegin(&pSPI->txFifo);
		if(result)
			return result;

		result = FifoPacketSyncedPush(&pSPI->txFifo, SPI_HEADER_SIZE, pHeader->_data);
		if(result)
		{
			FifoPacketSyncedCancel(&pSPI->txFifo);
			return result;
		}

		result = FifoPacketSyncedPush(&pSPI->txFifo, dataLength, pData);
		if(result)
		{
			FifoPacketSyncedCancel(&pSPI->txFifo);
			return result;
		}

		for(uint8_t i = 0; i < pHeader->length - dataLength; i++)
		{
			result = FifoPacketSyncedPushByte(&pSPI->txFifo, 0);
			if(result)
			{
				FifoPacketSyncedCancel(&pSPI->txFifo);
				return result;
			}
		}

		result = FifoPacketSyncedComplete(&pSPI->txFifo);
		if(result)
		{
			FifoPacketSyncedCancel(&pSPI->txFifo);
			return result;
		}

		return startTransfer(pSPI);
	}

	return 0;
}

int16_t SPIAddSlave(SPI* pSPI, SPISlave slave, uint8_t* pSlaveId)
{
	if(pSPI->numSlavesUsed == SPI_MAX_SLAVES)
		return ERROR_SPI_NO_FREE_SLOT;

	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Pin = slave.csPin;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	GPIO_Init(slave.pCSPort, &gpioInit);

	slave.pCSPort->BSRRL = slave.csPin;

	*pSlaveId = pSPI->numSlavesUsed;

	pSPI->slaves[pSPI->numSlavesUsed] = slave;

	pSPI->numSlavesUsed++;

	return 0;
}

int16_t	SPISetExclusiveSlave(SPI* pSPI, uint8_t slaveId)
{
	TASK_MUTEX(pSPI->hExclusiveInUse)
	{
		if(pSPI->exclusiveSlave != 255)
			return ERROR_SPI_EXCLUSIVE_LOCK;

		pSPI->exclusiveSlave = slaveId;
	}

	return 0;
}

int16_t	SPIReleaseExclusiveSlave(SPI* pSPI)
{
	TASK_MUTEX(pSPI->hExclusiveInUse)
	{
		pSPI->exclusiveSlave = 255;
	}

	return 0;
}

int16_t SPISetExclusiveConfig(SPI* pSPI)
{
	if(pSPI->exclusiveSlave == 255)
		return ERROR_SPI_INVALID_SLAVE;

	if(!SPIIdle(pSPI))
		return ERROR_SPI_BUSY;

	// get slave
	SPISlave* pSlave = &pSPI->slaves[pSPI->exclusiveSlave];

	// configure SPI
	pSPI->pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_SPE;	// reset all bits except master, disable SPI
	pSPI->pRegister->CR1 |=  pSlave->prescaler | pSlave->cpol << 1 | pSlave->cpha;

	return 0;
}

int16_t SPINotifySlaveCallbacks(SPI* pSPI)
{
	uint16_t bytesRead;
	int16_t result = 0;

	while(pSPI->rxFifo.numPackets > 0)
	{
		result = FifoPacketSyncedPop(&pSPI->rxFifo, &bytesRead, pSPI->data.pCbData, pSPI->data.cbSize);
		if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
		{
			FifoPacketSyncedDrop(&pSPI->rxFifo);
			continue;
		}

		SPIHeader* pHeader = (SPIHeader*)pSPI->data.pCbData;
		uint8_t* pData = pSPI->data.pCbData + SPI_HEADER_SIZE;

		if(pHeader->slaveId >= pSPI->numSlavesUsed)
			continue;

		if(pSPI->slaves[pHeader->slaveId].callback == 0)
			continue;

		(*pSPI->slaves[pHeader->slaveId].callback)(pHeader, pData, pSPI->slaves[pHeader->slaveId].pUserData);
	}

	return 0;
}

int16_t SPIGetPrescaler(SPI* pSPI, uint8_t slaveId, uint16_t* pPrescaler)
{
	if(slaveId >= pSPI->numSlavesUsed)
		return ERROR_SPI_INVALID_SLAVE;

	*pPrescaler = pSPI->slaves[slaveId].prescaler;

	return 0;
}

int16_t SPISetPrescaler(SPI* pSPI, uint8_t slaveId, uint16_t prescaler)
{
	if(slaveId >= pSPI->numSlavesUsed)
		return ERROR_SPI_INVALID_SLAVE;

	pSPI->slaves[slaveId].prescaler = prescaler;

	return 0;
}

uint16_t SPINumPacketsToProcess(SPI* pSPI)
{
	return pSPI->txFifo.numPackets;
}

uint8_t SPIIdle(SPI* pSPI)
{
	return (pSPI->txFifo.numPackets == 0 && pSPI->busy == 0);
}

static int16_t startTransfer(SPI* pSPI)
{
	if(pSPI->busy)
		return 0;

	pSPI->busy = 1;

	uint16_t bytesRead;

	// pop full data into rx (misuse)
	int16_t result = FifoPacketSyncedPop(&pSPI->txFifo, &bytesRead, pSPI->data.pDmaDataRx, pSPI->data.dmaSize);
	if(result == ERROR_FIFO_PACKET_SYNCED_EMPTY)
	{
		pSPI->busy = 0;
		return 0;
	}
	if(result)
	{
		pSPI->busy = 0;
		return result;
	}

	pSPI->currentPacket = *((SPIHeader*)pSPI->data.pDmaDataRx);
	uint16_t length = pSPI->currentPacket.length;

	SPISlave* pSlave = &pSPI->slaves[pSPI->currentPacket.slaveId];

	// select slave
	pSlave->pCSPort->BSRRH = pSlave->csPin;

	// copy only data to correct tx buffer
	MemCpy(pSPI->data.pDmaDataRx + SPI_HEADER_SIZE, pSPI->data.pDmaDataTx, length);

	// configure SPI
	pSPI->pRegister->CR1 = SPI_CR1_MSTR | SPI_CR1_SSM | SPI_CR1_SSI | SPI_CR1_SPE;	// reset all bits except master, disable SPI
	pSPI->pRegister->CR1 |=  pSlave->prescaler | pSlave->cpol << 1 | pSlave->cpha;

	// configure DMA
	pSPI->pDmaTx->NDTR = length;
	pSPI->pDmaRx->NDTR = length;

	DMASetEn(pSPI->pDmaRx);
	DMASetEn(pSPI->pDmaTx);

	return 0;
}
