/*
 * spi.c
 *
 *  Created on: 16.03.2011
 *      Author: AndreR
 */

#include <avr/interrupt.h>

#include "motor_data.h"
#include "spi.h"

#define SPI_DATA_SIZE MOTOR_DATA_SIZE/2+3

#define START_FLAG	0x3C
#define END_FLAG	0xA5

uint8_t spiRxBuf[SPI_DATA_SIZE];
uint8_t spiTxBuf[SPI_DATA_SIZE];
uint8_t spiRxProcBuf[SPI_DATA_SIZE];
volatile uint8_t* pRx;
volatile uint8_t* pTx;

volatile uint8_t spiActive;
volatile uint8_t newData;

ISR(SPI_STC_vect, ISR_NAKED)
{
	asm volatile(
		"out %0, r11		\n\t"	// this must be ultra-fast as the SPI transmitter in the AVR is not buffered
		"push r30			\n\t"
		"push r31			\n\t"
		"in r11, %0			\n\t"
		"lds r30, pRx		\n\t"
		"lds r31, pRx+1		\n\t"
		"st Z+, r11			\n\t"
		"sts pRx, r30		\n\t"
		"sts pRx+1, r31		\n\t"
		"lds r30, pTx		\n\t"
		"lds r31, pTx+1		\n\t"
		"ld r11, Z+			\n\t"
		"sts pTx, r30		\n\t"
		"sts pTx+1, r31		\n\t"
		"pop r31			\n\t"
		"pop r30			\n\t"
		"reti				\n\t"
			:
			: "M"(_SFR_IO_ADDR(SPDR)) );
}

ISR(PCINT0_vect)
{
	if(PINB & (1 << PB2))	// CS
	{
		// CS went high, slave deselected
		*pRx = SPDR;
		pTx = spiTxBuf+1;
		pRx = spiRxBuf;
		spiActive = 0;
		newData = 1;
	}
	else
	{
		// CS went low, slave selected
		SPDR = START_FLAG;
		nextTx = *spiTxBuf;
		spiActive = 1;
	}
}

void SPIInit()
{
	// init SPI
	DDRB |= (1 << PB4);

	spiActive = 0;
	newData = 0;
	SPCR = (1 << SPIE) | (1 << SPE);

	// init PCINT for CS pin
	PCMSK0 |= (1 << PCINT2);
	PCICR |= (1 << PCIE0);

	for(uint8_t i = 0; i < SPI_DATA_SIZE; i++)
	{
		spiRxBuf[i] = 0x00;
		spiTxBuf[i] = 0x00;
		spiRxProcBuf[i] = 0x00;
	}

	pRx = spiRxBuf;
	pTx = spiTxBuf;
	SPDR = START_FLAG;
}

static void setTx(uint8_t* pData, uint8_t length)
{
	uint8_t i;
	uint8_t sendCRC = 0;

	if(spiActive)
		return;

	for(i = 0; i < length; i++)
	{
		sendCRC ^= pData[i];
		spiTxBuf[i] = pData[i];
	}

	spiTxBuf[i] = sendCRC;
	spiTxBuf[i+1] = END_FLAG;
}

static uint8_t getRx(uint8_t* pData, uint8_t length)
{
	uint8_t i;
	uint8_t crc = 0;

	if(spiActive)
		return 1;

	if(newData == 0)
		return 1;

	newData = 0;

	for(i = 0; i < SPI_DATA_SIZE; i++)
	{
		spiRxProcBuf[i] = spiRxBuf[i];
	}

	if(spiRxProcBuf[0] != START_FLAG || spiRxProcBuf[length+2] != END_FLAG)
		return 1;	// Error, FLAG mismatch

	for(i = 1; i < length+1; i++)
	{
		crc ^= spiRxProcBuf[i];
	}

	if(crc != spiRxProcBuf[i])
		return 1;	// Error, CRC mismatch

	for(i = 0; i < length; i++)
	{
		pData[i] = spiRxProcBuf[i+1];
	}

	return 0;
}

void SPIExchange(uint8_t* pMisoData, uint8_t* pMosiData, uint8_t length)
{
	setTx(pMisoData, length/2);
	getRx(pMosiData, length/2);
}
