/*
 * cli.c
 *
 *  Created on: 29.05.2010
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"

#include "stm32f4xx_rcc.h"

#include "FreeRTOS.h"
#include "task.h"

#include "atmega/motor_prog.h"
#include "atmega/power_prog.h"
#include "atmega/kicker_prog.h"
#include "hal/buzzer.h"
#include "hal/motor.h"
#include "power.h"
#include "kicker.h"
#include "hal/led.h"
#include "commands.h"
#include "util/fastmath.h"
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "util/critical.h"
#include "util/fifo8_synced.h"
#include "hal/usart1.h"
#include "hal/usart2.h"
#include "motor_ir.h"
#include "hal/enc.h"
#include "self_check.h"
#include "ctrl.h"
#include "ctrl_motor.h"
#include "ctrl_fusion.h"
#include "ctrl_test_step.h"
#include "ctrl_fusion_vel.h"
#include "dribbler.h"
#include "hal/l3g.h"
#include "hal/lis3.h"
#include "util/boot.h"
#include "config.h"
#include "util/crc.h"

#include "network.h"

#include <stdio.h>
#include <string.h>

CLIGlobal cli;

static void processCommand();

#define COLOR_NONE	0
#define COLOR_RED	1
#define COLOR_GREEN	2
#define COLOR_BLUE	3

#define COLOR_MEDIA COLOR_BLUE
#define COLOR_MAIN COLOR_NONE

static void echoColor(uint8_t color)
{
	TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
	{
		switch(color)
		{
		case COLOR_RED:
			sprintf((char*)cli.procBuf, "\e[0;31m");
			break;
		case COLOR_GREEN:
			sprintf((char*)cli.procBuf, "\e[0;32m");
			break;
		case COLOR_BLUE:
			sprintf((char*)cli.procBuf, "\e[0;36m");
			break;
		default:
			sprintf((char*)cli.procBuf, "\e[0m");
			break;
		}

		FifoSyncedPush(&usart1.txFifo, strlen((char*)cli.procBuf), cli.procBuf);
	}
}

static void echoTargetColor(uint8_t addr)
{
	switch(addr)
	{
	case INTERCOM_ADDRESS_MEDIA:
		echoColor(COLOR_MEDIA);
		break;
	default:
		echoColor(COLOR_MAIN);
		break;
	}
}

// IRQ Level: USART1 IRQ
static void usart1Callback()
{
	// called on new complete packet in usart1 RX fifo
}

// IRQ Level: Task
static void consoleCallback()
{
	// called after ConsolePrint
}

// IRQ Level: Task
static void intercomCallback(IntercomHeader* pHeader, uint16_t length, uint8_t* pData, void* pUserData)
{
	pUserData = pUserData;

	// called after new console intercom rx
	TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
	{
		echoTargetColor(pHeader->command);
		FifoSyncedPush(&usart1.txFifo, length, pData);
		echoTargetColor(cli.currentTarget);

		USART1StartTx();
	}
}


void CLIInit()
{
	cli.currentTarget = INTERCOM_ADDRESS_MAIN;
	cli.hProcBufMutex = xSemaphoreCreateRecursiveMutex();

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];
	static uint8_t printData[CONSOLE_PRINT_SIZE];
	static uint8_t scanData[CONSOLE_SCAN_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.pPrintBuf = printData;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.pScanBuf = scanData;
	conData.scanSize = CONSOLE_SCAN_SIZE;
	conData.txCallback = &consoleCallback;
	conData.pUserData = 0;

	ConsoleInit(conData);

	usart1.callback = &usart1Callback;

	IntercomClient client;
	client.pUserData = 0;
	client.rxCallback = &intercomCallback;
	client.section = INTERCOM_SECTION_CONSOLE;

	IntercomAddClient(&intercom1, client);
}

void CLITask(void* params)
{
	params = params;

	portTickType xLastWakeTime;
	portTickType xLastLEDTime;

	int16_t result = 0;
	uint16_t length;

	xLastWakeTime = xTaskGetTickCount();
	xLastLEDTime = xLastWakeTime;

	LEDMustOff(LED_MAIN_RED);

	while(1)
	{
		TASK_MUTEX_RECURSIVE(cli.hProcBufMutex)
		{
			// check user input to terminal
			while(FifoPacketSyncedNumPackets(&usart1.rxFifo) > 0)
			{
				// get packet
				result = FifoPacketSyncedPop(&usart1.rxFifo, &length, cli.procBuf, CLI_PROC_SIZE-1);
				if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
				{
					FifoPacketSyncedDrop(&usart1.rxFifo);
					continue;
				}

				if(result)
					continue;

				ConsolePrint("\r\n");

				cli.procBuf[length] = '\0';

				// target selection logic
				if(strcmp((char*)cli.procBuf, "target main") == 0)
				{
					cli.currentTarget = INTERCOM_ADDRESS_MAIN;
					echoTargetColor(cli.currentTarget);
					ConsolePrint("New target: MAIN\r\n");
					continue;
				}

				if(strcmp((char*)cli.procBuf, "target media") == 0)
				{
					cli.currentTarget = INTERCOM_ADDRESS_MEDIA;
					echoTargetColor(cli.currentTarget);
					ConsolePrint("New target: MEDIA\r\n");
					continue;
				}

				// check current target
				if(cli.currentTarget == INTERCOM_ADDRESS_MAIN)
				{
					// process onboard
					result = FifoPacketSyncedBegin(&console.rxFifo);
					if(result)
						continue;

					result = FifoPacketSyncedPush(&console.rxFifo, length, cli.procBuf);
					if(result)
					{
						FifoPacketSyncedCancel(&console.rxFifo);
						continue;
					}

					FifoPacketSyncedComplete(&console.rxFifo);
				}
				else
				{
					// forward to intercom
					IntercomHeader header;
					header.address = cli.currentTarget;
					header.command = 0;
					header.section = INTERCOM_SECTION_CONSOLE;

					IntercomSend(&intercom1, &header, length, cli.procBuf);
				}
			}

			// check data from main console to terminal
			while(FifoPacketSyncedNumPackets(&console.txFifo) > 0)
			{
				echoTargetColor(INTERCOM_ADDRESS_MAIN);

				result = FifoPacketSyncedPop(&console.txFifo, &length, cli.procBuf, CLI_PROC_SIZE);
				if(result == ERROR_FIFO_PACKET_SYNCED_NOT_ENOUGH_MEMORY)
				{
					FifoPacketSyncedDrop(&console.txFifo);
					continue;
				}

				if(result)
					continue;

				FifoSyncedPush(&usart1.txFifo, length, cli.procBuf);

				echoTargetColor(cli.currentTarget);

				USART1StartTx();
			}
		}

		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		// Heartbeat LED
		if(xLastWakeTime - xLastLEDTime > 500/portTICK_RATE_MS)
		{
			xLastLEDTime = xLastWakeTime;

			LEDToggle(LED_MAIN_GREEN);
		}

		vTaskDelayUntil(&xLastWakeTime, 50/portTICK_RATE_MS);
	}
}

static void sensorsCommands()
{
	if(ConsoleCmpCmd("sensors"))
	{
		ConsolePrint("Acc Output: X %+07.3fm/s, Y %+07.3fm/s, Z %+07.3fm/s\r\n", lis3.acc[0], lis3.acc[1], lis3.acc[2]);
		ConsolePrint("Gyr Output: X %+07.3fdps,  Y %+07.3fdps,  Z %+07.3fdps\r\n", l3g.vel[0], l3g.vel[1], l3g.vel[2]);
	}
}

static void miscCommands()
{
	if(ConsoleCmpCmd("bootloader"))
	{
		ConsolePrint("Entering bootloader...\r\n");

		BootStartBootloader(&intercom1);
	}

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 42);
	}

	float f = 21.0f;
	if(ConsoleScanCmd("float %f", &f) == 1)
	{
		ConsolePrint("Float: %f\r\n", f);
	}

	if(ConsoleCmpCmd("float"))
	{
		ConsolePrint("Float: %f\r\n", 42.0f);
	}

	if(ConsoleCmpCmd("usart2 stats"))
	{
		ConsolePrint("--- USART2 ---\r\n");
		ConsolePrint("Rx invalid: %d\r\n", usart2.stats.rxInvalid);
	}

	if(ConsoleCmpCmd("led get"))
	{
		ConsolePrint("LED State: 0x%X\r\n", LEDGetState());
	}

	if(ConsoleCmpCmd("enc get"))
	{
		for(uint8_t i = 0; i < 4; i++)
		{
			ConsolePrint("M%d: %f rad/s\r\n", i, encoder.angularVelocity[i]);
		}
	}

	if(ConsoleCmpCmd("task list"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tState\tPrio\tStack\tID");

		ConsoleLockPrintBuf();
		vTaskList((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	if(ConsoleCmpCmd("task stats"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tTicks\t\tRatio");

		ConsoleLockPrintBuf();
		vTaskGetRunTimeStats((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	uint8_t reg;
	if(ConsoleScanCmd("nrf read %hhu", &reg) == 1)
	{
		uint8_t value;
		NRF24ReadRegister(&network.nrf24, reg, &value);
		ConsolePrint("Reg 0x%X: 0x%X\r\n", reg, value);
	}

	if(ConsoleScanCmd("dribbler %f", &f) == 1)
	{
		ConsolePrint("Dribbler: %f RPM\r\n", f);
		DribblerSetTarget(f);
	}

	if(ConsoleCmpCmd("wifi stats"))
	{
		ConsolePrint("-- TX --\r\n");
		ConsolePrint("Bytes:   %10u/%10u\r\n", network.nrf24.stats.txBytesSent, network.nrf24.stats.txBytesLost);
		ConsolePrint("Packets: %10u/%10u\r\n", network.nrf24.stats.txPacketsSent, network.nrf24.stats.txPacketsLost);
		ConsolePrint("Retrans: %10u\r\n",network.nrf24.stats.retransmissions);
		ConsolePrint("-- RX --\r\n");
		ConsolePrint("Bytes:   %10u/%10u\r\n", network.nrf24.stats.rxBytesRecv, network.nrf24.stats.rxBytesLost);
		ConsolePrint("Packets: %10u/%10u\r\n", network.nrf24.stats.rxPacketsRecv, network.nrf24.stats.rxPacketsLost);
	}

	if(ConsoleCmpCmd("float16"))
	{
		float x = 1.0f;
		float16_t f16 = floatToHalfFloat(x);

		float y = halfFloatToFloat(f16);

		ConsolePrint("In: %f, F16: 0x%04X, out: %f\r\n", x, f16, y);
	}

	if(ConsoleCmpCmd("sensors open"))
	{
		ConsolePrint("Opening sensors log file\r\n");

		IntercomHeader iHeader;
		iHeader.section = INTERCOM_SECTION_SENSORS;
		iHeader.address = INTERCOM_ADDRESS_MEDIA;
		iHeader.command = 1;

		IntercomSend(&intercom1, &iHeader, 0, 0);
	}

	if(ConsoleCmpCmd("sensors close"))
	{
		ConsolePrint("Closing sensors log file\r\n");

		IntercomHeader iHeader;
		iHeader.section = INTERCOM_SECTION_SENSORS;
		iHeader.address = INTERCOM_ADDRESS_MEDIA;
		iHeader.command = 2;

		IntercomSend(&intercom1, &iHeader, 0, 0);
	}

	if(ConsoleCmpCmd("net hello"))
	{
		ConsolePrint("Saying hello to network print\r\n");
		NetworkPrint("Hello World");
	}

	if(ConsoleCmpCmd("motor runtime"))
	{
		ConsolePrint("Runtime: %f\r\n", motorIr.runtime);
	}

	uint8_t u8;
	if(ConsoleScanCmd("set botid %hhu", &u8) == 1)
	{
		NRF24ChangeId(&network.nrf24, u8);
		config.botId = u8;

		ConsolePrint("Changed BotID to %hhu\r\n", u8);
	}

	if(ConsoleScanCmd("set channel %hhu", &u8) == 1)
	{
		NRF24ChangeChannel(&network.nrf24, u8);
		config.channel = u8;

		ConsolePrint("Changed channel to %hhu\r\n", u8);
	}

	float f32;
	if(ConsoleScanCmd("set gyro scale %f", &f32) == 1)
	{
		config.gyroScale = f32;

		ConsolePrint("Gyro scale set to: %f\r\n", f32);
	}

	uint8_t speed;
	if(ConsoleScanCmd("set speed %hhu", &speed) == 1)
	{
		switch(speed)
		{
			case 0: speed = NRF24_SPEED_250K; break;
			case 1: speed = NRF24_SPEED_1M; break;
			default: speed = NRF24_SPEED_2M; break;
		}

		NRF24ChangeSpeed(&network.nrf24, speed);
		config.speed = speed;

		if(config.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(config.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(config.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");
	}

	if(ConsoleCmpCmd("config write"))
	{
		int16_t result = ConfigSave();
		ConsolePrint("Config saved to FLASH (0x%hX)\r\n", result);
	}

	if(ConsoleCmpCmd("config"))
	{
		ConsolePrint("ID: %hhu\r\n", config.botId);
		ConsolePrint("Channel: %hhu\r\n", config.channel);
		if(config.speed == NRF24_SPEED_250K)
			ConsolePrint("Speed: 250k\r\n");
		if(config.speed == NRF24_SPEED_1M)
			ConsolePrint("Speed: 1M\r\n");
		if(config.speed == NRF24_SPEED_2M)
			ConsolePrint("Speed: 2M\r\n");
		ConsolePrint("Gyro scale: %f\r\n", config.gyroScale);
	}

	if(ConsoleCmpCmd("crc"))
	{
		ConsolePrint("Testing CRC calculation\r\n");

		uint32_t data[16];
		memset(data, 0, sizeof(uint32_t)*16);

		uint32_t crc = CRC32Checksum(data, 16);

		ConsolePrint("CRC32: %08X\r\n", crc);
	}
}

static void atMegaCommands()
{
	if(ConsoleCmpCmd("move sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&motorIr.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&motorIr.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&motorIr.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&motorIr.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", motorIr.atmega.signature[0], motorIr.atmega.signature[1], motorIr.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", motorIr.atmega.fuses[0], motorIr.atmega.fuses[1], motorIr.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("move prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&motorIr.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&motorIr.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&motorIr.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&motorIr.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&motorIr.atmega, motor_prog_data, MOTOR_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&motorIr.atmega, motor_prog_data, MOTOR_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&motorIr.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", motorIr.atmega.signature[0], motorIr.atmega.signature[1], motorIr.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", motorIr.atmega.fuses[0], motorIr.atmega.fuses[1], motorIr.atmega.fuses[2]);
	}


	if(ConsoleCmpCmd("move verify"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&motorIr.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&motorIr.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&motorIr.atmega);
				if(result == 0)
				{
					result = ATMegaVerifyProgram(&motorIr.atmega, motor_prog_data, MOTOR_PROG_SIZE);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&motorIr.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", motorIr.atmega.signature[0], motorIr.atmega.signature[1], motorIr.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", motorIr.atmega.fuses[0], motorIr.atmega.fuses[1], motorIr.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("move fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&motorIr.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&motorIr.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&motorIr.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&motorIr.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&motorIr.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", motorIr.atmega.signature[0], motorIr.atmega.signature[1], motorIr.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", motorIr.atmega.fuses[0], motorIr.atmega.fuses[1], motorIr.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("power sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&power.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&power.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&power.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&power.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("power prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&power.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&power.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&power.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&power.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&power.atmega, power_prog_data, POWER_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&power.atmega, power_prog_data, POWER_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&power.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("power fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&power.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&power.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&power.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&power.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&power.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("kicker sig"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&kicker.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&kicker.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&kicker.atmega);
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&kicker.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("kicker prog"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&kicker.atmega);
		if(result == 0)
		{
			result = ATMegaChipErase(&kicker.atmega);
			if(result == 0)
			{
				result = ATMegaReadSignature(&kicker.atmega);
				if(result == 0)
				{
					result = ATMegaReadFuses(&kicker.atmega);
					if(result == 0)
					{
						result = ATMegaLoadProgram(&kicker.atmega, kicker_prog_data, KICKER_PROG_SIZE);
						if(result == 0)
						{
							result = ATMegaVerifyProgram(&kicker.atmega, kicker_prog_data, KICKER_PROG_SIZE);
						}
					}
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&kicker.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	}

	if(ConsoleCmpCmd("kicker fuse"))
	{
		ConsolePrint("Trying to enter prog mode\r\n");

		vTaskDelay(20/portTICK_RATE_MS);

		int16_t result = ATMegaEnterProgramming(&kicker.atmega);
		if(result == 0)
		{
			result = ATMegaReadSignature(&kicker.atmega);
			if(result == 0)
			{
				result = ATMegaReadFuses(&kicker.atmega);
				if(result == 0)
				{
					result = ATMegaWriteFuses(&kicker.atmega, 0xC0, 0xDF, 0xF9);
				}
			}
		}

		vTaskDelay(20/portTICK_RATE_MS);

		ATMegaExitProgramming(&kicker.atmega);

		ConsolePrint("Result: 0x%X\r\n", result);
		ConsolePrint("Signature: %X %X %X\r\n", kicker.atmega.signature[0], kicker.atmega.signature[1], kicker.atmega.signature[2]);
		ConsolePrint("Fuses: %X %X %X\r\n", kicker.atmega.fuses[0], kicker.atmega.fuses[1], kicker.atmega.fuses[2]);
	}
}

static void checkCommands()
{
	if(ConsoleCmpCmd("check hall"))
	{
		ConsolePrint("Starting motor hall sensor test...");

		SelfCheckMotorHallResult result;
		SelfCheckMotorHall(&result);

		ConsolePrint("Done");

		for(uint8_t m = 0; m < 5; m++)
		{
			ConsolePrint("\r\nM%d: \r\n", m);
			ConsolePrint("Current: ");

			for(uint8_t i = 0; i < 6; i++)
				ConsolePrint("%dmA, ", result.M[m].current[i]);

			ConsolePrint("\r\nHall:    ");

			for(uint8_t i = 0; i < 6; i++)
				ConsolePrint("%d, ", result.M[m].hallEvents[i]);
		}

		ConsolePrint("\r\n");
	}

	if(ConsoleCmpCmd("check gear"))
	{
		ConsolePrint("Self checking gear\r\n");

		SelfCheckGearResult result;
		SelfCheckGear(&result);

		for(uint8_t m = 0; m < 5; m++)
		{
			ConsolePrint("M%d Fwd: M: %.2f, E: %.2f, R: %.3f\r\n", m, result.M[m].fwdMotorAngle, result.M[m].fwdEncAngle, result.M[m].fwdRatio);
			ConsolePrint( "   Rev: M: %.2f, E: %.2f, R: %.3f\r\n", m, result.M[m].revMotorAngle, result.M[m].revEncAngle, result.M[m].revRatio);
		}
	}
}

static void motorCommands()
{
	int16_t i16[5];

	if(ConsoleCmpCmd("motor get"))
	{
		ConsolePrint("--- Motor infos ---\r\n");
		ConsolePrint("M 1: %hd mA\r\n", motorIr.atData.current[0]);
		ConsolePrint("M 2: %hd mA\r\n", motorIr.atData.current[1]);
		ConsolePrint("M 3: %hd mA\r\n", motorIr.atData.current[2]);
		ConsolePrint("M 4: %hd mA\r\n", motorIr.atData.current[3]);
		ConsolePrint("M 5: %hd mA\r\n", motorIr.atData.current[4]);
		ConsolePrint("F  : %hhX\r\n", motorIr.atData.fault);
	}

	if(ConsoleScanCmd("motor power %hd %hd", i16, i16+1) == 2)
	{
		ConsolePrint("Motor %hd power: %hd\r\n", i16[0], i16[1]);

		MotorSetPower(i16[0], i16[1]);
	}

	if(ConsoleCmpCmd("motor hall"))
	{
		i16[0] = M1Pos;
		i16[1] = M2Pos;
		i16[2] = M3Pos;
		i16[3] = M4Pos;
		i16[4] = M5Pos;

		for(uint8_t i = 0; i < 5; i++)
			ConsolePrint("Motor %d hall pos: %d\r\n", i, i16[i]);
	}

	if(ConsoleScanCmd("motor phase %hd %hd %hd", i16, i16+1, i16+2) == 3)
	{
		ConsolePrint("Setting motor %hd to phase %hd, power: %hd\r\n", i16[0], i16[1], i16[2]);

		MotorSetPhase(i16[0], i16[1], i16[2]);
	}

	if(ConsoleCmpCmd("motor state"))
	{
		for(uint8_t i = 0; i < 5; i++)
		{
			ConsolePrint("M%d: %d\r\n", i, *motor.M[i].ccr);
		}
	}

	if(ConsoleCmpCmd("motor coms zero"))
	{
		ConsolePrint("Setting motor coms to zero\r\n", *i16);
		MotorZeroAllComs();
	}

	if(ConsoleScanCmd("motor coms %hd", i16) == 1)
	{
		ConsolePrint("Motor %hd coms: %u\r\n", *i16, motor.M[*i16].coms);
	}

	if(ConsoleCmpCmd("motor clear faults"))
	{
		ConsolePrint("Clearing motor faults\r\n");

		for(uint8_t i = 0; i < 5; i++)
			MotorClearDriverFault(i);
	}

	if(ConsoleCmpCmd("motor ex"))
	{
		int16_t result = ATMegaExchange(&motorIr.atmega, motorIr.atData.exchange.mosi, motorIr.atData.exchange.miso, MOTOR_DATA_SIZE);
		ConsolePrint("Motor ATMegaExchange code: 0x%hX\r\n", result);
	}
}

static void kickerCommands()
{
	int16_t i16[3];
	float f32[1];

	if(ConsoleCmpCmd("kicker status"))
	{
		ConsolePrint("--- Kicker Status ---\r\n");
		ConsolePrint("Cap: %hd\r\n", kicker.latestStatus.cap);
		ConsolePrint("Chg: %hd\r\n", kicker.latestStatus.chg);
		ConsolePrint("IR:  %hd\r\n", kicker.latestStatus.ir);
	}

	if(ConsoleScanCmd("kicker fire1 %f", f32) == 1)
	{
		ConsolePrint("Kicker fire1, duration %fus\r\n", *f32);

		KickerFire(*f32, KICKER_DEVICE_STRAIGHT);
	}

	if(ConsoleScanCmd("kicker fire2 %f", f32) == 1)
	{
		ConsolePrint("Kicker fire2, duration %fus\r\n", *f32);

		KickerFire(*f32, KICKER_DEVICE_CHIP);
	}

	if(ConsoleScanCmd("kicker charge %hd %hd %hd", i16, i16+1, i16+2) == 3)
	{
		uint16_t on = i16[0];
		uint16_t off = i16[1];
		uint16_t time = i16[2];
		ConsolePrint("Charging, on: %hd, off: %hd, time: %hdms\r\n", on, off, time);

		KickerLoadManual(on, off, time);
	}

	if(ConsoleScanCmd("kicker auto %hd", i16) == 1)
	{
		uint16_t maxCap = *i16;
		ConsolePrint("Enabling Auto-Charge, limit: %hd\r\n", maxCap);

		KickerLoadAuto(maxCap);
	}

	if(ConsoleCmpCmd("kicker discharge"))
	{
		ConsolePrint("Doing soft discharge\r\n");

		KickerDischarge();
	}
}

static void powerCommands()
{
	uint8_t u8;

	if(ConsoleCmpCmd("power get"))
	{
		ConsolePrint("--- Supply infos ---\r\n");
		ConsolePrint("I Vcc:   %hd mA\r\n", power.atData.iVCC);
		ConsolePrint("I 5V:    %hd mA\r\n", power.atData.i5V);
		ConsolePrint("I 3.3V:  %hd mA\r\n", power.atData.i3V3);
		ConsolePrint("U Cell1: %hd mV\r\n", power.atData.uCell[0]);
		ConsolePrint("U Cell2: %hd mV\r\n", power.atData.uCell[1]);
		ConsolePrint("U Cell3: %hd mV\r\n", power.atData.uCell[2]);
		ConsolePrint("U Cell4: %hd mV\r\n", power.atData.uCell[3]);
	}

	if(ConsoleScanCmd("power 5 %hhd", &u8))
	{
		ConsolePrint("Setting 5V power to %hhd\r\n", u8);

		PowerEnable5VSupply(u8);
	}
}

static void irCommands()
{
	if(ConsoleCmpCmd("ir get"))
	{
		ConsolePrint("IR Level:     %hd\r\n", motorIr.atData.irLevel);
		ConsolePrint("IR Threshold: %hd\r\n", motorIr.atData.irThreshold);
	}
}

static void printMat(arm_matrix_instance_f32* pMat)
{
	for(uint16_t r = 0; r < pMat->numRows; r++)
	{
		ConsolePrint("[ ");
		for(uint16_t c = 0; c < pMat->numCols; c++)
		{
			ConsolePrint("%4.04f ", pMat->pData[r*pMat->numCols+c]);
		}
		ConsolePrint("]\r\n");
	}
}

static void controllerCommands()
{
	float f32[5];
	uint8_t u8;

	if(ConsoleCmpCmd("sigma"))
	{
		printMat(&ctrlFusion.kfX.Sigma);
		printMat(&ctrlFusion.kfY.Sigma);
		printMat(&ctrlFusion.kfW.Sigma);
	}

	if(ConsoleCmpCmd("state"))
	{
		printMat(&ctrlFusion.kfX.x);
		printMat(&ctrlFusion.kfY.x);
		printMat(&ctrlFusion.kfW.x);
	}

/*	uint8_t motor;

	if(ConsoleCmpCmd("ctrl off"))
	{
		controller.mode = CONTROLLER_MODE_OFF;
	}

	if(ConsoleCmpCmd("ctrl pid"))
	{
		controller.mode = CONTROLLER_MODE_PID;
	}

	if(ConsoleCmpCmd("ctrl torque"))
	{
		controller.mode = CONTROLLER_MODE_TORQUE;
	}

	if(ConsoleScanCmd("torque %hhd %f", &motor, f32) == 2)
	{
		ConsolePrint("Motor %hhd: %fmA\r\n", motor, *f32);

		controller.torque.setpoint[motor] = *f32;
	}
*/
	if(ConsoleScanCmd("pid set %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("X: %f, Y: %f, W: %f\r\n", f32[0], f32[1], f32[2]);

		CtrlMotorSetTarget(f32[0], f32[1], f32[2]);
		CtrlFusionVelSetVelTarget(f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("fusion_vel pt1 %f", f32) == 1)
	{
		ConsolePrint("New PT1 time constant: %f\r\n", f32[0]);

		CtrlFusionVelSetPT1TimeConstant(f32[0]);
	}

	if(ConsoleScanCmd("fusion pid %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("Kp: %f, Ki: %f, Kd:%f\r\n", f32[0], f32[1], f32[2]);

		CtrlFusionSetParams(f32[0], f32[1], f32[2]);
		CtrlFusionVelSetParams(f32[0], f32[1], f32[2]);
	}

	if(ConsoleScanCmd("structure %f %f %f %f %f", f32, f32+1, f32+2, f32+3, f32+4) == 5)
	{
		ConsolePrint("New structural parameters:\r\n");
		ConsolePrint("Front angle: %fdeg\r\n", f32[0]);
		ConsolePrint("Back angle: %fdeg\r\n", f32[1]);
		ConsolePrint("Bot radius: %fm\r\n", f32[2]);
		ConsolePrint("Wheel radius: %fm\r\n", f32[3]);
		ConsolePrint("Mass: %fkg\r\n", f32[4]);

		CtrlSetParams(f32[0], f32[1], f32[2], f32[3], f32[4]);
	}

	if(ConsoleScanCmd("fusion ex %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("New state uncertainty: %f %f %f\r\n", f32[0], f32[1], f32[2]);

		CtrlFusionSetStateUncertainty(f32);
	}

	if(ConsoleScanCmd("fusion ez %f %f %f", f32, f32+1, f32+2) == 3)
	{
		ConsolePrint("New sensor uncertainty: %f %f %f\r\n", f32[0], f32[1], f32[2]);

		CtrlFusionSetSensorUncertainty(f32);
	}

	if(ConsoleCmpCmd("spline"))
	{
		ConsolePrint("Setting test spline\r\n");

		CtrlSpline2D spline;
		spline.options = CTRL_SPLINE_OPTION_CLEAR | CTRL_SPLINE_OPTION_APPEND;
		spline.tEnd = floatToHalfFloat(5.0f);

		HermiteSpline x, y;
		HermiteSplineCreate(&x, 0, 2, 0, 0, 5);
		HermiteSplineCreate(&y, 0, 2, 0, 0, 5);

		spline.splines[0].a = floatToHalfFloat(x.a);
		spline.splines[0].b = floatToHalfFloat(x.b);
		spline.splines[0].c = floatToHalfFloat(x.c);
		spline.splines[0].d = floatToHalfFloat(x.d);
		spline.splines[1].a = floatToHalfFloat(y.a);
		spline.splines[1].b = floatToHalfFloat(y.b);
		spline.splines[1].c = floatToHalfFloat(y.c);
		spline.splines[1].d = floatToHalfFloat(y.d);

		CtrlFusionAddSplineCmdXY(&spline);
	}

	if(ConsoleScanCmd("test step %f", f32) == 1)
	{
		ConsolePrint("Motor step. Output: %f\r\n", f32[0]);

		CtrlTestStepSet(f32[0]);
		CtrlFusionVelTestStep(f32[0]);
	}

	if(ConsoleCmpCmd("net gather"))
	{
		CtrlFusionVelGatherData();
	}

	if(ConsoleScanCmd("net log %hhu", &u8) == 1)
	{
		NetworkPrint("Logging %hhu", u8);

		CtrlFusionVelEnableLog(u8);
	}

	if(ConsoleCmpCmd("net input"))
	{
		CtrlFusionVelNetPrintInput();
	}
}

static void buzzerCommands()
{
	uint32_t u32;

	if(ConsoleCmpCmd("buzz on"))
	{
		ConsolePrint("Enabling buzzer\r\n");

		BuzzerEnable(1);
	}

	if(ConsoleCmpCmd("buzz off"))
	{
		ConsolePrint("Disabling buzzer\r\n");

		BuzzerEnable(0);
	}

	if(ConsoleScanCmd("buzz freq %d", &u32) == 1)
	{
		ConsolePrint("Setting buzzer frequency to: %dHz\r\n", u32);

		BuzzerSetFrequency(u32);
	}

	if(ConsoleScanCmd("buzz lvl %d", &u32) == 1)
	{
		ConsolePrint("Setting buzzer level to: %d\r\n", u32);

		BuzzerSetLevel(u32);
	}
}

static void processCommand()
{
	sensorsCommands();
	miscCommands();
	atMegaCommands();
	checkCommands();
	motorCommands();
	kickerCommands();
	powerCommands();
	controllerCommands();
	buzzerCommands();
	irCommands();

	if(ConsoleCmpCmd("version"))
	{
		ConsolePrint("Firmware Version: 0.0.0 - dev\r\n");
	}
}
