/*
 * ctrl.c
 *
 *  Created on: 06.03.2013
 *      Author: AndreR
 */

#include "ctrl.h"

#include "stm32f4xx.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_tim.h"
#include "misc.h"

#include "constants.h"
#include "hal/enc.h"
#include "hal/motor.h"
#include "hal/l3g.h"
#include "hal/lis3.h"
#include "motor_ir.h"
#include "kicker.h"
#include "commands.h"
#include "network.h"
#include "power.h"
#include "config.h"
#include "task.h"
#include "dribbler.h"

#include <string.h>

Ctrl ctrl;

static float pXYW2Motor[12];
static float pMotor2XYW[12];
static float pForceXYW2Motor[12];
static float pForceMotor2XYW[12];

void TIM1_BRK_TIM9_IRQHandler()
{
	TIM9->SR = 0;

	uint8_t event = CTRL_EVENT_RUN;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void encUpdated()
{
	uint8_t event = CTRL_EVENT_UPDATE_ENC;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void gyrUpdated()
{
	uint8_t event = CTRL_EVENT_UPDATE_GYR;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void accUpdated()
{
	uint8_t event = CTRL_EVENT_UPDATE_ACC;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void powerUpdated()
{
	uint8_t event = CTRL_EVENT_UPDATE_POWER;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void posUpdated()
{
	uint8_t event = CTRL_EVENT_UPDATE_POS;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

static void motorUpdated()
{
	uint8_t event = CTRL_EVENT_UPDATE_MOTOR;
	int32_t yield = 0;
	xQueueSendFromISR(ctrl.hEventQueue, &event, &yield);
	if(yield == pdTRUE)
		taskYIELD();
}

void CtrlInit()
{
	memset(&ctrl, 0, sizeof(Ctrl));

	ctrl.dT = 0.01f;

	// construct matrices
	ctrl.matXYW2Motor.numCols = 3;
	ctrl.matXYW2Motor.numRows = 4;
	ctrl.matXYW2Motor.pData = pXYW2Motor;
	ctrl.matMotor2XYW.numCols = 4;
	ctrl.matMotor2XYW.numRows = 3;
	ctrl.matMotor2XYW.pData = pMotor2XYW;
	ctrl.matForceXYW2Motor.numCols = 3;
	ctrl.matForceXYW2Motor.numRows = 4;
	ctrl.matForceXYW2Motor.pData = pForceXYW2Motor;
	ctrl.matForceMotor2XYW.numCols = 4;
	ctrl.matForceMotor2XYW.numRows = 3;
	ctrl.matForceMotor2XYW.pData = pForceMotor2XYW;

	ctrl.state.matPos.numCols = 1;
	ctrl.state.matPos.numRows = 3;
	ctrl.state.matPos.pData = ctrl.state.pos;
	ctrl.state.matVel.numCols = 1;
	ctrl.state.matVel.numRows = 3;
	ctrl.state.matVel.pData = ctrl.state.vel;
	ctrl.state.matAcc.numCols = 1;
	ctrl.state.matAcc.numRows = 3;
	ctrl.state.matAcc.pData = ctrl.state.acc;

	// build structural data
	CtrlSetParams(30.0f, 45.0f, 0.072f, 0.02f, 4.0f); // CtrlSetParams(30.0f, 45.0f, 0.0785f, 0.025f);

	// init event queue
	ctrl.hEventQueue = xQueueCreate(10, sizeof(uint8_t));

	// register encoder update callback
	EncSetUpdateCallback(&encUpdated);

	// register acc and gyr callbacks
	L3GSetCallback(&gyrUpdated);
	LIS3SetCallback(&accUpdated);

	// register power callback
	PowerSetCallback(&powerUpdated);

	// register pos callback
	NetworkSetPosCallback(&posUpdated);

	// register motor current callback
	MotorIrSetCallback(&motorUpdated);

	// init precision timer for accurate controller dT
	RCC_APB2PeriphClockCmd(RCC_APB2Periph_TIM9, ENABLE);

	TIM_TimeBaseInitTypeDef timeBaseInit;
	timeBaseInit.TIM_ClockDivision = TIM_CKD_DIV1;
	timeBaseInit.TIM_CounterMode = TIM_CounterMode_Up;
	timeBaseInit.TIM_Period = 20;	// 100Hz
	timeBaseInit.TIM_Prescaler = 41999;	// = 0.5ms
	timeBaseInit.TIM_RepetitionCounter = 0;
	TIM_TimeBaseInit(TIM9, &timeBaseInit);

	TIM_ITConfig(TIM9, TIM_IT_Update, ENABLE);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = TIM1_BRK_TIM9_IRQn;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_CTRL_TIMER;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	NVIC_Init(&nvicInit);

	TIM_Cmd(TIM9, ENABLE);
}

void CtrlTask(void* params)
{
	(void) params;

	uint8_t event;

	while(1)
	{
		xQueueReceive(ctrl.hEventQueue, &event, portMAX_DELAY);

		switch(event)
		{
			case CTRL_EVENT_RUN:
			{
				if(ctrl.pInstance == 0)
					continue;
				if(ctrl.pInstance->pControllerFunction == 0)
					continue;

				(*ctrl.pInstance->pControllerFunction)(&ctrl.input, &ctrl.output);

				for(uint8_t i = 0; i < 4; i++)
				{
					MotorSetPower(i, (int16_t)(ctrl.output.motorPWM[i]*50.0f));
				}

				// clear driver faults
				for(uint8_t i = 0; i < 5; i++)
				{
					if(motorIr.atData.fault & (1 << i))
					{
						MotorClearDriverFault(i);
					}
				}

				motorIr.atData.fault = 0;

				// send status
				if(rtosHfTickCount - ctrl.lastHfStatusTime > 50000/RTOS_HF_TICK_RATE)	// 50ms
				{
					ctrl.lastHfStatusTime = rtosHfTickCount;

					SystemStatusV2 status;
					memset(&status, 0, sizeof(SystemStatusV2));

					if(motorIr.atData.irLevel < motorIr.atData.irThreshold)
						status.barrierInterrupted = 1;
					else
						 status.barrierInterrupted = 0;

					if(dribbler.speedReached)
						status.bits |= 0x08;

					if(dribbler.overload)
						status.bits |= 0x10;

					status.kickerLevel = kicker.data.uCap;

					if(ctrl.state.updated[0])
					{
						ctrl.state.updated[0] = 0;
						status.bits |= 0x01;
						status.position[0] = (int16_t)(ctrl.state.pos[0]*1000.0f + ctrl.state.vel[0]*50.0f);
						status.position[1] = (int16_t)(ctrl.state.pos[1]*1000.0f + ctrl.state.vel[1]*50.0f);
						status.position[2] = (int16_t)(ctrl.state.pos[2]*1000.0f + ctrl.state.vel[2]*50.0f);
					}

					if(ctrl.state.updated[1])
					{
						ctrl.state.updated[1] = 0;
						status.bits |= 0x02;
						status.velocity[0] = (int16_t)(ctrl.state.vel[0]*1000.0f);
						status.velocity[1] = (int16_t)(ctrl.state.vel[1]*1000.0f);
						status.velocity[2] = (int16_t)(ctrl.state.vel[2]*1000.0f);
					}

					if(ctrl.state.updated[2])
					{
						ctrl.state.updated[2] = 0;
						status.bits |= 0x04;
						status.acceleration[0] = (int16_t)(ctrl.state.acc[0]*1000.0f);
						status.acceleration[1] = (int16_t)(ctrl.state.acc[1]*1000.0f);
						status.acceleration[2] = (int16_t)(ctrl.state.acc[2]*1000.0f);
					}

					PacketHeader header;
					header.section = SECTION_SYSTEM;
					header.cmd = CMD_SYSTEM_STATUS_V2;
					header.len = LEN_SYSTEM_STATUS_V2;

					if(ctrl.sendLog)
						NetworkSendPacket(&header, (uint8_t*)&status);
				}
			}
			break;
			case CTRL_EVENT_UPDATE_MOTOR:
			{
				for(uint8_t i = 0; i < 4; i++)
				{
					ctrl.input.motor.current[i] = motorIr.atData.current[i]*0.001f;

					if(motor.M[i].ccw)
						ctrl.input.motor.current[i] *= -1;
				}

				ctrl.input.motor.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_ENC:
			{
				memcpy(ctrl.input.enc.motorVel, encoder.angularVelocity, 4*sizeof(float));
				ctrl.input.enc.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_GYR:
			{
				memcpy(ctrl.input.gyr.rotVel, l3g.vel, sizeof(float)*3);
				ctrl.input.gyr.rotVel[0] *= config.gyroScale;
				ctrl.input.gyr.rotVel[1] *= config.gyroScale;
				ctrl.input.gyr.rotVel[2] *= config.gyroScale;
				ctrl.input.gyr.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_ACC:
			{
				memcpy(ctrl.input.acc.linAcc, lis3.acc, sizeof(float)*3);
				ctrl.input.acc.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_POS:
			{
				ctrl.input.pos.pos[0] = network.lastVisionPos[0]*0.001f;
				ctrl.input.pos.pos[1] = network.lastVisionPos[1]*0.001f;
				ctrl.input.pos.pos[2] = network.lastVisionPos[2]*0.001f;
				ctrl.input.pos.updated = 1;
			}
			break;
			case CTRL_EVENT_UPDATE_POWER:
			{
				ctrl.input.power.voltage = power.atData.uCell[0]+power.atData.uCell[1]+power.atData.uCell[2]+power.atData.uCell[3];
				ctrl.input.power.voltage *= 0.001f;
				ctrl.input.power.updated = 1;
			}
			break;
			case CTRL_EVENT_SET_CTRL:
			{
				if(ctrl.pInstance)
				{
					if(ctrl.pInstance->pExit)
						(*ctrl.pInstance->pExit)();
				}

				for(uint8_t i = 0; i < 4; i++)
				{
					MotorSetPower(i, 0);
				}

				ctrl.pInstance = ctrl.pNewInstance;

				if(ctrl.pInstance)
				{
					ctrl.dT = ((float)ctrl.pInstance->deltaT)*0.0005f;
					TIM9->ARR = ctrl.pInstance->deltaT;
					TIM9->EGR = TIM_EGR_UG;

					if(ctrl.pInstance->pInit)
						(*ctrl.pInstance->pInit)();
				}
			}
			break;
			default:
				break;
		}
	}
}

void CtrlSetController(CtrlInstance* pInstance)
{
	ctrl.pNewInstance = pInstance;

	uint8_t event = CTRL_EVENT_SET_CTRL;
	xQueueSend(ctrl.hEventQueue, &event, portMAX_DELAY);
}

void CtrlSetParams(float frontAngle, float backAngle, float botRadius, float wheelRadius, float mass)
{
	ctrl.botParams.frontAngle = frontAngle;
	ctrl.botParams.backAngle = backAngle;
	ctrl.botParams.botRadius = botRadius;
	ctrl.botParams.wheelRadius = wheelRadius;
	ctrl.botParams.mass = mass;

	// TODO: inertia is approximated by evenly distributed cylinder
	ctrl.botParams.momentOfInertia = 0.5f*ctrl.botParams.mass*ctrl.botParams.botRadius*ctrl.botParams.botRadius;

	float alpha = frontAngle*PI/180.0f;
	float beta = backAngle*PI/180.0f;

	float theta[4];
	theta[0] = alpha;
	theta[1] = PI-alpha;
	theta[2] = PI+beta;
	theta[3] = 2*PI-beta;

	// construct matrix for XYW to Motor (VM) conversion
	for(uint8_t i = 0; i < 4; i++)
	{
		pXYW2Motor[(i*3)+0] = -sin(theta[i]);
		pXYW2Motor[(i*3)+1] = cos(theta[i]);
		pXYW2Motor[(i*3)+2] = botRadius;

		pForceXYW2Motor[(i*3)+0] = -sin(theta[i]);
		pForceXYW2Motor[(i*3)+1] = cos(theta[i]);
		pForceXYW2Motor[(i*3)+2] = 1.0f/botRadius;
	}

	// calculate pseudo-inverse of XYW2Motor
	float pXYW2MotorTransposed[12];
	arm_matrix_instance_f32 matXYW2MotorTransposed = { .numCols = 4, .numRows = 3, .pData = pXYW2MotorTransposed };
	float pTri[9];
	arm_matrix_instance_f32 matTri = { .numCols = 3, .numRows = 3, .pData = pTri };
	float pTriInv[9];
	arm_matrix_instance_f32 matTriInv = { .numCols = 3, .numRows = 3, .pData = pTriInv };

	arm_mat_trans_f32(&ctrl.matXYW2Motor, &matXYW2MotorTransposed);
	arm_mat_mult_f32(&matXYW2MotorTransposed, &ctrl.matXYW2Motor, &matTri);
	arm_mat_inverse_f32(&matTri, &matTriInv);
	arm_mat_mult_f32(&matTriInv, &matXYW2MotorTransposed, &ctrl.matMotor2XYW);

	// calculate pseudo-inverse of ForceXYW2Motor
	arm_mat_trans_f32(&ctrl.matForceXYW2Motor, &matXYW2MotorTransposed);
	arm_mat_mult_f32(&matXYW2MotorTransposed, &ctrl.matForceXYW2Motor, &matTri);
	arm_mat_inverse_f32(&matTri, &matTriInv);
	arm_mat_mult_f32(&matTriInv, &matXYW2MotorTransposed, &ctrl.matForceMotor2XYW);
}

void CtrlSetLog(uint8_t enable)
{
	ctrl.sendLog = enable;
}
