/*
 * init.c
 *
 *  Created on: 17.05.2011
 *      Author: AndreR
 */

#include "init.h"

#include "stm32f4xx_iwdg.h"

#include "FreeRTOS.h"
#include "task.h"

#include "atmega/kicker_prog.h"
#include "atmega/motor_prog.h"
#include "atmega/power_prog.h"

#include "cli.h"
#include "network.h"
#include "spi1_2_3.h"
#include "power.h"
#include "kicker.h"
#include "util/console.h"
#include "intercom1.h"
#include "intercom_constants.h"
#include "motor_ir.h"
#include "ctrl.h"
#include "dribbler.h"
#include "sensors.h"
#include "ctrl_motor.h"
#include "ctrl_fusion.h"
#include "ctrl_test_step.h"
#include "ctrl_fusion_vel.h"
#include "config.h"
#include "util/crc.h"
#include "presenter.h"

#include "hal/motor.h"
#include "hal/enc.h"
#include "hal/buzzer.h"
#include "hal/exti.h"
#include "hal/usart1.h"
#include "hal/usart2.h"
#include "hal/hw_sync1.h"


void InitTask(void* params)
{
	params = params;

	PowerPreInit();
	PowerEnable5VSupply(1);

	// Configure and start Independent watchdog
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Enable);
	IWDG_SetPrescaler(IWDG_Prescaler_8);
	IWDG_SetReload(2500);	// approx. 500ms
	IWDG_WriteAccessCmd(IWDG_WriteAccess_Disable);

	// Initialize shared resources
	// --- HAL ---
	ConfigInit();
	EncInit();
	MotorInit();
	BuzzerInit();
	ExtiInit();
	USART1Init();
	USART2Init();
	CRCInit();

	// --- GATEKEEPERS ---
	SPI1_2_3Init();

	// --- High-Level Systems ---
	Intercom1Init();
	CLIInit();
	PowerInit();
	SensorsInit();
	DribblerInit();
	NetworkInit();
	KickerInit();
	MotorIrInit();
	CtrlInit();
	PresenterInit();

//	CtrlSetController(0);
//	CtrlSetController(&ctrlMotorInstance);
	CtrlSetController(&ctrlFusionInstance);
//	CtrlSetController(&ctrlFusionVelInstance);
//	CtrlSetController(&ctrlTestStepInstance);

	// -- CLI
	ConsolePrint("\f\e[0;36mInitializing...\e[0m\r\n");

	xTaskCreate(CLITask, (const signed char* const)"CLI", 1536, NULL, 1, NULL);
	ConsolePrint("CLI: \e[0;32mOK\e[0m\r\n");

	ConsolePrint("Waiting for co-processors...");
	HWSyncBarrier(&hwSync1);
	ConsolePrint("Ready\r\n");

//	uint32_t uid1 = *((uint32_t*)(0x1FFF7A10 + 0));	// CPU ID register (96 bit)
//	uint32_t uid2 = *((uint32_t*)(0x1FFF7A10 + 4));
//	uint32_t uid3 = *((uint32_t*)(0x1FFF7A10 + 8));

//	ConsolePrint("CPU ID: 0x%08X 0x%08X 0x%08X\r\n", uid3, uid2, uid1);

	xTaskCreate(USARTTask, (const signed char* const)"USART2", 1024, &usart2, 4, NULL);
	vTaskDelay(100/portTICK_RATE_MS);	// give USART some time to reset RX
	xTaskCreate(Intercom1Task, (const signed char* const)"IC", 1024, NULL, 1, NULL);

	// -- SPI 1, 2, 3
	xTaskCreate(SPI1_2_3Task, (const signed char* const)"SPI", 256, NULL, 4, NULL);

	ConsolePrint("SPI 1 + 2 + 3: \e[0;32mOK\e[0m\r\n");

	// Create Tasks, MAX PRIORITY: 4
	xTaskCreate(PowerTask, (const signed char* const)"Power", 512, NULL, 1, NULL);
	xTaskCreate(SensorsTask, (const signed char* const)"Sensors", 256, NULL, 2, NULL);
	xTaskCreate(MotorIrTask, (const signed char* const)"MotorIr", 256, NULL, 1, NULL);
	xTaskCreate(NRF24RxTask, (const signed char* const)"NRF24RX", 1536, &network.nrf24, 4, NULL);
	xTaskCreate(DribblerTask, (const signed char* const)"Dribbler", 512, NULL, 2, NULL);
	xTaskCreate(CtrlTask, (const signed char* const)"CTRLNew", 1024, NULL, 3, NULL);

	if(kicker.attached)
	{
		ConsolePrint("Kicker attached. Starting task\r\n");
		xTaskCreate(KickerTask, (const signed char* const)"Kicker", 512, NULL, 2, NULL);
	}
	else
	{
		ConsolePrint("Kicker not attached. Not starting kicker task\r\n");
	}

	ConsolePrint("%d Tasks started\r\n", uxTaskGetNumberOfTasks());
	ConsolePrint("HEAP usage: %d/%d Byte\r\n", configTOTAL_HEAP_SIZE-xPortGetFreeHeapSize(), configTOTAL_HEAP_SIZE);
	ConsolePrint("Initialization done.\r\n");

	vTaskSuspend(NULL);	// Disable this task
}
