/*
 * power.c
 *
 * Power control and overcurrent watchdog.
 *
 *  Created on: 25.10.2010
 *      Author: AndreR
 */

#include "power.h"

#include "stm32f4xx.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_iwdg.h"

#include "FreeRTOS.h"
#include "task.h"

#include "commands.h"
#include "network.h"
#include "hal/motor.h"
#include "atmega/power_prog.h"
#include "hal/buzzer.h"
#include "hal/led.h"
#include "kicker.h"
#include "util/console.h"
#include "config.h"
#include "intercom1.h"
#include "intercom_constants.h"

#include "spi1_2_3.h"

#define POWER_5V_ENABLE (1 << 12)	// GPIOB
#define POWER_ATMEGA_CS (1 << 11)	// GPIOA
#define POWER_ATMEGA_RST (1 << 12)	// GPIOA

PowerGlobal power;

void PowerPreInit()
{
	GPIO_InitTypeDef gpioInit;

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_Pin = POWER_5V_ENABLE;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	GPIO_Init(GPIOB, &gpioInit);
}

void PowerInit()
{
	GPIO_InitTypeDef gpioInit;

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_Pin = POWER_ATMEGA_CS | POWER_ATMEGA_RST;
	gpioInit.GPIO_Speed = GPIO_Speed_50MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	GPIO_Init(GPIOA, &gpioInit);

	GPIOA->BSRRL = POWER_ATMEGA_CS;
	GPIOA->BSRRL = POWER_ATMEGA_RST;

	ATMegaInit(&power.atmega, &spi3, GPIOA, POWER_ATMEGA_CS, GPIOA, POWER_ATMEGA_RST);

	power.state = POWER_STATE_GOOD;
	power.atData.uCell[0] = 4000;
	power.atData.uCell[1] = 4000;
	power.atData.uCell[2] = 4000;
	power.atData.uCell[3] = 4000;

	power.blinkRed = 0;
	power.buzzerOn = 0;
	power.cb = 0;
}

void PowerTask(void* params)
{
	params = params;

	portTickType xLastWakeTime = xTaskGetTickCount();
	portTickType xLogTime = xLastWakeTime;
	portTickType xBeepTime = xLastWakeTime;
	portTickType xLEDTime = xLastWakeTime;
	portTickType xLowBatTimeStart = 0;
	portTickType xOffBatTimeStart = 0;

	uint8_t newState = POWER_STATE_GOOD;

	uint16_t beepDelay = 100;
	uint8_t beepIter = 0;
	uint16_t beepCycles[4] = {100, 250, 100, 2000};

	uint16_t ledDelay = 100;
	uint8_t ledIter = 0;
	uint16_t ledCycles[6] = {100, 100, 100, 100, 200, 800};

	int16_t result = 0;

	// -- Power
	uint8_t programUpdated = 0;
	result = ATMegaUpdateProgram(&power.atmega, power_prog_data, POWER_PROG_SIZE, &programUpdated);

	ConsolePrint("--- Power AVR ---\r\n");
	ConsolePrint("Signature: %X %X %X\r\n", power.atmega.signature[0], power.atmega.signature[1], power.atmega.signature[2]);
	ConsolePrint("Fuses: %X %X %X\r\n", power.atmega.fuses[0], power.atmega.fuses[1], power.atmega.fuses[2]);
	if(programUpdated)
		ConsolePrint("\e[0;36mUpdated power program\e[0m\r\n");
	if(result)
		ConsolePrint("\e[0;31mPower program update failed (0x%X)\e[0m\r\n", result);

	NetworkPrint("Version: 1.0.7");

	while(1)
	{
//		IWDG_ReloadCounter();

		if(xOffBatTimeStart != 0 && xLastWakeTime - xOffBatTimeStart > 10000/portTICK_RATE_MS)
			newState = POWER_STATE_OFF;
		else if(xLowBatTimeStart != 0 && xLastWakeTime - xLowBatTimeStart > 10000/portTICK_RATE_MS)
			newState = POWER_STATE_LOW;
		else
			newState = POWER_STATE_GOOD;

		// Power level state machine
		if(power.state == POWER_STATE_GOOD && newState == POWER_STATE_LOW)
		{
			// enable buzzer
			power.buzzerOn = 1;

			power.state = newState;
		}

		if(power.state != POWER_STATE_OFF && newState == POWER_STATE_OFF)
		{
			power.buzzerOn = 0;
			power.blinkRed = 1;

			power.state = newState;

			MotorEmergencyShutdown();

			KickerLoadManual(6, 78, 1);
			vTaskDelay(200/portTICK_RATE_MS);
			KickerFire(5e6f, KICKER_DEVICE_STRAIGHT);

			PowerEnable5VSupply(0);
		}

		if(power.buzzerOn)
		{
			if(xLastWakeTime - xBeepTime > beepDelay/portTICK_RATE_MS)
			{
				xBeepTime = xLastWakeTime;

				beepDelay = beepCycles[beepIter++];
				beepIter %= 4;
				if(beepIter % 2 == 0)
					BuzzerEnable(0);
				else
					BuzzerEnable(1);
			}
		}
		else
		{
			BuzzerEnable(0);
		}

		if(power.blinkRed)
		{
			if(xLastWakeTime - xLEDTime > ledDelay/portTICK_RATE_MS)
			{
				xLEDTime = xLastWakeTime;

				ledDelay = ledCycles[ledIter++];
				ledIter %= 6;
				if(ledIter % 2 == 0)
					LEDMustOff(LED_MAIN_RED);
				else
					LEDMustOn(LED_MAIN_RED);
			}
		}
		else
		{
			LEDMustOff(LED_MAIN_RED);
		}

		if(xLastWakeTime - xLogTime > 500/portTICK_RATE_MS)
		{
			xLogTime = xLastWakeTime;

			// send config to media
			IntercomHeader icHeader;
			icHeader.address = INTERCOM_ADDRESS_MEDIA;
			icHeader.section = INTERCOM_SECTION_PRESENTER;
			icHeader.command = IC_PRESENTER_CONFIG;

			IntercomSend(&intercom1, &icHeader, sizeof(Config), (uint8_t*)&config);

			SystemPowerLog log;
			log.iVCC = power.atData.iVCC;
			log.i5V = power.atData.i5V;
			log.i3V3 = power.atData.i3V3;
			log.u1 = power.atData.uCell[0];
			log.u2 = power.atData.uCell[1];
			log.u3 = power.atData.uCell[2];
			log.u4 = power.atData.uCell[3];

			PacketHeader header;
			header.section = SECTION_SYSTEM;
			header.cmd = CMD_SYSTEM_POWER_LOG;
			header.len = LEN_SYSTEM_POWER_LOG;

			NetworkSendPacket(&header, (uint8_t*)&log);

			icHeader.command = IC_PRESENTER_POWER;

			IntercomSend(&intercom1, &icHeader, sizeof(SystemPowerLog), (uint8_t*)&log);

			ATMegaExchange(&power.atmega, power.atData.exchange.mosi, power.atData.exchange.miso, POWER_DATA_SIZE);

			if(power.atData.uCell[0] > 5000)
				ConsolePrint("Cell 0 abnormal: %d\r\n", power.atData.uCell[0]);
			if(power.atData.uCell[1] > 5000)
				ConsolePrint("Cell 1 abnormal: %d\r\n", power.atData.uCell[1]);
			if(power.atData.uCell[2] > 5000)
				ConsolePrint("Cell 2 abnormal: %d\r\n", power.atData.uCell[2]);
			if(power.atData.uCell[3] > 16000)
				ConsolePrint("Cell 3 abnormal: %d\r\n", power.atData.uCell[3]);
			if(power.atData.iVCC > 10000)
				ConsolePrint("iVCC abnormal: %d\r\n", power.atData.iVCC);
			if(power.atData.i3V3 > 10000)
				ConsolePrint("i3V3 abnormal: %d\r\n", power.atData.i3V3);
			if(power.atData.i5V > 10000)
				ConsolePrint("i5V abnormal: %d\r\n", power.atData.i5V);

			uint8_t i;
			uint8_t low = 0;
			uint8_t off = 0;
			for(i = 0; i < 4; i++)
			{
				if(power.atData.uCell[i] < 3200)
					low = 1;
				if(power.atData.uCell[i] < 3000)
					off = 1;
			}

			if(power.atData.uCell[0] < 100 && power.atData.uCell[1] < 100 && power.atData.uCell[2] < 100 && power.atData.uCell[3] > 10000)
			{
				// Stationary power supply connected
			}
			else
			{
				if(low)
				{
					if(xLowBatTimeStart == 0)
						xLowBatTimeStart = xLastWakeTime;
				}
				else
					xLowBatTimeStart = 0;

				if(off)
				{
					if(xOffBatTimeStart == 0)
						xOffBatTimeStart = xLastWakeTime;
				}
				else
					xOffBatTimeStart = 0;
			}

			if(power.cb)
				(*power.cb)();
		}

		vTaskDelayUntil(&xLastWakeTime, 10/portTICK_RATE_MS);
	}
}

void PowerEnable5VSupply(uint8_t enable)
{
	if(enable)
		GPIOB->BSRRH = POWER_5V_ENABLE;
	else
		GPIOB->BSRRL = POWER_5V_ENABLE;
}

uint8_t PowerIs5VSupplyEnabled()
{
	if(GPIOB->ODR & POWER_5V_ENABLE)
		return 0;
	else
		return 1;
}

void PowerSetCallback(PowerUpdatedCallback cb)
{
	power.cb = cb;
}
