/*
 * audio.c
 *
 *  Created on: 08.12.2012
 *      Author: AndreR
 */

#include "audio.h"

#include "stm32f4xx.h"
#include "stm32f4xx_gpio.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_spi.h"
#include "misc.h"

#include "spi3.h"
#include "constants.h"

#include "FreeRTOS.h"
#include "task.h"

#include "util/console.h"
#include "fatfs/ff.h"
#include "errors.h"

#include <string.h>

typedef struct _RIFFHeader
{
	uint32_t chunkID;
	uint32_t chunkSize;
	uint32_t format;
} RIFFHeader;

typedef struct _FMTChunk
{
	uint32_t chunkID;
	uint32_t chunkSize;
	uint16_t audioFormat;
	uint16_t numChannels;
	uint32_t sampleRate;
	uint32_t byteRate;
	uint16_t blockAlign;
	uint16_t bitsPerSample;
} FMTChunk;

typedef struct _DataChunk
{
	uint32_t chunkId;
	uint32_t chunkSize;
} DataChunk;

Audio audio;

#define AUDIO_ACTION_HIGH_EMPTY 1
#define AUDIO_ACTION_LOW_EMPTY 2

#define AUDIO_BUF_SAMPLES 8820
static uint16_t audioBuf[AUDIO_BUF_SAMPLES]; // sufficient for 100ms of 44.1kHz, 16-Bit stereo data

void DMA1_Stream4_IRQHandler()
{
	uint8_t action = AUDIO_ACTION_HIGH_EMPTY;

	if(DMA1->HISR & DMA_HISR_TCIF4)
	{
		DMA1->HIFCR = DMA_HIFCR_CTCIF4;

		// means high part is empty
	}

	if(DMA1->HISR & DMA_HISR_HTIF4)
	{
		DMA1->HIFCR = DMA_HIFCR_CHTIF4;

		action = AUDIO_ACTION_LOW_EMPTY;
	}

	int32_t higherTaskWoken;

	xQueueSendFromISR(audio.hActionQueue, &action, &higherTaskWoken);

	if(higherTaskWoken)
		taskYIELD();
}

int16_t AudioPlayWAV(const char* pFilename)
{
	FRESULT fresult = FR_OK;

	SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, DISABLE);

	fresult = f_open(&audio.hAudioFile, pFilename, FA_READ);
	if(fresult != FR_OK)
		return ERROR_AUDIO_INVALID_FILE;

	RIFFHeader riffHeader;
	uint32_t bytesRead;
	fresult = f_read(&audio.hAudioFile, &riffHeader, sizeof(RIFFHeader), (UINT*)&bytesRead);
	if(fresult != FR_OK || bytesRead != sizeof(RIFFHeader))
		return ERROR_AUDIO_INVALID_FILE;

	if(riffHeader.chunkID != 0x46464952) // "RIFF"
	{
		return ERROR_AUDIO_INVALID_FORMAT;
	}

	if(riffHeader.format != 0x45564157) // "WAVE"
	{
		return ERROR_AUDIO_INVALID_FORMAT;
	}

	FMTChunk fmtChunk;
	fresult = f_read(&audio.hAudioFile, &fmtChunk, sizeof(FMTChunk), (UINT*)&bytesRead);
	if(fresult != FR_OK || bytesRead != sizeof(FMTChunk))
		return ERROR_AUDIO_INVALID_FILE;

	if(fmtChunk.chunkID != 0x20746d66) // "fmt "
	{
		return ERROR_AUDIO_INVALID_FORMAT;
	}

	ConsolePrint("NumChannels:   %d\r\n", fmtChunk.numChannels);
	ConsolePrint("SampleRate:    %d\r\n", fmtChunk.sampleRate);
	ConsolePrint("BitsPerSample: %d\r\n", fmtChunk.bitsPerSample);

	if(fmtChunk.audioFormat != 1 || fmtChunk.numChannels != 2 || fmtChunk.sampleRate != 44100
			|| fmtChunk.bitsPerSample != 16)
	{
		return ERROR_AUDIO_INVALID_FORMAT;
	}

	if(fmtChunk.chunkSize+8 > sizeof(FMTChunk))
		f_lseek(&audio.hAudioFile, sizeof(RIFFHeader)+fmtChunk.chunkSize+8);

	DataChunk dataChunk;
	fresult = f_read(&audio.hAudioFile, &dataChunk, sizeof(DataChunk), (UINT*)&bytesRead);
	if(fresult != FR_OK || bytesRead != sizeof(DataChunk))
		return ERROR_AUDIO_INVALID_FILE;

	if(dataChunk.chunkId != 0x61746164) // "data"
	{
		return ERROR_AUDIO_INVALID_FORMAT;
	}

	ConsolePrint("DataSize: %d Bytes\r\n", dataChunk.chunkSize);
	uint16_t secs = dataChunk.chunkSize/(fmtChunk.numChannels*fmtChunk.sampleRate*fmtChunk.bitsPerSample/8);
	ConsolePrint("Play Time: %d:%02d\r\n", secs/60, secs%60);

	// seek to data start
	f_lseek(&audio.hAudioFile, sizeof(RIFFHeader)+fmtChunk.chunkSize+16);

	// load DMA buffer
	fresult = f_read(&audio.hAudioFile, audioBuf, AUDIO_BUF_SAMPLES*2, (UINT*)&bytesRead);
	if(fresult != FR_OK)
	{
		return ERROR_AUDIO_INVALID_FILE;
	}

	xQueueReset(audio.hActionQueue);

	SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, ENABLE);

	audio.filePlaying = 1;

	return 0;
}

void AudioStop()
{
	if(audio.filePlaying == 0)
		return;

	SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, DISABLE);
	f_close(&audio.hAudioFile);
	audio.filePlaying = 0;
}

void AudioSetLevel(uint8_t level)
{
	uint8_t buf8[2];

	SPIHeader header;
	header.slaveId = audio.slaveId;
	header.length = 2;
	header.cmd = 0;

	buf8[0] = 0x10;
	buf8[1] = 128+level;
	SPIPutPacket(&spi3, &header, 2, buf8);

	buf8[0] = 0x11;
	buf8[1] = 128+level;
	SPIPutPacket(&spi3, &header, 2, buf8);

}

void AudioTask(void* params)
{
	params = params;

	audio.hActionQueue = xQueueCreate(1, 1);
	audio.filePlaying = 0;

	RCC_APB1PeriphClockCmd(RCC_APB1Periph_SPI2, ENABLE);

	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_AF;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_Speed = GPIO_Speed_25MHz;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_Pin = GPIO_Pin_12 | GPIO_Pin_13 | GPIO_Pin_15; // WS, CK, SD
	GPIO_Init(GPIOB, &gpioInit);

	gpioInit.GPIO_Pin = GPIO_Pin_6; // MCK
	GPIO_Init(GPIOC, &gpioInit);

	GPIO_PinAFConfig(GPIOB, GPIO_PinSource12, GPIO_AF_SPI2);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource13, GPIO_AF_SPI2);
	GPIO_PinAFConfig(GPIOB, GPIO_PinSource15, GPIO_AF_SPI2);
	GPIO_PinAFConfig(GPIOC, GPIO_PinSource6, GPIO_AF_SPI2);

	I2S_InitTypeDef i2sInit;
	i2sInit.I2S_AudioFreq = I2S_AudioFreq_44k;
	i2sInit.I2S_Standard = I2S_Standard_MSB;
	i2sInit.I2S_DataFormat = I2S_DataFormat_16bextended;
	i2sInit.I2S_CPOL = I2S_CPOL_Low;
	i2sInit.I2S_Mode = I2S_Mode_MasterTx;
	i2sInit.I2S_MCLKOutput = I2S_MCLKOutput_Enable;
	I2S_Init(SPI2, &i2sInit);

	/* Set the parameters to be configured */
	DMA_InitTypeDef dmaInit;
	dmaInit.DMA_Channel = DMA_Channel_0;
	dmaInit.DMA_PeripheralBaseAddr = (uint32_t)&SPI2->DR;
	dmaInit.DMA_Memory0BaseAddr = (uint32_t)audioBuf;
	dmaInit.DMA_DIR = DMA_DIR_MemoryToPeripheral;
	dmaInit.DMA_BufferSize = AUDIO_BUF_SAMPLES;
	dmaInit.DMA_PeripheralInc = DMA_PeripheralInc_Disable;
	dmaInit.DMA_MemoryInc = DMA_MemoryInc_Enable;
	dmaInit.DMA_PeripheralDataSize = DMA_PeripheralDataSize_HalfWord;
	dmaInit.DMA_MemoryDataSize = DMA_MemoryDataSize_HalfWord;
	dmaInit.DMA_Mode = DMA_Mode_Circular;
	dmaInit.DMA_Priority = DMA_Priority_Medium;
	dmaInit.DMA_FIFOMode = DMA_FIFOMode_Enable;
	dmaInit.DMA_FIFOThreshold = DMA_FIFOThreshold_Full;
	dmaInit.DMA_MemoryBurst = DMA_MemoryBurst_Single;
	dmaInit.DMA_PeripheralBurst = DMA_PeripheralBurst_Single;
	DMA_Init(DMA1_Stream4, &dmaInit);

	DMA_Cmd(DMA1_Stream4, ENABLE);

	DMA_ITConfig(DMA1_Stream4, DMA_IT_TC, ENABLE);
	DMA_ITConfig(DMA1_Stream4, DMA_IT_HT, ENABLE);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannel = DMA1_Stream4_IRQn;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_AUDIO_DMA;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&nvicInit);

	SPISlave slave;
	slave.pCSPort = GPIOB;
	slave.csPin = GPIO_Pin_5;
	slave.cpha = 0;
	slave.cpol = 0;
	slave.prescaler = SPI_BaudRatePrescaler_256;
	slave.callback = 0;
	slave.pUserData = 0;

	SPIAddSlave(&spi3, slave, &audio.slaveId);

	I2S_Cmd(SPI2, ENABLE);

	vTaskDelay(100/portTICK_RATE_MS);

	memset(audioBuf, 0, sizeof(uint16_t)*AUDIO_BUF_SAMPLES);

	SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, ENABLE);

	vTaskDelay(10/portTICK_RATE_MS);

	AudioSetLevel(0);

	vTaskDelay(100/portTICK_RATE_MS);

	SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, DISABLE);

	xQueueReset(audio.hActionQueue);

	uint8_t action;
	FRESULT fresult;
	uint32_t bytesRead;

	while(42)
	{
		xQueueReceive(audio.hActionQueue, &action, portMAX_DELAY);

		if(action == AUDIO_ACTION_LOW_EMPTY)
		{
			fresult = f_read(&audio.hAudioFile, audioBuf, AUDIO_BUF_SAMPLES, (UINT*)&bytesRead);
		}
		else
		{
			fresult = f_read(&audio.hAudioFile, audioBuf+(AUDIO_BUF_SAMPLES/2), AUDIO_BUF_SAMPLES, (UINT*)&bytesRead);
		}

		if(fresult != FR_OK)
		{
			ConsolePrint("Reading audio data failed\r\n");
		}

		if(bytesRead < AUDIO_BUF_SAMPLES)
		{
			ConsolePrint("End of WAV file reached\r\n");
		}

		if(fresult != FR_OK || bytesRead < AUDIO_BUF_SAMPLES)
		{
			SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, DISABLE);
			f_close(&audio.hAudioFile);
			audio.filePlaying = 0;
		}
	}
}
