/*
 * cli.c
 *
 *  Created on: 19.11.2012
 *      Author: AndreR
 */

#include "cli.h"

#include "util/console.h"
#include "intercom_constants.h"
#include "hal/led.h"
#include "hal/rtc.h"
#include "hal/tft.h"
#include "hal/ov7725.h"
#include "hal/usart2.h"
#include "hal/sram.h"
#include "hal/rtos_hf_timer.h"
#include "intercom1.h"
#include "spi3.h"

#include "FreeRTOS.h"
#include "task.h"
#include "display.h"
#include "audio.h"
#include "gui.h"

#include "usbh_def.h"
#include "usbh_hid_core.h"
#include "usb.h"

CLI cli;

#define SRAM_TEST_SIZE 512*1024
static volatile EXT_SRAM uint16_t sram[SRAM_TEST_SIZE];

void processCommand();

void CLIInit()
{
	cli.hProcBufMutex = xSemaphoreCreateRecursiveMutex();

	static uint8_t rxData[CONSOLE_RX_SIZE];
	static uint8_t txData[CONSOLE_TX_SIZE];
	static uint8_t printData[CONSOLE_PRINT_SIZE];
	static uint8_t scanData[CONSOLE_SCAN_SIZE];

	ConsoleData conData;
	conData.pDataTx = txData;
	conData.txSize = CONSOLE_TX_SIZE;
	conData.pDataRx = rxData;
	conData.rxSize = CONSOLE_RX_SIZE;
	conData.pPrintBuf = printData;
	conData.printSize = CONSOLE_PRINT_SIZE;
	conData.pScanBuf = scanData;
	conData.scanSize = CONSOLE_SCAN_SIZE;
	conData.txCallback = 0;
	conData.pUserData = 0;

	ConsoleInit(conData);

	static uint8_t adapterBuf[CONSOLE_INTERCOM_ADAPTER_BUF_SIZE];

	ConsoleIntercomAdapterInit(&cli.adapter, &intercom1, adapterBuf,
			CONSOLE_INTERCOM_ADAPTER_BUF_SIZE, INTERCOM_ADDRESS_MEDIA, INTERCOM_ADDRESS_MAIN);
}

void CLITask(void* params)
{
	params = params;

	portTickType xLastWakeTime;
	portTickType xLastLEDTime;

	xLastWakeTime = xTaskGetTickCount();
	xLastLEDTime = xLastWakeTime;

	LEDMustOff(LED_MAIN_RED);

	while(1)
	{
		if(ConsoleIsCmdReady())
		{
			processCommand();

			ConsoleCmdProcessed();
		}

		// Heartbeat LED
		if(xLastWakeTime - xLastLEDTime > 500/portTICK_RATE_MS)
		{
			xLastLEDTime = xLastWakeTime;

			LEDToggle(LED_MAIN_GREEN);
		}

		vTaskDelayUntil(&xLastWakeTime, 50/portTICK_RATE_MS);
	}
}

void processCommand()
{
	static const uint16_t bufLen = 128;
	static uint8_t buf8[128];
	static uint16_t buf16[10];
	static uint32_t buf32[10];
	uint32_t time;

	if(ConsoleCmpCmd("test"))
	{
		ConsolePrint("Number: %d\r\n", 43);
	}

	if(ConsoleCmpCmd("test2"))
	{
		ConsolePrint("0123456789012345678901234567890123456789012345678901234567890123456789012345678901234567890123456789\r\n");
	}

	if(ConsoleCmpCmd("task list"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tState\tPrio\tStack\tID");

		ConsoleLockPrintBuf();
		vTaskList((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	if(ConsoleCmpCmd("task stats"))
	{
		char* printBuf = ConsoleGetPrintBuf();

		ConsolePrint("\r\nTask\t\tTicks\t\tRatio");

		ConsoleLockPrintBuf();
		vTaskGetRunTimeStats((signed char*)printBuf);
		ConsoleSendPrintBuf();
		ConsoleUnlockPrintBuf();
	}

	if(ConsoleCmpCmd("time"))
	{
		DateTime time = RTCGetDateTime();

		ConsolePrint("%02u.%02u.%4u %02u:%02u:%02u\r\n", time.day, time.month, time.year, time.hour, time.min, time.sec);
	}

	if(ConsoleScanCmd("settime %u", &time) == 1)
	{
		RTCSetTime(time);

		DateTime now = RTCGetDateTime();

		ConsolePrint("%02u.%02u.%4u %02u:%02u:%02u\r\n", now.day, now.month, now.year, now.hour, now.min, now.sec);
	}

	uint16_t tft[7];
	int32_t numArgs = 0;
	if((numArgs = ConsoleScanCmd("tft wr %hX %hX %hX %hX %hX %hX %hX", tft, tft+1, tft+2, tft+3, tft+4, tft+5, tft+6)) > 0)
	{
		ConsolePrint("REG: 0x%X\r\n", tft[0]);
		if(numArgs > 1)
			ConsolePrint("Param1: 0x%X\r\n", tft[1]);
		if(numArgs > 2)
			ConsolePrint("Param2: 0x%X\r\n", tft[2]);
		if(numArgs > 3)
			ConsolePrint("Param3: 0x%X\r\n", tft[3]);
		if(numArgs > 4)
			ConsolePrint("Param4: 0x%X\r\n", tft[4]);
		if(numArgs > 5)
			ConsolePrint("Param5: 0x%X\r\n", tft[5]);
		if(numArgs > 6)
			ConsolePrint("Param6: 0x%X\r\n", tft[6]);

		*TFT_CMD_REG = tft[0];
		if(numArgs > 1)
			*TFT_DATA_REG = tft[1];
		if(numArgs > 2)
			*TFT_DATA_REG = tft[2];
		if(numArgs > 3)
			*TFT_DATA_REG = tft[3];
		if(numArgs > 4)
			*TFT_DATA_REG = tft[4];
		if(numArgs > 5)
			*TFT_DATA_REG = tft[5];
		if(numArgs > 6)
			*TFT_DATA_REG = tft[6];
	}

	if(ConsoleScanCmd("tft rd %hX %hd", tft, tft+1) == 2)
	{
		ConsolePrint("REG: 0x%X\r\n", tft[0]);

		*TFT_CMD_REG = tft[0];

		for(uint16_t i = 0; i < tft[1]; i++)
			ConsolePrint("PAR%u: 0x%X\r\n", i, *TFT_DATA_REG);
	}

	if(ConsoleScanCmd("tft brightness %d", buf32) == 1)
	{
		ConsolePrint("TFT brightness set to: %d\r\n");
		TFTSetBrightness((uint8_t)buf32[0]);
	}

	uint32_t rgb;
	if(ConsoleScanCmd("tft fill %X", &rgb) == 1)
	{
		ConsolePrint("Sending data to display\r\n");

		TFTFillAll(rgb);
	}


	static uint16_t graphObjX = 1;
	static uint16_t graphObjY = 1;
	uint8_t graphObjRotation;
	uint8_t graphObjType;
	/*
	if(ConsoleScanCmd("testimg %ld %ld %d", &graphObjX, &graphObjY, &graphObjRotation) == 3)
	{
		if (graphObjRotation==90)
			graphObjRotation=1;
		if (graphObjRotation==180)
			graphObjRotation=2;
		if (graphObjRotation==270)
			graphObjRotation=3;

		if (graphObjRotation>3 )
			ConsolePrint("not supported rotation mode\r\n");
		else
		{
			switch (graphObjRotation)
			{
				case 0: ConsolePrint("Test image is not rotated\r\n"); break;
				case 1: ConsolePrint("Test image is rotated by 90 degrees anti-clockwise\r\n"); break;
				case 2: ConsolePrint("Test image is rotated by 180 degrees\r\n"); break;
				case 3: ConsolePrint("Test image is rotated by 90 degrees clockwise\r\n"); break;
			} // switch
		}//else

		TFTDrawTestImage(graphObjX, graphObjY, (uint8_t)graphObjRotation);
	}
*/

	static uint32_t graphObjBGColor, graphObjFGColor;

/*
	if (ConsoleScanCmd("write at %hd %hd %X %X %127[^\n]", &graphObjX, &graphObjY, &graphObjBGColor, &graphObjFGColor, buf8)== 5)
	{
		ConsolePrint("Writing %s at ( %d | %d ) with background color %X and foreground color %X\r\n", buf8, graphObjX, graphObjY,  graphObjBGColor, graphObjFGColor );
		GUIDrawString( graphObjX, graphObjY, buf8, graphObjBGColor, graphObjFGColor, 1 );
	}
*/
	static uint16_t btnWidth;

	if (ConsoleScanCmd("button at %hd %hd %hd %X %X %127[^\n]", &graphObjX, &graphObjY, &btnWidth, &graphObjBGColor, &graphObjFGColor, buf8)==6 )
	{
		ConsolePrint("Drawing button that reads %s at ( %d | %d ) with minimum width %d background color %X and foreground color %X\r\n", buf8, graphObjX, graphObjY, btnWidth, graphObjBGColor, graphObjFGColor );
		//DisplayAddButton( graphObjX, graphObjY, buf8, graphObjBGColor, graphObjFGColor, btnWidth );
	}

	if (ConsoleScanCmd("reset button %d", buf32) == 1)
	{
		ConsolePrint("removed touch sensitivity for button %d\r\n", *buf32);
		//DisplayResetButton(*buf32);
	}

	if(ConsoleCmpCmd("reset all buttons"))
	{
		ConsolePrint("removed touch sensitivity for all buttons\r\n");
		//DisplayResetAllButtons();
	}

/*
	if(ConsoleScanCmd("draw num pad %X %X %d %d", &graphObjBGColor, &graphObjFGColor, &graphObjType, &graphObjRotation) == 4)
	{
		ConsolePrint("Drawing num pad with BG color %d, FG color%d and rotation %d\r\n", graphObjBGColor, graphObjFGColor, graphObjRotation);
		GUIDrawNumPad(graphObjBGColor, graphObjFGColor, graphObjType, graphObjRotation);
	}


	if(ConsoleScanCmd("init num pad %d %d %d", &graphObjBGColor, &graphObjFGColor, &graphObjRotation) == 3)
	{
		ConsolePrint("Activating touch sensitivity for num pad with BG color %d, FG color%d and rotation %d", graphObjBGColor, graphObjFGColor, graphObjRotation);
		DisplayAddNumPad(uint32_t bGColor, uint32_t fGColor, uint8_t rotation);
	}
*/

	if(ConsoleCmpCmd("tft pattern"))
	{
		ConsolePrint("Drawing test pattern\r\n");

		TFTPattern();
	}

	if(ConsoleScanCmd("tft tracking %d", buf32) == 1)
	{
		ConsolePrint("Tracking demo state: %d\r\n", *buf32);

		display.trackingDemo = *buf32;
	}

	if(ConsoleScanCmd("tft rect %u %u %u %u %x", buf32, buf32+1, buf32+2, buf32+3, buf32+4) == 5)
	{
		ConsolePrint("TFT rect: (%u,%u) (%u,%u), Color: %06X\r\n", buf32[0], buf32[1], buf32[2], buf32[3], buf32[4]);

		TFTDrawRect(RectU16Coordinates(buf32[0], buf32[1], buf32[2], buf32[3]), buf32[4]);
	}

	if(ConsoleScanCmd("tft fill %u %u %u %u %x", buf32, buf32+1, buf32+2, buf32+3, buf32+4) == 5)
	{
		ConsolePrint("TFT fill part: (%u,%u) (%u,%u), Color: %06X\r\n", buf32[0], buf32[1], buf32[2], buf32[3], buf32[4]);

		TFTFillRect(RectU16Coordinates(buf32[0], buf32[1], buf32[2], buf32[3]), buf32[4]);
	}

	if(ConsoleCmpCmd("ts test"))
	{
		uint8_t buf[16];

		SPIHeader header;
		header.cmd = 0;
		header.length = 2;
		header.slaveId = display.slaveId;

		buf[0] = 0x01 << 1 | 0x01;

		SPIPutPacket(&spi3, &header, 1, buf);

		ConsolePrint("TS Data req sent\r\n");
	}

	if(ConsoleCmpCmd("audio on"))
	{
		SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, ENABLE);
	}

	if(ConsoleCmpCmd("audio off"))
	{
		SPI_I2S_DMACmd(SPI2, SPI_I2S_DMAReq_Tx, DISABLE);
	}

	if(ConsoleScanCmd("audio lvl %d", buf32) == 1)
	{
		AudioSetLevel(buf32[0]);
	}

	if(ConsoleCmpCmd("audio stop"))
	{
		AudioStop();
	}

	if(ConsoleCmpCmd("usart2 state"))
	{
		ConsolePrint("TX: %d\r\n", usart2.txState);
	}

	if(ConsoleCmpCmd("cam"))
	{
		ConsolePrint("Reading camera MFG\r\n");

		uint8_t mfg1;
		uint8_t mfg2;
		if(OV7725ReadReg(OV7725_MIDL, &mfg1))
			ConsolePrint("Reading OV7725_MIDL failed\r\n");
		if(OV7725ReadReg(OV7725_MIDH, &mfg2))
			ConsolePrint("Reading OV7725_MIDH failed\r\n");

		ConsolePrint("Manufacturer: 0x%X 0x%X\r\n", mfg1, mfg2);
	}

	if(ConsoleCmpCmd("cam acq"))
	{
		ConsolePrint("Acquiring CAM image...\r\n");

		if(OV7725AcquireImage() == 0)
		{
			*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;

			uint16_t* img = (uint16_t*)OV7725GetPicBuf();

			for(uint32_t p = 0; p < 76800; p++)
			{
				*TFT_DATA_REG = img[p];
			}

			ConsolePrint("Success\r\n");
		}
		else
			ConsolePrint("Failed\r\n");
	}

	if(ConsoleScanCmd("cam stream %hhd", buf8) == 1)
	{
		ConsolePrint("Setting stream mode to: %hhd\r\n", *buf8);

		OV7725SetMode(*buf8);
	}

	if(ConsoleScanCmd("cam wr %hX %hhX", buf16, buf8) == 2)
	{
		ConsolePrint("Addr 0x%hX: 0x%hhX...", *buf16, *buf8);

		if(OV7725WriteReg(*buf16, *buf8))
			ConsolePrint("Error\r\n");
		else
			ConsolePrint("OK\r\n");
	}

	if(ConsoleCmpCmd("sram test"))
	{
		ConsolePrint("SRAM test. Addr: 0x%X\r\n", (uint32_t)sram);

		uint16_t res1 = 0;
		uint16_t res2 = 0;

		sram[2] = 0x0000;
		vTaskDelay(2/portTICK_RATE_MS);
		res1 = sram[2];
		vTaskDelay(2/portTICK_RATE_MS);
		sram[2] = 0xFFFF;
		vTaskDelay(2/portTICK_RATE_MS);
		res2 = sram[2];

		ConsolePrint("Out: 0x0000, In: 0x%hX\r\n", res1);
		ConsolePrint("Out: 0xFFFF, In: 0x%hX\r\n", res2);


		for(uint32_t i = 0; i < SRAM_TEST_SIZE; i++)
		{
			sram[i] = 0x0000;

			vTaskDelay(2/portTICK_RATE_MS);
		}

		for(uint32_t i = 0; i < SRAM_TEST_SIZE; i++)
		{
			res1 = sram[i];
			if(res1 != 0x0000)
			{
				ConsolePrint("Test failed at address: 0x%X, expected: 0x0000, read: 0x%hX\r\n", i, res1);
//				break;
			}

			vTaskDelay(10/portTICK_RATE_MS);
		}

		ConsolePrint("Test done\r\n");
	}

	if(ConsoleScanCmd("hid %hhd", buf8))
	{
		ConsolePrint("Requesting HID Descriptor for Interface %hhd\r\n", *buf8);

		USBH_HIDDesc_TypeDef hid;

		int16_t result = USBH_Get_HID_Descriptor(&USB_OTG_Core, &USB_Host, &hid, *buf8);
		if(result)
		{
			ConsolePrint("Error: 0x%hX\r\n", result);
			return;
		}

		USBH_HID_PrintHIDDescriptor(&hid);

		ConsolePrint("Requesting HID Report Descriptor\r\n");

		result = USBH_Get_HID_ReportDescriptor(&USB_OTG_Core, &USB_Host, USB_OTG_Core.host.Rx_Buffer, hid.wItemLength, *buf8);
		if(result)
		{
			ConsolePrint("Error: 0x%X\r\n", result);
			return;
		}

		uint8_t* pReport = USB_OTG_Core.host.Rx_Buffer;

		USBH_HID_PrintHIDReport(pReport, hid.wItemLength);

		uint16_t reportSize = USBH_HID_GetReportSize(pReport, hid.wItemLength);

		ConsolePrint("Report Size: %hd\r\n", reportSize);
	}

//	if(sd.mounted)
	{
		FRESULT fresult;

		if(ConsoleCmpCmd("pwd"))
		{
			f_getcwd((TCHAR*)buf8, bufLen);

			ConsolePrint("%s\r\n", buf8);
		}

		if(ConsoleCmpCmd("ls"))
		{
			DIR dir;
			FILINFO finfo;
			finfo.lfname = (TCHAR*)buf8;
			finfo.lfsize = bufLen-1;

			f_getcwd((TCHAR*)buf8, bufLen);

			ConsolePrint("Content of: %s\r\n", buf8);

			fresult = f_opendir(&dir, "");
			if(fresult != FR_OK)
				return;

			do
			{
				fresult = f_readdir(&dir, &finfo);

				if(finfo.fname[0] == 0)
					break;

				if(finfo.fattrib & AM_DIR)
					ConsolePrint("%s/\r\n", *finfo.lfname ? finfo.lfname : finfo.fname);
				else
					ConsolePrint("%s\r\n", *finfo.lfname ? finfo.lfname : finfo.fname);
			}
			while(fresult == FR_OK);
		}

		if(ConsoleScanCmd("cd %127[^\n]", buf8) == 1)
		{
			f_chdir((TCHAR*)buf8);

			f_getcwd((TCHAR*)buf8, bufLen);

			ConsolePrint("%s\r\n", buf8);
		}

		if(ConsoleScanCmd("touch %127[^\n]", buf8) == 1)
		{
			FIL file;
			fresult = f_open(&file, (TCHAR*)buf8, FA_CREATE_NEW);
			if(fresult == FR_EXIST)
				ConsolePrint("File already exists\r\n");
			f_close(&file);
		}

		if(ConsoleScanCmd("mkdir %127[^\n]", buf8) == 1)
		{
			fresult = f_mkdir((TCHAR*)buf8);
			if(fresult == FR_OK)
				ConsolePrint("Created dir: %s\r\n", buf8);
		}

		if(ConsoleScanCmd("rm %127[^\n]", buf8) == 1)
		{
			fresult = f_unlink((TCHAR*)buf8);
			if(fresult != FR_OK)
				ConsolePrint("Could not remove object\r\n");
		}

		if(ConsoleScanCmd("play %127[^\n]", buf8) == 1)
		{
			if(AudioPlayWAV((char*)buf8) != 0)
				ConsolePrint("Playing WAV failed\r\n");
		}

		if(ConsoleScanCmd("read %u %127[^\n]", buf32, buf8) == 2)
		{
			ConsolePrint("Reading %uB from file: %s\r\n", *buf32, buf8);

			FIL file;
			fresult = f_open(&file, (TCHAR*)buf8, FA_READ);
			if(fresult != FR_OK)
			{
				ConsolePrint("Open file failed\r\n");
				return;
			}

			uint32_t bytesRead;
			uint32_t timeStart = rtosHfTickCount;
			fresult = f_read(&file, (void*)sram, *buf32, (UINT*)&bytesRead);
			uint32_t timeEnd = rtosHfTickCount;
			ConsolePrint("Result: %d\r\n", fresult);
			ConsolePrint("Bytes read: %u\r\n", bytesRead);
			float timeMs = ((float)timeEnd-timeStart)*0.02f;
			ConsolePrint("Time: %fms\r\n", timeMs);
		}
	}
}
