/*
 * display.c
 *
 *  Created on: 06.12.2012
 *      Author: AndreR
 */

#include "display.h"

#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"
#include "util/console.h"
#include "util/bit_banding.h"

#include "spi3.h"
#include "hal/tft.h"
#include "constants.h"

#include "stm32f4xx_exti.h"
#include "stm32f4xx_syscfg.h"
#include "misc.h"

#include "hal/tft.h"

#include "cli.h"

#include "gui.h"
#include <string.h>
#include <math.h>

Display display;

#define CMD_SET_MODE	1
#define CMD_READ_STATUS	2
#define CMD_MEAS_XYZ	3
#define CMD_READ_XYZ	4

#define TOUCH_X_MIN		2048	// 204.8*10
#define TOUCH_X_MAX		63488	// 204.8*310
#define TOUCH_Y_MIN		2730	// 273.066*10
#define TOUCH_Y_MAX		62805	// 273.066*230
#define TOUCH_X_RANGE	(TOUCH_X_MAX-TOUCH_X_MIN)
#define TOUCH_Y_RANGE	(TOUCH_Y_MAX-TOUCH_Y_MIN)

typedef struct _TSMeas
{
	uint32_t x;
	uint32_t y;
	uint32_t z1;
	uint32_t z2;
} TSMeas;

// touch event active
void EXTI2_IRQHandler()
{
	BBPeriphSet((uint32_t)&EXTI->PR, 2);

	signed portBASE_TYPE xHigherPriorityTaskWoken = pdFALSE;

	xSemaphoreGiveFromISR(display.hTouch, &xHigherPriorityTaskWoken);

	if(xHigherPriorityTaskWoken == pdTRUE)
		taskYIELD();
}

static void tsCallback (SPIHeader* pHeader, uint8_t* pData, void* pUserData)
{
	pUserData = pUserData;

	if(pHeader->cmd == CMD_READ_XYZ)
	{
		TSMeas result;
		result.x =  pData[1] << 8;
		result.x |= pData[2];
		result.y =  pData[3] << 8;
		result.y |= pData[4];
		result.z1 =  pData[5] << 8;
		result.z1 |= pData[6];
		result.z2 =  pData[7] << 8;
		result.z2 |= pData[8];

		result.y = 65536-result.y;	// Y starts at 0 in lower left corner, so we invert it to match pixel positions

		float x = result.x;
		float z1 = result.z1;
		float z2 = result.z2;

		display.touchZ1 = result.z1;
		display.touchZ2 = result.z2;

		display.touchZ = (x/65536.0f)*((z2/z1)-1.0f);
		if(display.touchZ < 2.0f && z1 > 1000.0f && z2 < 65000.0f
			&& result.x > TOUCH_X_MIN && result.x < TOUCH_X_MAX && result.y > TOUCH_Y_MIN && result.y < TOUCH_Y_MAX)
		{
			display.touchX = ((result.x-TOUCH_X_MIN)*LCD_WIDTH)/TOUCH_X_RANGE;
			display.touchY = ((result.y-TOUCH_Y_MIN)*LCD_HEIGHT)/TOUCH_Y_RANGE;

			if(display.touchDown == 0)
				display.touchPressed = 1;

			display.touchDown = 1;
		}
		else
		{
			if(display.touchDown)
				display.touchReleased = 1;

			display.touchDown = 0;
		}
	}
}

void DisplayInit()
{
	vSemaphoreCreateBinary(display.hTouch);

	// TP-IRQ: E2
	GPIO_InitTypeDef gpioInit;
	gpioInit.GPIO_Mode = GPIO_Mode_IN;
	gpioInit.GPIO_OType = GPIO_OType_OD;
	gpioInit.GPIO_PuPd = GPIO_PuPd_UP;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_Pin = GPIO_Pin_2;
	GPIO_Init(GPIOE, &gpioInit);

	SYSCFG_EXTILineConfig(EXTI_PortSourceGPIOE, EXTI_PinSource2);

	EXTI_InitTypeDef extiInit;
	EXTI_StructInit(&extiInit);
	extiInit.EXTI_Line = EXTI_Line2;
	extiInit.EXTI_LineCmd = ENABLE;
	extiInit.EXTI_Mode = EXTI_Mode_Interrupt;
	extiInit.EXTI_Trigger = EXTI_Trigger_Rising_Falling;
	EXTI_Init(&extiInit);

	NVIC_InitTypeDef nvicInit;
	nvicInit.NVIC_IRQChannelCmd = ENABLE;
	nvicInit.NVIC_IRQChannelPreemptionPriority = IRQL_EXTI2;
	nvicInit.NVIC_IRQChannelSubPriority = 0;
	nvicInit.NVIC_IRQChannel = EXTI2_IRQn;
	NVIC_Init(&nvicInit);

	display.slaveId = 0;
	display.touchDown = 0;
	display.touchPressed = 0;
	display.touchReleased = 0;

	SPISlave tsSlave;
	tsSlave.pCSPort = GPIOD;
	tsSlave.csPin = GPIO_Pin_6;
	tsSlave.cpol = 0;
	tsSlave.cpha = 0;
	tsSlave.prescaler = SPI_BaudRatePrescaler_2;
	tsSlave.callback = &tsCallback;
	tsSlave.pUserData = 0;
	SPIAddSlave(&spi3, tsSlave, &display.slaveId);

	SPIHeader header;
	header.cmd = CMD_SET_MODE;
	header.length = 12;
	header.slaveId = display.slaveId;

	uint8_t buf[14];
	buf[0] = 0x03 << 1;
	buf[1] = 0xFF; // 16 samples, median filter
	buf[2] = 0x07 << 1;
	buf[3] = 0x60; // set PUR to 1ms
	buf[4] = 0x0B << 1; // write Mode Config Register
	buf[5] = 0x06; // avg filter on x,y,z1,z2
	buf[6] = 0x04 << 1;
	buf[7] = 0x55; // ADC sample time 128us on all channels
	buf[8] = 0x05 << 1;
	buf[9] = 0x44; // Panel setup time 1ms
	buf[10]= 0x06 << 1;
	buf[11]= 0x00; // Delayed conversion 1ms
	buf[12]= 0x01 << 1;
	buf[13]= 0xC0;	//

	SPIPutPacket(&spi3, &header, 12, buf);

	display.trackingDemo = 0;
}

void DisplayTask(void* params)
{
	params = params;

	uint8_t buf[16];

	SPIHeader header;
	header.slaveId = display.slaveId;

	while(42)
	{
		header.cmd = CMD_MEAS_XYZ;
		header.length = 1;
		buf[0] = 0x74 << 1;
		SPIPutPacket(&spi3, &header, 1, buf);

		vTaskDelay(10/portTICK_RATE_MS);

		header.cmd = CMD_READ_XYZ;
		header.length = 9;
		buf[0] = 0x52 << 1 | 0x01;
		SPIPutPacket(&spi3, &header, 1, buf);

//		if(display.touchDown)
//			ConsolePrint("Touch: (%d,%d), Z: %f (%d,%d)\r\n", display.touchX, display.touchY, display.touchZ, display.touchZ1, display.touchZ2);

		GUIHandler();

		// the tracking demo enables exactly one pixel where the display was last touched
		// used to test touch-to-display coordinate mapping
		if(display.trackingDemo && display.touchDown)
		{
			*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;

			for(uint16_t y = 0; y < 240; y++)
			{
				for(uint16_t x = 0; x < 320; x++)
				{
					if(display.touchX == x && display.touchY == y)
						TFTWritePixel(0xFF00);
					else
						TFTWritePixel(0);
				}
			}
		}

	}
}
