/*
 * tft.c
 *
 *  Created on: 04.12.2012
 *      Author: AndreR
 */
#include <string.h>
#include "tft.h"

#include "stm32f4xx.h"
#include "stm32f4xx_rcc.h"
#include "stm32f4xx_fsmc.h"
#include "stm32f4xx_gpio.h"

#include "FreeRTOS.h"
#include "task.h"
#include "gui.h"

#define LCD_Panel_Data_Width				0//0:18bit; 1:24bit;
#define LCD_Pixel_Data_Interface_Format		3 // 16-bit (565 format)

#define LCD_DotClk_Polarity			1 //0:Data latch in rising edge; 1:Data latch in falling edge
#define LCD_Hsync_Polarity			0 //0:Active low; 1:Active high
#define LCD_Vsync_Polarity			0 //0:Active low; 1:Active high

#define LCD_Panel_Mode				0 //0:Hsync+Vsync +DE mode; 1:TTL mode
#define LCD_TFT_Type				0 //0 or 1:TFT mode; 2:Serial RGB mode; 3:Serial RGB+dummy mode

#define LCD_Panel_Mode1	(LCD_Panel_Data_Width<<5)|(0<<4)|(0<<3)|(LCD_DotClk_Polarity<<2)|(LCD_Hsync_Polarity<<1)|(LCD_Vsync_Polarity)
#define LCD_Panel_Mode2	(LCD_Panel_Mode<<7)|(LCD_TFT_Type<<5)

//Horizontal Cycle = Horizontal Pulse Width + Horizontal Back Porch + Horizontal Display Period + Horizontal Front Porch
//Horizontal Start Position = Horizontal Pulse Width + Horizontal Back Porch
#define LCD_HORI_FRONT_PORCH		70//0x90//0x30//2
#define LCD_HORI_BACK_PORCH			8//0x05//0x0F//2
#define LCD_HORI_PULSE_WIDTH		2//0x1e//0x60//41

//Vertical Cycle = Vertical Pulse Width + Vertical Back Porch + Vertical Display Period + Vertical Front Porch
//Vertical Start Position = Vertical Pulse Width + Vertical Back Porch
#define LCD_VERT_FRONT_PORCH		13//0x20//3//2
#define LCD_VERT_BACK_PORCH			8//2//0//2
#define LCD_VERT_PULSE_WIDTH		2//2//10

#define LCD_HORI_CYCLE				(LCD_HORI_PULSE_WIDTH + LCD_HORI_BACK_PORCH + LCD_WIDTH + LCD_HORI_FRONT_PORCH)
#define LCD_VERT_CYCLE				(LCD_VERT_PULSE_WIDTH + LCD_VERT_BACK_PORCH + LCD_HEIGHT + LCD_VERT_FRONT_PORCH)

#define LCD_HORI_START				(LCD_HORI_PULSE_WIDTH + LCD_HORI_BACK_PORCH)
#define LCD_VERT_START				(LCD_VERT_PULSE_WIDTH + LCD_VERT_BACK_PORCH)

#define TXT_LEN(a)            strlen((const char *)a)*8

RectU16 TFTDisplayRect = { .a = {0, 0}, .b = {LCD_HPS, LCD_VPS} };

void TFTSetPixelArea(RectU16 rect, uint8_t rotation);

static void regInit()
{
	*TFT_CMD_REG = TFT_REG_SOFT_RESET;	// soft reset
	vTaskDelay(10/portTICK_RATE_MS);

	// set LSHIFT frequency (DotClk)
	*TFT_CMD_REG = TFT_REG_SET_LSHIFT_FREQ;
	*TFT_DATA_REG = 0x01;
	*TFT_DATA_REG = 0x90;
	*TFT_DATA_REG = 0xFF;

	// Set LCD mode
	*TFT_CMD_REG = TFT_REG_SET_LCD_MODE;
	*TFT_DATA_REG = LCD_Panel_Mode1; //  24Bit data width, dot clk polatrity 1, vsync active low, hsync active low
	*TFT_DATA_REG = LCD_Panel_Mode2; // Mode: hsync+vsync +DE, TFT mode
	*TFT_DATA_REG = LCD_HPS >> 8;
	*TFT_DATA_REG = LCD_HPS;
	*TFT_DATA_REG = LCD_VPS >> 8;
	*TFT_DATA_REG = LCD_VPS;
	*TFT_DATA_REG = 0x00; // RGB sequence

	// Set Horizontal Period
	*TFT_CMD_REG = TFT_REG_SET_HORI_PERIOD;
	*TFT_DATA_REG = LCD_HORI_CYCLE >> 8;
	*TFT_DATA_REG = LCD_HORI_CYCLE;
	*TFT_DATA_REG = LCD_HORI_FRONT_PORCH >> 8;
	*TFT_DATA_REG = LCD_HORI_FRONT_PORCH;
	*TFT_DATA_REG = LCD_HORI_PULSE_WIDTH;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;

	// Set Vertical Period
	*TFT_CMD_REG = TFT_REG_SET_VERT_PERIOD;
	*TFT_DATA_REG = LCD_VERT_CYCLE >> 8;
	*TFT_DATA_REG = LCD_VERT_CYCLE;
	*TFT_DATA_REG = LCD_VERT_FRONT_PORCH >> 8;
	*TFT_DATA_REG = LCD_VERT_FRONT_PORCH;
	*TFT_DATA_REG = LCD_VERT_PULSE_WIDTH;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;

	// Set GPIO value
	*TFT_CMD_REG = TFT_REG_SET_GPIO_VALUE;
	*TFT_DATA_REG = 0x0F;

	vTaskDelay(100/portTICK_RATE_MS);

	// Display on
	*TFT_CMD_REG = TFT_REG_SET_DISPLAY_ON;

	vTaskDelay(100/portTICK_RATE_MS);

	// Set column address
	*TFT_CMD_REG = TFT_REG_SET_COLUMN_ADDRESS;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = LCD_HPS >>8;
	*TFT_DATA_REG = LCD_HPS;

	// Set Page address
	*TFT_CMD_REG = TFT_REG_SET_PAGE_ADDRESS;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = LCD_VPS >> 8;
	*TFT_DATA_REG = LCD_VPS;

	// Set address mode
	*TFT_CMD_REG = TFT_REG_SET_ADDRESS_MODE;
	*TFT_DATA_REG = 0x00;

	// Set pixel data interface
	*TFT_CMD_REG = TFT_REG_SET_PIXEL_DATA_INTERFACE;
	*TFT_DATA_REG = LCD_Pixel_Data_Interface_Format;

	// Set GPIO config
	*TFT_CMD_REG = TFT_REG_SET_GPIO_CONF;
	*TFT_DATA_REG = 0x0F; // GPIO 0-3 as output
	*TFT_DATA_REG = 0x01; // normal GPIOs

	// Set GPIO value
	*TFT_CMD_REG = TFT_REG_SET_GPIO_VALUE;
	*TFT_DATA_REG = 0x0F;

	// Set PWM config
	*TFT_CMD_REG = TFT_REG_SET_PWM_CONF;
	*TFT_DATA_REG = 0x0E;
	*TFT_DATA_REG = 0xFF;
	*TFT_DATA_REG = 0x01;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x0F;
}

#define DEG_0 0
#define DEG_90 1
#define DEG_180 2
#define DEG_270 3

/**
 * Set paint area. This includes rect.b.x and rect.b.y.
 */
void TFTSetPixelArea(RectU16 rect, uint8_t rotation)
{
	if(rotation > 3)
		rotation=0;

	// 1. Set mode for frame buffer
	*TFT_CMD_REG = TFT_REG_SET_ADDR_MODE;

	switch (rotation)
	{
	case DEG_0: *TFT_DATA_REG= 0; 		break;
	case DEG_90: *TFT_DATA_REG= 0xA0;  	break;   //BIT7 and BIT5 are set
	case DEG_180: *TFT_DATA_REG= 0xC0; 	break;  // BIT7 and BIT6 are set
	case DEG_270: *TFT_DATA_REG= 0x60;	break; // BIT6 and BIT5 are set
	}

	// 2.1 Set Column Address
	*TFT_CMD_REG = TFT_REG_SET_COL_ADDR;

	*TFT_DATA_REG = rect.a.x >> 8;
	*TFT_DATA_REG = rect.a.x;
	*TFT_DATA_REG = (rect.b.x) >> 8;
	*TFT_DATA_REG =  rect.b.x;

	// 2.2 Set Page Address
	*TFT_CMD_REG = TFT_REG_SET_PG_ADDR;
	*TFT_DATA_REG = rect.a.y >> 8;
	*TFT_DATA_REG = rect.a.y;
	*TFT_DATA_REG = (rect.b.y) >> 8;
	*TFT_DATA_REG =  rect.b.y;
}

void TFTWritePixel(uint32_t color)
{
	*TFT_DATA_REG =TFT_COLOR(color);
}


void TFTInit()
{
	FSMC_NORSRAMInitTypeDef fsmcInit;
	FSMC_NORSRAMTimingInitTypeDef fsmcTiming;
	FSMC_NORSRAMTimingInitTypeDef fsmcTimingWrite;
	GPIO_InitTypeDef gpioInit;

	// Configure RESET Pin
	GPIOG->BSRRL = GPIO_Pin_15;

	gpioInit.GPIO_Mode = GPIO_Mode_OUT;
	gpioInit.GPIO_Speed = GPIO_Speed_2MHz;
	gpioInit.GPIO_OType = GPIO_OType_PP;
	gpioInit.GPIO_PuPd = GPIO_PuPd_NOPULL;
	gpioInit.GPIO_Pin = GPIO_Pin_15;
	GPIO_Init(GPIOG, &gpioInit);

	vTaskDelay(1000/portTICK_RATE_MS);	// power up delay for display

	// Generate RESET pulse
	GPIOG->BSRRH = GPIO_Pin_15;
	vTaskDelay(10/portTICK_RATE_MS);
	GPIOG->BSRRL = GPIO_Pin_15;
	vTaskDelay(10/portTICK_RATE_MS);

	/*-- FSMC Configuration ------------------------------------------------------*/
	fsmcTiming.FSMC_AddressSetupTime = 0;
	fsmcTiming.FSMC_AddressHoldTime = 0;
	fsmcTiming.FSMC_DataSetupTime = 200; // display is very slow as long as PLL is not set
	fsmcTiming.FSMC_BusTurnAroundDuration = 0;
	fsmcTiming.FSMC_CLKDivision = 0;
	fsmcTiming.FSMC_DataLatency = 0;
	fsmcTiming.FSMC_AccessMode = FSMC_AccessMode_A;

	fsmcInit.FSMC_Bank = FSMC_Bank1_NORSRAM1;
	fsmcInit.FSMC_DataAddressMux = FSMC_DataAddressMux_Disable;
	fsmcInit.FSMC_MemoryType = FSMC_MemoryType_SRAM;
	fsmcInit.FSMC_MemoryDataWidth = FSMC_MemoryDataWidth_16b;
	fsmcInit.FSMC_BurstAccessMode = FSMC_BurstAccessMode_Disable;
	fsmcInit.FSMC_AsynchronousWait = FSMC_AsynchronousWait_Disable;
	fsmcInit.FSMC_WaitSignalPolarity =	FSMC_WaitSignalPolarity_Low;
	fsmcInit.FSMC_WrapMode = FSMC_WrapMode_Disable;
	fsmcInit.FSMC_WaitSignalActive = FSMC_WaitSignalActive_BeforeWaitState;
	fsmcInit.FSMC_WriteOperation = FSMC_WriteOperation_Enable;
	fsmcInit.FSMC_WaitSignal = FSMC_WaitSignal_Disable;
	fsmcInit.FSMC_ExtendedMode = FSMC_ExtendedMode_Disable;
	fsmcInit.FSMC_WriteBurst = FSMC_WriteBurst_Disable;
	fsmcInit.FSMC_ReadWriteTimingStruct = &fsmcTiming;
	fsmcInit.FSMC_WriteTimingStruct = &fsmcTiming;

	FSMC_NORSRAMInit(&fsmcInit);

	/*!< Enable FSMC Bank1_SRAM1 Bank */
	FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM1, ENABLE);

	*TFT_CMD_REG = 0x01; // soft reset
	vTaskDelay(10/portTICK_RATE_MS);

	// init display PLL to 110Mhz
	*TFT_CMD_REG = TFT_REG_SET_PLL_MN;
	*TFT_DATA_REG = 0x20;
	*TFT_DATA_REG = 0x02;
	*TFT_DATA_REG = 0x04;

	*TFT_CMD_REG = TFT_REG_SET_PLL;
	*TFT_DATA_REG = 0x01;
	vTaskDelay(100/portTICK_RATE_MS); // wait for PLL lock
	*TFT_CMD_REG = TFT_REG_SET_PLL;
	*TFT_DATA_REG = 0x03;

	FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM1, DISABLE);

	fsmcTimingWrite = fsmcTiming;
	fsmcTimingWrite.FSMC_DataSetupTime = 2;	// write can be very fast :)

	fsmcTiming.FSMC_DataSetupTime = 3; // reading is not so fast :/

	fsmcInit.FSMC_ExtendedMode = FSMC_ExtendedMode_Enable;
	fsmcInit.FSMC_WriteTimingStruct = &fsmcTimingWrite;

	FSMC_NORSRAMInit(&fsmcInit);

	FSMC_NORSRAMCmd(FSMC_Bank1_NORSRAM1, ENABLE);

	vTaskDelay(2000/portTICK_RATE_MS);

	regInit();
}

void TFTSetBrightness(uint8_t level)
{
	// Set PWM config
	*TFT_CMD_REG = TFT_REG_SET_PWM_CONF;
	*TFT_DATA_REG = 0x0E;  // frequency
	*TFT_DATA_REG = level;  //  PWM duty cycle
	*TFT_DATA_REG = 0x01;  //  PWM enabled & PWM controlled by host
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x00;
	*TFT_DATA_REG = 0x0F;
}

void TFTFillAll(uint32_t rgb)
{
	TFTSetPixelArea(TFTDisplayRect, 0);

	*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;

	for(uint16_t y = 0; y < 240; y++)
	{
		for(uint16_t x = 0; x < 320; x++)
		{
			TFTWritePixel(rgb);
		}
	}
}

void TFTFillRect(RectU16 rect, uint32_t rgb)
{
	TFTSetPixelArea(rect, 0);

	*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;

	for(uint32_t x = 0; x < RectU16Size(&rect); x++)
	{
		TFTWritePixel(rgb);
	}
}

void TFTDrawRect(RectU16 rect, uint32_t rgb)
{
	TFTFillRect(RectU16Coordinates(rect.a.x, rect.a.y, rect.b.x, rect.a.y), rgb);
	TFTFillRect(RectU16Coordinates(rect.b.x, rect.a.y, rect.b.x, rect.b.y), rgb);
	TFTFillRect(RectU16Coordinates(rect.a.x, rect.b.y, rect.b.x, rect.b.y), rgb);
	TFTFillRect(RectU16Coordinates(rect.a.x, rect.a.y, rect.a.x, rect.b.y), rgb);
}

void TFTBox()
{
	TFTSetPixelArea(TFTDisplayRect, 0);

	*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;

	for(uint32_t y = 0; y < 240; y++)
	{
		for(uint32_t x = 0; x < 320; x++)
		{
			if(y == 0 || y == 239 || x == 0 || x == 319)
				TFTWritePixel(0xFF0000);
			else
				TFTWritePixel(0);
		}
	}
}

void TFTPattern()
{
	uint32_t rgb;

	TFTSetPixelArea(TFTDisplayRect, 0);

	*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;

	for(uint32_t y = 0; y < 240; y++)
	{
		for(uint32_t x = 0; x < 320; x++)
		{
			if(y < 80)
			{
				rgb = x << 14;
			}
			else if(y < 160)
			{
				rgb = x << 6;
			}
			else
			{
				rgb = x >> 2;
			}

			TFTWritePixel(rgb);
		}
	}

}

/*
void TFTDrawRectangle(uint32_t x1, uint32_t y1, uint32_t x2, uint32_t y2)
{
	// *TFT_CMD_REG=
}*/


/*
void TFTDrawTestImage(uint32_t x1, uint32_t y1, uint8_t rotation)
{
	uint16_t i;
	setPixelArea( x1,  y1, 30, 30, rotation);
	// Write Image
	*TFT_CMD_REG = TFT_REG_WRITE_MEMORY_START;
	for (i = 0; i < 0x384; i++)
	{
		*TFT_DATA_REG = Testbild_Bot[i];  // bmp to C program says it produces RGB565 format already
	}
}
*/





/*
void TFT;
int i;
	i= display.touchX;
*/


