/**
 ******************************************************************************
 * @file    usb_hcd.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   Host Interface Layer
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software 
 * distributed under the License is distributed on an "AS IS" BASIS, 
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include "usb_core.h"
#include "usb_hcd.h"
#include "usb_conf.h"
#include "usb_bsp.h"

#include "FreeRTOS.h"
#include "task.h"

#include "errors.h"
#include "util/console.h"

#include <string.h>

/**
 * @brief  HCD_Init 
 *         Initialize the HOST portion of the driver.
 * @param  pdev: Selected device
 * @param  base_address: OTG base address
 * @retval Status
 */
uint32_t HCD_Init(USB_OTG_CORE_HANDLE *pdev, USB_OTG_CORE_ID_TypeDef coreID)
{
	uint8_t i = 0;
	pdev->host.ConnSts = 0;
	pdev->host.hcdCb = 0;
	pdev->host.hEventQueue = xQueueCreate(USB_OTG_MAX_TX_FIFOS, sizeof(uint32_t));

	for(i = 0; i < USB_OTG_MAX_TX_FIFOS; i++)
	{
		pdev->host.hc[i].errCnt = 0;
//		pdev->host.XferCnt[i] = 0;
		pdev->host.hc[i].hcStatus = HC_IDLE;
		pdev->host.hc[i].hURBMutex = xSemaphoreCreateMutex();
		vSemaphoreCreateBinary(pdev->host.hc[i].hSync);
	}
	pdev->host.hc[0].max_packet = 8;

	USB_OTG_SelectCore(pdev, coreID);
#ifndef DUAL_ROLE_MODE_ENABLED
	USB_OTG_DisableGlobalInt(pdev);
	USB_OTG_CoreInit(pdev);

	/* Force Host Mode*/
	USB_OTG_SetCurrentMode(pdev, HOST_MODE);
	USB_OTG_CoreInitHost(pdev);
	USB_OTG_EnableGlobalInt(pdev);
#endif

	return 0;
}

/**
 * @brief  HCD_GetCurrentSpeed
 *         Get Current device Speed.
 * @param  pdev : Selected device
 * @retval Status
 */

uint32_t HCD_GetCurrentSpeed(USB_OTG_CORE_HANDLE *pdev)
{
	USB_OTG_HPRT0_TypeDef HPRT0;
	HPRT0.d32 = USB_OTG_READ_REG32(pdev->regs.HPRT0);

	return HPRT0.b.prtspd;
}

/**
 * @brief  HCD_ResetPort
 *         Issues the reset command to device
 * @param  pdev : Selected device
 * @retval Status
 */
uint32_t HCD_ResetPort(USB_OTG_CORE_HANDLE *pdev)
{
	/*
	 Before starting to drive a USB reset, the application waits for the OTG 
	 interrupt triggered by the debounce done bit (DBCDNE bit in OTG_FS_GOTGINT), 
	 which indicates that the bus is stable again after the electrical debounce 
	 caused by the attachment of a pull-up resistor on DP (FS) or DM (LS).
	 */

	USB_OTG_ResetPort(pdev);
	return 0;
}

/**
 * @brief  HCD_IsDeviceConnected
 *         Check if the device is connected.
 * @param  pdev : Selected device
 * @retval Device connection status. 1 -> connected and 0 -> disconnected
 * 
 */
uint32_t HCD_IsDeviceConnected(USB_OTG_CORE_HANDLE *pdev)
{
	return (pdev->host.ConnSts);
}

/**
 * @brief  HCD_GetCurrentFrame 
 *         This function returns the frame number for sof packet
 * @param  pdev : Selected device
 * @retval Frame number
 * 
 */
uint32_t HCD_GetCurrentFrame(USB_OTG_CORE_HANDLE *pdev)
{
	return (USB_OTG_READ_REG32(&pdev->regs.HREGS->HFNUM) & 0xFFFF);
}

/**
 * @brief  HCD_GetHCState 
 *         This function returns the HC Status 
 * @param  pdev: Selected device
 * @retval HC_STATUS
 * 
 */
HC_STATUS HCD_GetHCState(USB_OTG_CORE_HANDLE *pdev, uint8_t ch_num)
{
	return pdev->host.hc[ch_num].hcStatus;
}

/**
 * @brief  HCD_HC_Init 
 *         This function prepare a HC and start a transfer
 * @param  pdev: Selected device
 * @param  hc_num: Channel number 
 * @retval status 
 */
uint32_t HCD_HC_Init(USB_OTG_CORE_HANDLE *pdev, uint8_t hc_num)
{
	return USB_OTG_HC_Init(pdev, hc_num);
}

void URBInit(URB* pURB, uint8_t pipe, uint8_t* pData, uint32_t length, uint32_t flags)
{
	memset(pURB, 0, sizeof(URB));

	pURB->pipe = pipe;
	pURB->pData = pData;
	pURB->length = length;
	pURB->flags = flags;
	pURB->syncTimeout = portMAX_DELAY;
}

void URBSetCallback(URB* pURB, URBCompleteCallback cb, void* pContext)
{
	pURB->cb = cb;
	pURB->pContext = pContext;
}

static uint8_t findHCNum(USB_OTG_CORE_HANDLE* pDev, uint8_t pipe)
{
	uint8_t hcNum;
	for(hcNum = 0; hcNum < USB_OTG_MAX_TX_FIFOS; hcNum++)
	{
		if((pDev->host.channel[hcNum] & 0x7FFF/*HC_USED_MASK*/) == pipe)
			break;
	}

	if(hcNum == USB_OTG_MAX_TX_FIFOS)
		return 0xFF;

	return hcNum;
}

int16_t HCDIsPipeActive(USB_OTG_CORE_HANDLE* pDev, uint8_t pipe)
{
	uint8_t hcNum = findHCNum(pDev, pipe);
	if(hcNum == 0xFF)
		return 0;

	if(pDev->host.hc[hcNum].pURB == 0)
		return 0;
	else
		return 1;
}

uint32_t HCDGetMaxTransferSize(USB_OTG_CORE_HANDLE* pDev, uint8_t pipe)
{
	uint8_t hcNum = findHCNum(pDev, pipe);
	if(hcNum == 0xFF)
		return 0;

	return MAX_PACKET_COUNT * pDev->host.hc[hcNum].max_packet;
}

static void syncURBCallback(URB* pURB, void* pContext)
{
	(void)pURB;

	xSemaphoreHandle* phSync = (xSemaphoreHandle*)pContext;

	xSemaphoreGive(*phSync);
}

int16_t HCDSubmitURB(USB_OTG_CORE_HANDLE* pDev, URB* pURB)
{
	uint32_t hcNum = findHCNum(pDev, pURB->pipe);
	if(hcNum == 0xFF)
		return ERROR_USB_INVALID_PIPE;

	if(pURB->length > MAX_PACKET_COUNT * pDev->host.hc[hcNum].max_packet)
		return ERROR_USB_URB_TOO_BIG;

	pURB->state = URB_QUEUED;

	if(pURB->flags & URB_FLAG_SYNC)
	{
		URB* pLastURB;
		for(pLastURB = pURB; pLastURB->pNext; pLastURB = pLastURB->pNext);

		URBSetCallback(pLastURB, &syncURBCallback, &pDev->host.hc[hcNum].hSync);
		xSemaphoreTake(pDev->host.hc[hcNum].hSync, 0);
	}

	xSemaphoreTake(pDev->host.hc[hcNum].hURBMutex, portMAX_DELAY);

	URB* pCurrentURB = pDev->host.hc[hcNum].pURB;

	if(pCurrentURB)
	{
		// there is already some data being processed on this EP
		for(; pCurrentURB->pNext != 0; pCurrentURB = pCurrentURB->pNext);	// find last URB in list

		pCurrentURB->pNext = pURB;
	}
	else
	{
		// currently no URB in progress on this EP
		pDev->host.hc[hcNum].pURB = pURB;

		USB_OTG_HC_StartXfer(pDev, hcNum);
	}

	xSemaphoreGive(pDev->host.hc[hcNum].hURBMutex);

//	ConsolePrint("Submit: EP: 0x%02X, URB: %d, Len: %d\r\n", pURB->pipe, pURB->state, pURB->length);

	if(pURB->flags & URB_FLAG_SYNC)
	{
		if(xSemaphoreTake(pDev->host.hc[hcNum].hSync, pURB->syncTimeout) == pdFALSE)
		{
			// timeout expired :(
			pURB->cb = 0;
			pURB->state = URB_TIMEOUT;
			xQueueSend(pDev->host.hEventQueue, &hcNum, portMAX_DELAY);

			return ERROR_USB_URB_TIMEOUT;
		}
	}

	return 0;
}

void HCDTask(void* params)
{
	USB_OTG_CORE_HANDLE* pDev = (USB_OTG_CORE_HANDLE*)params;

	uint32_t ch;

	while(1)
	{
		xQueueReceive(pDev->host.hEventQueue, &ch, portMAX_DELAY);

		xSemaphoreTake(pDev->host.hc[ch].hURBMutex, portMAX_DELAY);

		URB* pCurrentURB = pDev->host.hc[ch].pURB;

//		ConsolePrint("EP: 0x%02X, URB: %d, Tx: %d/%d\r\n", pCurrentURB->pipe, pCurrentURB->state, pCurrentURB->transferred, pCurrentURB->length);

		// process callbacks
		if(pCurrentURB->cb)
			(*pCurrentURB->cb)(pCurrentURB, pCurrentURB->pContext);

		if(pCurrentURB->pNext)
		{
			pCurrentURB = pCurrentURB->pNext;

			pDev->host.hc[ch].pURB = pCurrentURB;

			pCurrentURB->transferred = 0;
			pCurrentURB->pNext = 0;
			pCurrentURB->state = URB_QUEUED;

			USB_OTG_HC_StartXfer(pDev, ch);
		}
		else
			pDev->host.hc[ch].pURB = 0;

		xSemaphoreGive(pDev->host.hc[ch].hURBMutex);
	}
}
