/**
 ******************************************************************************
 * @file    usbh_msc_bot.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   This file includes the mass storage related functions
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include "usbh_ioreq.h"
#include "usbh_def.h"
#include "usb_hcd_int.h"

#include "usbh_msc.h"

#include "errors.h"
#include "util/console.h"

#define USB_REQ_BOT_RESET                0xFF
#define USB_REQ_GET_MAX_LUN              0xFE
#define USBH_MSC_BOT_CBW_PACKET_LENGTH    31
#define USBH_MSC_CSW_LENGTH               13

/**
 * @brief  USBH_MSC_BOTReset
 *         This request is used to reset the mass storage device and its
 *         associated interface. This class-specific request shall ready the
 *         device for the next CBW from the host.
 * @param  pdev: Selected device
 * @retval USBH_Status : Status of class request handled.
 */
int16_t USBH_MSC_BOTReset(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	phost->Control.setup.b.bmRequestType = USB_H2D | USB_REQ_TYPE_CLASS | USB_REQ_RECIPIENT_INTERFACE;
	phost->Control.setup.b.bRequest = USB_REQ_BOT_RESET;
	phost->Control.setup.b.wValue.w = 0;
	phost->Control.setup.b.wIndex.w = 0;
	phost->Control.setup.b.wLength.w = 0;

	return USBHCtrlReq(pdev, phost, &phost->Control.setup, 0);
}

int16_t USBH_MSC_BOTTransfer(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost, USBH_CBW_Block* pCBW, uint8_t* pData, USBH_CSW_Block* pCSW)
{
	usbhMsc.bot.transferState = BOT_STATE_SEND_CBW;

	uint32_t dataLength = pCBW->field.CBWTransferLength;
	uint32_t errorCount = 0;

	while(usbhMsc.bot.transferState != BOT_STATE_COMPLETE)
	{
		switch(usbhMsc.bot.transferState)
		{
			case BOT_STATE_SEND_CBW:
			{
				URBInit(&usbhMsc.bot.urb, usbhMsc.setup.bulkOutEp, pCBW->CBWArray, USBH_MSC_BOT_CBW_PACKET_LENGTH, URB_FLAG_SYNC);
				HCDSubmitURB(pdev, &usbhMsc.bot.urb);

				if(dataLength > 0)
				{
					if((pCBW->field.CBWFlags & USB_REQ_DIR_MASK) == USB_D2H)
						usbhMsc.bot.transferState = BOT_STATE_DATA_IN;
					else
						usbhMsc.bot.transferState = BOT_STATE_DATA_OUT;
				}
				else
					usbhMsc.bot.transferState = BOT_STATE_SEND_CSW;
			}
			break;
			case BOT_STATE_DATA_IN:
			{
				uint32_t transferred = 0;
				uint8_t* pDataProc = pData;

				// IN data stage
				while(transferred < dataLength)
				{
					uint32_t transferSize = dataLength - transferred;
					if(transferSize > HCDGetMaxTransferSize(pdev, usbhMsc.setup.bulkInEp))
						transferSize = HCDGetMaxTransferSize(pdev, usbhMsc.setup.bulkInEp);

					URBInit(&usbhMsc.bot.urb, usbhMsc.setup.bulkInEp, pDataProc, transferSize, URB_FLAG_SYNC);
					HCDSubmitURB(pdev, &usbhMsc.bot.urb);
					if(usbhMsc.bot.urb.state == URB_DONE)
					{
						pDataProc += transferSize;
						transferred += transferSize;
						continue;
					}

					if(usbhMsc.bot.urb.state == URB_STALL)
					{
						errorCount++;
						USBH_ClrFeature(pdev, phost, usbhMsc.setup.bulkInEp);
						usbhMsc.bot.transferState = BOT_STATE_SEND_CSW;
						break;
					}
				}

				if(usbhMsc.bot.urb.state == URB_DONE)
					usbhMsc.bot.transferState = BOT_STATE_SEND_CSW;
			}
			break;
			case BOT_STATE_DATA_OUT:
			{
				uint32_t transferred = 0;
				uint8_t* pDataProc = pData;

				// OUT data stage
				while(transferred < dataLength)
				{
					uint32_t transferSize = dataLength - transferred;
					if(transferSize > HCDGetMaxTransferSize(pdev, usbhMsc.setup.bulkOutEp))
						transferSize = HCDGetMaxTransferSize(pdev, usbhMsc.setup.bulkOutEp);

					URBInit(&usbhMsc.bot.urb, usbhMsc.setup.bulkOutEp, pDataProc, transferSize, URB_FLAG_SYNC);
					HCDSubmitURB(pdev, &usbhMsc.bot.urb);
					if(usbhMsc.bot.urb.state == URB_DONE)
					{
						pDataProc += transferSize;
						transferred += transferSize;
						continue;
					}

					if(usbhMsc.bot.urb.state == URB_STALL)
					{
						errorCount++;
						USBH_ClrFeature(pdev, phost, usbhMsc.setup.bulkOutEp);
						usbhMsc.bot.transferState = BOT_STATE_SEND_CSW;
						break;
					}
				}

				if(usbhMsc.bot.urb.state == URB_DONE)
					usbhMsc.bot.transferState = BOT_STATE_SEND_CSW;
			}
			break;
			case BOT_STATE_SEND_CSW:
			{
				// Status stage
				URBInit(&usbhMsc.bot.urb, usbhMsc.setup.bulkInEp, pCSW->CSWArray, USBH_MSC_CSW_LENGTH, URB_FLAG_SYNC);
				HCDSubmitURB(pdev, &usbhMsc.bot.urb);
				if(usbhMsc.bot.urb.state == URB_DONE)
				{
					usbhMsc.bot.transferState = BOT_STATE_COMPLETE;

					if(pCSW->field.CSWStatus == USBH_MSC_FAIL)
						return ERROR_USB_MSC_BOT_ERROR;
				}

				if(usbhMsc.bot.urb.state == URB_STALL)
				{
					errorCount++;
					USBH_ClrFeature(pdev, phost, usbhMsc.setup.bulkInEp);
				}
				else
				{
					if(usbhMsc.bot.urb.transferred != USBH_MSC_CSW_LENGTH || pCSW->field.CSWSignature != USBH_MSC_BOT_CSW_SIGNATURE || pCSW->field.CSWTag != pCBW->field.CBWTag || pCSW->field.CSWStatus == USBH_MSC_PHASE_ERROR)
					{
						// CSW is invalid, perform reset recovery
						errorCount++;
						USBH_MSC_BOTReset(pdev, phost);
						USBH_ClrFeature(pdev, phost, usbhMsc.setup.bulkInEp);
						USBH_ClrFeature(pdev, phost, usbhMsc.setup.bulkOutEp);
						usbhMsc.bot.transferState = BOT_STATE_SEND_CBW;
					}

					if(dataLength > 0 && pCSW->field.CSWDataResidue > 0)
					{
						// we wanted data but did not get them all
						// FIXME: dataResidue is sometimes set but data is valid, hum?!
						ConsolePrint("BOT data length: %d, Residue: %d\r\n", dataLength, pCSW->field.CSWDataResidue);
						/*
						pCBW->field.CBWTransferLength = pCSW->field.CSWDataResidue;
						pData += dataLength - pCSW->field.CSWDataResidue;
						usbhMsc.bot.state = BOT_STATE_SEND_CBW;
						*/
					}
				}

				if(errorCount > USBH_MSC_ERROR_RETRY_LIMIT)
					return ERROR_USB_MSC_BOT_ERROR;
			}
			break;
			default: break;
		}
	}

	return 0;
}
