/**
 ******************************************************************************
 * @file    usbh_msc_scsi.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   This file implements the SCSI commands
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */

/* Includes ------------------------------------------------------------------*/
#include "usbh_msc.h"
#include "usbh_ioreq.h"
#include "usbh_def.h"

#include "usbh_msc.h"

#define OPCODE_TEST_UNIT_READY            0X00
#define OPCODE_READ_CAPACITY10            0x25
#define OPCODE_MODE_SENSE6                0x1A
#define OPCODE_READ10                     0x28
#define OPCODE_WRITE10                    0x2A
#define OPCODE_REQUEST_SENSE              0x03

#define DESC_REQUEST_SENSE                0X00
#define ALLOCATION_LENGTH_REQUEST_SENSE   63
#define XFER_LEN_READ_CAPACITY10           8
#define XFER_LEN_MODE_SENSE6              63

#define MASK_MODE_SENSE_WRITE_PROTECT     0x80
#define MODE_SENSE_PAGE_CONTROL_FIELD     0x00
#define MODE_SENSE_PAGE_CODE              0x3F

#define CBW_CB_LENGTH                     16
#define CBW_LENGTH                        10
#define CBW_LENGTH_TEST_UNIT_READY         6

/**
 * @brief  USBH_MSC_TestUnitReady
 *         Issues 'Test unit ready' command to the device. Once the response
 *         received, it updates the status to upper layer.
 * @param  None
 * @retval Status
 */
uint8_t USBH_MSC_TestUnitReady(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	uint8_t index;
	USBH_MSC_Status_TypeDef status = USBH_MSC_OK;

	if(!HCD_IsDeviceConnected(pdev))
		return USBH_MSC_FAIL;

	usbhMsc.bot.CBWData.field.CBWTransferLength = 0;
	usbhMsc.bot.CBWData.field.CBWFlags = USB_EP_DIR_OUT;
	usbhMsc.bot.CBWData.field.CBWLength = CBW_LENGTH_TEST_UNIT_READY;
	for(index = CBW_CB_LENGTH; index != 0; index--)
	{
		usbhMsc.bot.CBWData.field.CBWCB[index] = 0x00;
	}
	usbhMsc.bot.CBWData.field.CBWCB[0] = OPCODE_TEST_UNIT_READY;

	if(USBH_MSC_BOTTransfer(pdev, phost, &usbhMsc.bot.CBWData, 0, &usbhMsc.bot.CSWData))
	{
		status = USBH_MSC_FAIL;
	}

	return status;
}

/**
 * @brief  USBH_MSC_ReadCapacity10
 *         Issue the read capacity command to the device. Once the response
 *         received, it updates the status to upper layer
 * @param  None
 * @retval Status
 */
uint8_t USBH_MSC_ReadCapacity10(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	uint8_t index;
	USBH_MSC_Status_TypeDef status = USBH_MSC_OK;

	if(!HCD_IsDeviceConnected(pdev))
		return USBH_MSC_FAIL;

	usbhMsc.bot.CBWData.field.CBWTransferLength = XFER_LEN_READ_CAPACITY10;
	usbhMsc.bot.CBWData.field.CBWFlags = USB_EP_DIR_IN;
	usbhMsc.bot.CBWData.field.CBWLength = CBW_LENGTH;
	for(index = CBW_CB_LENGTH; index != 0; index--)
	{
		usbhMsc.bot.CBWData.field.CBWCB[index] = 0x00;
	}
	usbhMsc.bot.CBWData.field.CBWCB[0] = OPCODE_READ_CAPACITY10;

	if(USBH_MSC_BOTTransfer(pdev, phost, &usbhMsc.bot.CBWData, usbhMsc.scsi.dataInBuffer, &usbhMsc.bot.CSWData))
	{
		status = USBH_MSC_FAIL;
	}
	else
	{
		/*assign the capacity*/
		(((uint8_t*)&usbhMsc.storageParams.capacity)[3]) = usbhMsc.scsi.dataInBuffer[0];
		(((uint8_t*)&usbhMsc.storageParams.capacity)[2]) = usbhMsc.scsi.dataInBuffer[1];
		(((uint8_t*)&usbhMsc.storageParams.capacity)[1]) = usbhMsc.scsi.dataInBuffer[2];
		(((uint8_t*)&usbhMsc.storageParams.capacity)[0]) = usbhMsc.scsi.dataInBuffer[3];

		/*assign the page length*/
		(((uint8_t*)&usbhMsc.storageParams.pageLength)[1]) = usbhMsc.scsi.dataInBuffer[6];
		(((uint8_t*)&usbhMsc.storageParams.pageLength)[0]) = usbhMsc.scsi.dataInBuffer[7];
	}

	return status;
}

/**
 * @brief  USBH_MSC_ModeSense6
 *         Issue the Mode Sense6 Command to the device. This function is used
 *          for reading the WriteProtect Status of the Mass-Storage device.
 * @param  None
 * @retval Status
 */
uint8_t USBH_MSC_ModeSense6(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	uint8_t index;
	USBH_MSC_Status_TypeDef status = USBH_MSC_OK;

	if(!HCD_IsDeviceConnected(pdev))
		return USBH_MSC_FAIL;

	usbhMsc.bot.CBWData.field.CBWTransferLength = XFER_LEN_MODE_SENSE6;
	usbhMsc.bot.CBWData.field.CBWFlags = USB_EP_DIR_IN;
	usbhMsc.bot.CBWData.field.CBWLength = CBW_LENGTH;
	for(index = CBW_CB_LENGTH; index != 0; index--)
	{
		usbhMsc.bot.CBWData.field.CBWCB[index] = 0x00;
	}
	usbhMsc.bot.CBWData.field.CBWCB[0] = OPCODE_MODE_SENSE6;
	usbhMsc.bot.CBWData.field.CBWCB[2] = MODE_SENSE_PAGE_CONTROL_FIELD | MODE_SENSE_PAGE_CODE;
	usbhMsc.bot.CBWData.field.CBWCB[4] = XFER_LEN_MODE_SENSE6;

	if(USBH_MSC_BOTTransfer(pdev, phost, &usbhMsc.bot.CBWData, usbhMsc.scsi.dataInBuffer, &usbhMsc.bot.CSWData))
	{
		status = USBH_MSC_FAIL;
	}
	else
	{
		/* Assign the Write Protect status */
		/* If WriteProtect = 0, Writing is allowed
		 If WriteProtect != 0, Disk is Write Protected */
		if(usbhMsc.scsi.dataInBuffer[2] & MASK_MODE_SENSE_WRITE_PROTECT)
		{
			usbhMsc.storageParams.writeProtect = 1;
		}
		else
		{
			usbhMsc.storageParams.writeProtect = 0;
		}
	}

	return status;
}

/**
 * @brief  USBH_MSC_RequestSense
 *         Issues the Request Sense command to the device. Once the response
 *         received, it updates the status to upper layer
 * @param  None
 * @retval Status
 */
uint8_t USBH_MSC_RequestSense(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	USBH_MSC_Status_TypeDef status = USBH_MSC_OK;

	uint8_t index;

	if(!HCD_IsDeviceConnected(pdev))
		return USBH_MSC_FAIL;

	usbhMsc.bot.CBWData.field.CBWTransferLength = ALLOCATION_LENGTH_REQUEST_SENSE;
	usbhMsc.bot.CBWData.field.CBWFlags = USB_EP_DIR_IN;
	usbhMsc.bot.CBWData.field.CBWLength = CBW_LENGTH;
	for(index = CBW_CB_LENGTH; index != 0; index--)
	{
		usbhMsc.bot.CBWData.field.CBWCB[index] = 0x00;
	}
	usbhMsc.bot.CBWData.field.CBWCB[0] = OPCODE_REQUEST_SENSE;
	usbhMsc.bot.CBWData.field.CBWCB[1] = DESC_REQUEST_SENSE;
	usbhMsc.bot.CBWData.field.CBWCB[4] = ALLOCATION_LENGTH_REQUEST_SENSE;

	if(USBH_MSC_BOTTransfer(pdev, phost, &usbhMsc.bot.CBWData, usbhMsc.scsi.dataInBuffer, &usbhMsc.bot.CSWData))
	{
		status = USBH_MSC_FAIL;
	}
	else
	{
		/* Get Sense data*/
		(((uint8_t*)&usbhMsc.storageParams.senseKey)[3]) = usbhMsc.scsi.dataInBuffer[0];
		(((uint8_t*)&usbhMsc.storageParams.senseKey)[2]) = usbhMsc.scsi.dataInBuffer[1];
		(((uint8_t*)&usbhMsc.storageParams.senseKey)[1]) = usbhMsc.scsi.dataInBuffer[2];
		(((uint8_t*)&usbhMsc.storageParams.senseKey)[0]) = usbhMsc.scsi.dataInBuffer[3];
	}

	return status;
}

/**
 * @brief  USBH_MSC_Write10
 *         Issue the write command to the device. Once the response received,
 *         it updates the status to upper layer
 * @param  dataBuffer : DataBuffer contains the data to write
 * @param  address : Address to which the data will be written
 * @param  nbOfbytes : NbOfbytes to be written
 * @retval Status
 */
uint8_t USBH_MSC_Write10(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost, uint8_t *dataBuffer, uint32_t address, uint32_t nbOfbytes)
{
	uint8_t index;
	USBH_MSC_Status_TypeDef status = USBH_MSC_OK;
	uint16_t nbOfPages;

	if(!HCD_IsDeviceConnected(pdev))
		return USBH_MSC_FAIL;

	usbhMsc.bot.CBWData.field.CBWTransferLength = nbOfbytes;
	usbhMsc.bot.CBWData.field.CBWFlags = USB_EP_DIR_OUT;
	usbhMsc.bot.CBWData.field.CBWLength = CBW_LENGTH;
	for(index = CBW_CB_LENGTH; index != 0; index--)
	{
		usbhMsc.bot.CBWData.field.CBWCB[index] = 0x00;
	}
	usbhMsc.bot.CBWData.field.CBWCB[0] = OPCODE_WRITE10;

	/*logical block address*/
	usbhMsc.bot.CBWData.field.CBWCB[2] = (((uint8_t*)&address)[3]);
	usbhMsc.bot.CBWData.field.CBWCB[3] = (((uint8_t*)&address)[2]);
	usbhMsc.bot.CBWData.field.CBWCB[4] = (((uint8_t*)&address)[1]);
	usbhMsc.bot.CBWData.field.CBWCB[5] = (((uint8_t*)&address)[0]);

	/*USBH_MSC_PAGE_LENGTH = 512*/
	nbOfPages = nbOfbytes / usbhMsc.storageParams.pageLength;

	/*Tranfer length */
	usbhMsc.bot.CBWData.field.CBWCB[7] = (((uint8_t *)&nbOfPages)[1]);
	usbhMsc.bot.CBWData.field.CBWCB[8] = (((uint8_t *)&nbOfPages)[0]);

	if(USBH_MSC_BOTTransfer(pdev, phost, &usbhMsc.bot.CBWData, dataBuffer, &usbhMsc.bot.CSWData))
	{
		status = USBH_MSC_FAIL;
	}

	return status;
}

/**
 * @brief  USBH_MSC_Read10
 *         Issue the read command to the device. Once the response received,
 *         it updates the status to upper layer
 * @param  dataBuffer : DataBuffer will contain the data to be read
 * @param  address : Address from which the data will be read
 * @param  nbOfbytes : NbOfbytes to be read
 * @retval Status
 */
uint8_t USBH_MSC_Read10(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost, uint8_t *dataBuffer, uint32_t address, uint32_t nbOfbytes)
{
	uint8_t index;
	USBH_MSC_Status_TypeDef status = USBH_MSC_OK;
	uint16_t nbOfPages;

	if(!HCD_IsDeviceConnected(pdev))
		return USBH_MSC_FAIL;

	usbhMsc.bot.CBWData.field.CBWTransferLength = nbOfbytes;
	usbhMsc.bot.CBWData.field.CBWFlags = USB_EP_DIR_IN;
	usbhMsc.bot.CBWData.field.CBWLength = CBW_LENGTH;

	for(index = CBW_CB_LENGTH; index != 0; index--)
	{
		usbhMsc.bot.CBWData.field.CBWCB[index] = 0x00;
	}

	usbhMsc.bot.CBWData.field.CBWCB[0] = OPCODE_READ10;

	/*logical block address*/

	usbhMsc.bot.CBWData.field.CBWCB[2] = (((uint8_t*)&address)[3]);
	usbhMsc.bot.CBWData.field.CBWCB[3] = (((uint8_t*)&address)[2]);
	usbhMsc.bot.CBWData.field.CBWCB[4] = (((uint8_t*)&address)[1]);
	usbhMsc.bot.CBWData.field.CBWCB[5] = (((uint8_t*)&address)[0]);

	/*USBH_MSC_PAGE_LENGTH = 512*/
	nbOfPages = nbOfbytes / usbhMsc.storageParams.pageLength;

	/*Tranfer length */
	usbhMsc.bot.CBWData.field.CBWCB[7] = (((uint8_t *)&nbOfPages)[1]);
	usbhMsc.bot.CBWData.field.CBWCB[8] = (((uint8_t *)&nbOfPages)[0]);

	if(USBH_MSC_BOTTransfer(pdev, phost, &usbhMsc.bot.CBWData, dataBuffer, &usbhMsc.bot.CSWData))
	{
		status = USBH_MSC_FAIL;
	}

	return status;
}
