/**
 ******************************************************************************
 * @file    usbh_core.c
 * @author  MCD Application Team
 * @version V2.1.0
 * @date    19-March-2012
 * @brief   This file implements the functions for the core state machine process
 *          the enumeration and the control transfer process
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; COPYRIGHT 2012 STMicroelectronics</center></h2>
 *
 * Licensed under MCD-ST Liberty SW License Agreement V2, (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at:
 *
 *        http://www.st.com/software_license_agreement_liberty_v2
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 ******************************************************************************
 */
/* Includes ------------------------------------------------------------------*/

#include "usbh_ioreq.h"
#include "usb_bsp.h"
#include "usbh_hcs.h"
#include "usbh_stdreq.h"
#include "usbh_core.h"
#include "usb_hcd_int.h"

#include <stdio.h>

void USBH_Disconnected(USB_OTG_CORE_HANDLE *pdev);
void USBH_Connected(USB_OTG_CORE_HANDLE *pdev);

USBH_HCD_INT_cb_TypeDef USBH_HCD_INT_cb = { 0, USBH_Connected, USBH_Disconnected, };

static USBH_Status USBH_HandleEnum(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost);

/**
 * @brief  USBH_Connected
 *         USB Connect callback function from the Interrupt.
 * @param  selected device
 * @retval Status
 */
void USBH_Connected(USB_OTG_CORE_HANDLE *pdev)
{
	pdev->host.ConnSts = 1;
}

/**
 * @brief  USBH_Disconnected
 *         USB Disconnect callback function from the Interrupt.
 * @param  selected device
 * @retval Status
 */
void USBH_Disconnected(USB_OTG_CORE_HANDLE *pdev)
{
	pdev->host.ConnSts = 0;
}

/**
 * @brief  USBH_Init
 *         Host hardware and stack initializations
 * @param  class_cb: Class callback structure address
 * @param  usr_cb: User callback structure address
 * @retval None
 */
void USBH_Init(USB_OTG_CORE_HANDLE *pdev, USB_OTG_CORE_ID_TypeDef coreID, USBH_HOST *phost, USBH_Usr_cb_TypeDef *usr_cb)
{
	/* Hardware Init */
	USB_OTG_BSP_Init(pdev);

	/* configure GPIO pin used for switching VBUS power */
	USB_OTG_BSP_ConfigVBUS(0);

	/* Host de-initializations */
	USBH_DeInit(pdev, phost);

	/*Register class and user callbacks */
	phost->class_cb = 0;
	phost->usr_cb = usr_cb;

	/* Start the USB OTG core */
	HCD_Init(pdev, coreID);

	pdev->host.hcdCb = &USBH_HCD_INT_cb;

	/* Upon Init call usr call back */
	phost->usr_cb->Init();

	/* Enable Interrupts */
	USB_OTG_BSP_EnableInterrupt(pdev);
}


void USBH_Reset(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	USB_OTG_ResetPort(pdev);

	USBH_DeInit(pdev, phost);

	uint8_t i = 0;

	for(i = 0; i < USB_OTG_MAX_TX_FIFOS; i++)
	{
		pdev->host.hc[i].errCnt = 0;
		pdev->host.hc[i].hcStatus = HC_IDLE;
	}
	pdev->host.hc[0].max_packet = 8;

#ifndef DUAL_ROLE_MODE_ENABLED
	USB_OTG_DisableGlobalInt(pdev);
	USB_OTG_CoreInit(pdev);

	/* Force Host Mode*/
	USB_OTG_SetCurrentMode(pdev, HOST_MODE);
	USB_OTG_CoreInitHost(pdev);
	USB_OTG_EnableGlobalInt(pdev);
#endif
}

/**
 * @brief  USBH_DeInit
 *         Re-Initialize Host
 * @param  None
 * @retval status: USBH_Status
 */
USBH_Status USBH_DeInit(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	/* Software Init */

	phost->gState = HOST_IDLE;
	phost->gStateBkp = HOST_IDLE;
	phost->EnumState = ENUM_IDLE;

	phost->Control.state = CTRL_SETUP;
	phost->Control.ep0size = USB_OTG_MAX_EP0_SIZE;

	phost->device_prop.address = USBH_DEVICE_ADDRESS_DEFAULT;
	phost->device_prop.speed = HPRT0_PRTSPD_FULL_SPEED;

	USBH_Free_Channel(pdev, phost->Control.hc_num_in);
	USBH_Free_Channel(pdev, phost->Control.hc_num_out);
	return USBH_OK;
}

/**
 * @brief  USBH_Process
 *         USB Host core main state machine process
 * @param  None
 * @retval None
 */
void USBH_Process(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	volatile USBH_Status status = USBH_FAIL;

	/* check for Host port events */
	if((HCD_IsDeviceConnected(pdev) == 0) && (phost->gState != HOST_IDLE))
	{
		if(phost->gState != HOST_DEV_DISCONNECTED)
		{
			phost->gState = HOST_DEV_DISCONNECTED;
		}
	}

	switch(phost->gState)
	{
		case HOST_IDLE:
			if(HCD_IsDeviceConnected(pdev))
			{
				phost->gState = HOST_DEV_ATTACHED;
				USB_OTG_BSP_mDelay(100);
			}
			break;

		case HOST_DEV_ATTACHED:
			phost->usr_cb->DeviceAttached();
			phost->Control.hc_num_out = USBH_Alloc_Channel(pdev, 0x00);
			phost->Control.hc_num_in = USBH_Alloc_Channel(pdev, 0x80);

			/* Reset USB Device */
			if(HCD_ResetPort(pdev) == 0)
			{
				phost->device_prop.speed = HCD_GetCurrentSpeed(pdev);

				phost->gState = HOST_ENUMERATION;

				/* Open Control pipes */
				USBH_Open_Channel(pdev, phost->Control.hc_num_in, phost->device_prop.address, phost->device_prop.speed, EP_TYPE_CTRL,
						phost->Control.ep0size);

				/* Open Control pipes */
				USBH_Open_Channel(pdev, phost->Control.hc_num_out, phost->device_prop.address, phost->device_prop.speed, EP_TYPE_CTRL,
						phost->Control.ep0size);
			}
			break;

		case HOST_ENUMERATION:
		{
			/* Check for enumeration status */
			USBH_Status usbhStatus = USBH_HandleEnum(pdev, phost);
			if(usbhStatus == USBH_OK)
			{
				/* The function shall return USBH_OK when full enumeration is complete */

				/* user callback for end of device basic enumeration */
				phost->usr_cb->EnumerationDone(pdev, phost);

				phost->gState = HOST_USR_INPUT;
			}

			if(usbhStatus == USBH_FAIL)
			{
				USBH_Reset(pdev, phost);

				phost->gState = HOST_IDLE;
			}
		}
			break;

		case HOST_USR_INPUT:
			phost->class_cb = phost->usr_cb->ClassSetup(pdev, phost);

			if(phost->class_cb == 0)
			{
				phost->gState = HOST_DEV_NOT_SUPPORTED;
				phost->usr_cb->DeviceNotSupported();
			}
			else
			{
				phost->gState = HOST_CLASS_INIT;
			}
			break;

		case HOST_CLASS_INIT:
			/* process class standard contol requests state machine */
			status = phost->class_cb->Init(pdev, phost);

			switch(status)
			{
				case USBH_OK:
					phost->gState = HOST_CLASS;
				break;
				case USBH_NOT_SUPPORTED:
				{
					phost->gState = HOST_DEV_NOT_SUPPORTED;
					phost->usr_cb->DeviceNotSupported();
				}
				break;
				default:
					phost->gState = HOST_ERROR_STATE;
				break;
			}
			break;
		case HOST_CLASS:
			/* process class state machine */
			status = phost->class_cb->Machine(pdev, phost);
			switch(status)
			{
				case USBH_OK: break;
				case USBH_NOT_SUPPORTED:
				{
					phost->gState = HOST_DEV_NOT_SUPPORTED;
					phost->usr_cb->DeviceNotSupported();
				}
				break;
				default:
					phost->gState = HOST_ERROR_STATE;
				break;
			}
			break;

		case HOST_DEV_NOT_SUPPORTED:
			break;

		case HOST_ERROR_STATE:
			/* Re-Initilaize Host for new Enumeration */
			USBH_DeInit(pdev, phost);
			phost->usr_cb->DeInit();
			phost->class_cb->DeInit(pdev, &phost->device_prop);
			break;

		case HOST_DEV_DISCONNECTED:
			/* Manage User disconnect operations*/
			phost->usr_cb->DeviceDisconnected();

			/* Re-Initilaize Host for new Enumeration */
			USBH_DeInit(pdev, phost);
			phost->usr_cb->DeInit();
			phost->class_cb->DeInit(pdev, &phost->device_prop);
			USBH_DeAllocate_AllChannel(pdev);
			phost->gState = HOST_IDLE;

			break;

		default:
			break;
	}
}

/**
 * @brief  USBH_HandleEnum
 *         This function includes the complete enumeration process
 * @param  pdev: Selected device
 * @retval USBH_Status
 */
static USBH_Status USBH_HandleEnum(USB_OTG_CORE_HANDLE *pdev, USBH_HOST *phost)
{
	USBH_Status Status = USBH_BUSY;

	switch(phost->EnumState)
	{
		case ENUM_IDLE:
			/* Get Device Desc for only 1st 8 bytes : To get EP0 MaxPacketSize */
			if(USBH_Get_DevDesc(pdev, phost, 8) == 0)
			{
				phost->Control.ep0size = phost->device_prop.Dev_Desc.bMaxPacketSize;

				/* Issue Reset  */
				HCD_ResetPort(pdev);
				phost->EnumState = ENUM_GET_FULL_DEV_DESC;

				/* modify control channels configuration for MaxPacket size */
				USBH_Modify_Channel(pdev, phost->Control.hc_num_out, 0, 0, 0, phost->Control.ep0size);
				USBH_Modify_Channel(pdev, phost->Control.hc_num_in, 0, 0, 0, phost->Control.ep0size);
			}
			else
			{
				return USBH_FAIL;
			}
			break;

		case ENUM_GET_FULL_DEV_DESC:
			/* Get FULL Device Desc  */
			if(USBH_Get_DevDesc(pdev, phost, USB_DEVICE_DESC_SIZE) == 0)
			{
				/* user callback for device descriptor available */
				phost->EnumState = ENUM_SET_ADDR;
			}
			break;

		case ENUM_SET_ADDR:
			/* set address */
			if(USBH_SetAddress(pdev, phost, USBH_DEVICE_ADDRESS) == 0)
			{
				USB_OTG_BSP_mDelay(2);
				phost->device_prop.address = USBH_DEVICE_ADDRESS;

				/* user callback for device address assigned */
				phost->EnumState = ENUM_GET_CFG_DESC;

				/* modify control channels to update device address */
				USBH_Modify_Channel(pdev, phost->Control.hc_num_in, phost->device_prop.address, 0, 0, 0);
				USBH_Modify_Channel(pdev, phost->Control.hc_num_out, phost->device_prop.address, 0, 0, 0);
			}
			break;

		case ENUM_GET_CFG_DESC:
			/* get standard configuration descriptor */
			if(USBH_Get_CfgDesc(pdev, phost, USB_CONFIGURATION_DESC_SIZE) == 0)
			{
				phost->EnumState = ENUM_GET_FULL_CFG_DESC;
			}
			break;

		case ENUM_GET_FULL_CFG_DESC:
			/* get FULL config descriptor (config, interface, endpoints) */
			if(USBH_Get_CfgDesc(pdev, phost, phost->device_prop.Cfg_Desc.wTotalLength) == 0)
			{
				phost->EnumState = ENUM_GET_MFC_STRING_DESC;
			}
			break;

		case ENUM_GET_MFC_STRING_DESC:
		{
			// Check that Manufacturer String is available
			if(phost->device_prop.Dev_Desc.iManufacturer != 0)
			{
				USBH_Get_StringDesc(pdev, phost, phost->device_prop.Dev_Desc.iManufacturer, phost->device_prop.manufacturerString, USB_MAX_STRING_SIZE);
			}
			else
			{
				sprintf((char*)phost->device_prop.manufacturerString, "N/A");
			}

			phost->EnumState = ENUM_GET_PRODUCT_STRING_DESC;
		}
		break;

		case ENUM_GET_PRODUCT_STRING_DESC:
		{
			// Check that Product string is available
			if(phost->device_prop.Dev_Desc.iProduct != 0)
			{
				USBH_Get_StringDesc(pdev, phost, phost->device_prop.Dev_Desc.iProduct, phost->device_prop.productString, USB_MAX_STRING_SIZE);
			}
			else
			{
				sprintf((char*)phost->device_prop.productString, "N/A");
			}

			phost->EnumState = ENUM_GET_SERIALNUM_STRING_DESC;
		}
		break;

		case ENUM_GET_SERIALNUM_STRING_DESC:
		{
			// Check that Serial number string is available
			if(phost->device_prop.Dev_Desc.iSerialNumber != 0)
			{
				USBH_Get_StringDesc(pdev, phost, phost->device_prop.Dev_Desc.iSerialNumber, phost->device_prop.serialNumberString, USB_MAX_STRING_SIZE);
			}
			else
			{
				sprintf((char*)phost->device_prop.serialNumberString, "N/A");
			}

			phost->EnumState = ENUM_SET_CONFIGURATION;
		}
		break;

		case ENUM_SET_CONFIGURATION:
			/* set configuration  (default config) */
			if(USBH_SetCfg(pdev, phost, phost->device_prop.Cfg_Desc.bConfigurationValue) == 0)
			{
				phost->EnumState = ENUM_DEV_CONFIGURED;
			}
			break;

		case ENUM_DEV_CONFIGURED:
			/* user callback for enumeration done */
			Status = USBH_OK;
			break;

		default:
			break;
	}

	return Status;
}
